Package: rdrobust
Type: Package
Title: Robust Data-Driven Statistical Inference in
        Regression-Discontinuity Designs
Version: 4.0.0
Date: 2026-05-15
Authors@R: c(person(given = c("Sebastian"),
                      family = "Calonico",
                      role = c("aut", "cre"),
                      email = "scalonico@ucdavis.edu"),
               person(given = c("Matias", "D."),
                      family = "Cattaneo",
                      role = "aut",
                      email = "matias.d.cattaneo@gmail.com"),
               person(given = c("Max", "H."),
                      family = "Farrell",
                      role = "aut",
                      email = "mhfarrell@gmail.com"),
               person(given = "Rocio",
                      family = "Titiunik",
                      role = "aut",
                      email = "rocio.titiunik@gmail.com"))
Description: Regression-discontinuity (RD) designs are quasi-experimental research designs popular in social, behavioral and natural sciences. The RD design is usually employed to study the (local) causal effect of a treatment, intervention or policy. This package provides tools for data-driven graphical and analytical statistical inference in RD designs: rdrobust() to construct local-polynomial point estimators and robust confidence intervals for average treatment effects at the cutoff in Sharp, Fuzzy and Kink RD settings, rdbwselect() to perform bandwidth selection for the different procedures implemented, and rdplot() to conduct exploratory data analysis (RD plots).
Encoding: UTF-8
Depends: R (>= 3.6.0)
License: GPL-3
URL: https://github.com/rdpackages/rdrobust
BugReports: https://github.com/rdpackages/rdrobust/issues
Imports: ggplot2, MASS
Suggests: broom, gridExtra, knitr, rmarkdown
VignetteBuilder: knitr
NeedsCompilation: no
Packaged: 2026-05-15 18:28:12 UTC; ncalonic
Author: Sebastian Calonico [aut, cre],
  Matias D. Cattaneo [aut],
  Max H. Farrell [aut],
  Rocio Titiunik [aut]
Maintainer: Sebastian Calonico <scalonico@ucdavis.edu>
Repository: CRAN
Date/Publication: 2026-05-16 08:10:02 UTC
Built: R 4.6.0; ; 2026-05-16 12:36:23 UTC; unix

Package: rdrobust
Type: Package
Title: Robust Data-Driven Statistical Inference in
        Regression-Discontinuity Designs
Version: 4.0.0
Date: 2026-05-15
Authors@R: c(person(given = c("Sebastian"),
                      family = "Calonico",
                      role = c("aut", "cre"),
                      email = "scalonico@ucdavis.edu"),
               person(given = c("Matias", "D."),
                      family = "Cattaneo",
                      role = "aut",
                      email = "matias.d.cattaneo@gmail.com"),
               person(given = c("Max", "H."),
                      family = "Farrell",
                      role = "aut",
                      email = "mhfarrell@gmail.com"),
               person(given = "Rocio",
                      family = "Titiunik",
                      role = "aut",
                      email = "rocio.titiunik@gmail.com"))
Description: Regression-discontinuity (RD) designs are quasi-experimental research designs popular in social, behavioral and natural sciences. The RD design is usually employed to study the (local) causal effect of a treatment, intervention or policy. This package provides tools for data-driven graphical and analytical statistical inference in RD designs: rdrobust() to construct local-polynomial point estimators and robust confidence intervals for average treatment effects at the cutoff in Sharp, Fuzzy and Kink RD settings, rdbwselect() to perform bandwidth selection for the different procedures implemented, and rdplot() to conduct exploratory data analysis (RD plots).
Encoding: UTF-8
Depends: R (>= 3.6.0)
License: GPL-3
URL: https://github.com/rdpackages/rdrobust
BugReports: https://github.com/rdpackages/rdrobust/issues
Imports: ggplot2, MASS
Suggests: broom, gridExtra, knitr, rmarkdown
VignetteBuilder: knitr
NeedsCompilation: no
Packaged: 2026-05-15 18:28:12 UTC; ncalonic
Author: Sebastian Calonico [aut, cre],
  Matias D. Cattaneo [aut],
  Max H. Farrell [aut],
  Rocio Titiunik [aut]
Maintainer: Sebastian Calonico <scalonico@ucdavis.edu>
Repository: CRAN
Date/Publication: 2026-05-16 08:10:02 UTC
Built: R 4.5.2; ; 2026-05-16 08:44:31 UTC; unix

export(rdbwselect)
export(rdrobust)
export(rdplot)
S3method(print,rdrobust)
S3method(summary,rdrobust)
S3method(print,summary.rdrobust)
S3method(plot,rdrobust)
S3method(coef,rdrobust)
S3method(vcov,rdrobust)
S3method(print,rdbwselect)
S3method(summary,rdbwselect)
S3method(print,summary.rdbwselect)
S3method(print,rdplot)
S3method(summary,rdplot)
S3method(print,summary.rdplot)
S3method(coef,rdplot)
S3method(broom::tidy,rdrobust)
S3method(broom::glance,rdrobust)
S3method(broom::tidy,rdbwselect)
S3method(broom::glance,rdbwselect)
S3method(broom::glance,rdplot)
import(ggplot2)
importFrom("MASS", "ginv")
importFrom("stats", "lm", "predict")
importFrom("utils", "capture.output")
importFrom("graphics", "abline", "arrows", "lines", "plot", "polygon")
importFrom("stats", "IQR", "complete.cases", "integrate", "median",
              "pnorm", "poly", "qnorm", "quantile", "sd", "var", "qt",
              "coef", "vcov", "setNames")

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(
  collapse  = TRUE,
  comment   = "#>",
  fig.width = 7,
  fig.height = 4.5
)

## ----data---------------------------------------------------------------------
library(rdrobust)
data(rdrobust_RDsenate)
df <- rdrobust_RDsenate
head(df)

## ----data-arg-basic-----------------------------------------------------------
r <- rdrobust(y = vote, x = margin, data = df)
round(r$coef, 3)

## ----data-arg-expr------------------------------------------------------------
r_cov <- rdrobust(y = vote, x = margin,
                  covs = cbind(class, termshouse, termssenate),
                  data = df)
round(r_cov$coef, 3)

# subset via an expression evaluated inside `data`
r_sub <- rdrobust(y = vote, x = margin,
                  subset = year >= 1950, data = df)
round(r_sub$coef, 3)

## ----covs-formula-factor------------------------------------------------------
df$classf <- factor(df$class, labels = c("A", "B", "C"))

r_fm <- rdrobust(y = vote, x = margin,
                 covs = ~ classf + termshouse + I(termssenate^2),
                 data = df)
round(r_fm$coef, 3)

## ----covs-char----------------------------------------------------------------
covs_names <- c("class", "termshouse", "termssenate")
r_ch <- rdrobust(y = vote, x = margin, covs = covs_names, data = df)
round(r_ch$coef, 3)

## ----covs-matrix--------------------------------------------------------------
Z <- with(df, cbind(class, termshouse, termssenate))
r_mat <- rdrobust(y = vote, x = margin, covs = Z, data = df)
round(r_mat$coef, 3)

## ----rdplot, fig.cap="RD plot: U.S. Senate incumbency advantage."-------------
rdplot(y = vote, x = margin, data = df,
       title   = "Senate: incumbency advantage",
       x.label = "Vote margin (t)",
       y.label = "Vote share (t+2)")

## ----hc-variants--------------------------------------------------------------
hc_fits <- lapply(c("nn", "hc0", "hc1", "hc2", "hc3"), function(v) {
  rdrobust(y = vote, x = margin, vce = v, data = df)
})
names(hc_fits) <- c("NN", "HC0", "HC1", "HC2", "HC3")

hc_tbl <- do.call(rbind, lapply(names(hc_fits), function(nm) {
  r <- hc_fits[[nm]]
  data.frame(vce    = nm,
             coef   = round(r$coef[2], 3),   # bias-corrected
             se_rb  = round(r$se[3],   3),   # robust SE
             pval   = round(r$pv[3],   3))
}))
print(hc_tbl, row.names = FALSE)

## ----cluster-counts-----------------------------------------------------------
cat("Number of clusters:", length(unique(df$state)), "\n")

## ----crv-variants-------------------------------------------------------------
r_cr1 <- rdrobust(y = vote, x = margin, cluster = state, vce = "cr1", data = df)
r_cr2 <- rdrobust(y = vote, x = margin, cluster = state, vce = "cr2", data = df)
r_cr3 <- rdrobust(y = vote, x = margin, cluster = state, vce = "cr3", data = df)

cl_tbl <- data.frame(
  vce    = c("CR1", "CR2", "CR3"),
  coef   = c(r_cr1$coef[2],  r_cr2$coef[2],  r_cr3$coef[2]),
  se_rb  = c(r_cr1$se[3],    r_cr2$se[3],    r_cr3$se[3]),
  pval   = c(r_cr1$pv[3],    r_cr2$pv[3],    r_cr3$pv[3])
)
cl_tbl[, 2:4] <- round(cl_tbl[, 2:4], 3)
print(cl_tbl, row.names = FALSE)

## ----cr3-summary--------------------------------------------------------------
summary(r_cr3)

## ----plot-basic, fig.cap="Binscatter with local polynomial fits and 95% CI bands."----
plot(r_cr1, df$vote, df$margin,
     title   = "Senate: incumbency advantage",
     x.label = "Vote margin (t)",
     y.label = "Vote share (t+2)")

## ----plot-effect, fig.height=6, fig.cap="Main plot with RD effect panel."-----
plot(r_cr1, df$vote, df$margin,
     show_effect = TRUE,
     title       = "Senate: incumbency advantage",
     x.label     = "Vote margin (t)",
     y.label     = "Vote share (t+2)")

## ----plot-cr3, fig.cap="CR3 clustered variance, with effect panel."-----------
plot(r_cr3, df$vote, df$margin,
     show_effect = TRUE,
     title       = "Senate: CR3 clustered SE",
     x.label     = "Vote margin (t)",
     y.label     = "Vote share (t+2)")

## ----rdplot-hide--------------------------------------------------------------
out <- rdplot(y = vote, x = margin, c = 0,
              hide = TRUE, ci = 95, data = df)
str(out$vars_bins, give.attr = FALSE)
str(out$vars_poly, give.attr = FALSE)
out$coef

## ----rdplot-manual, fig.cap="Custom RD plot built from rdplot() output."------
library(ggplot2)
poly_l <- subset(out$vars_poly, rdplot_x <  0)
poly_r <- subset(out$vars_poly, rdplot_x >= 0)

ggplot() +
  geom_vline(xintercept = 0, linewidth = 0.4) +
  geom_errorbar(data = out$vars_bins,
                aes(x = rdplot_mean_bin,
                    ymin = rdplot_ci_l, ymax = rdplot_ci_r),
                color = "grey60", width = 0) +
  geom_point(data = out$vars_bins,
             aes(x = rdplot_mean_bin, y = rdplot_mean_y),
             color = "grey40", size = 1.5) +
  geom_line(data = poly_l, aes(x = rdplot_x, y = rdplot_y)) +
  geom_line(data = poly_r, aes(x = rdplot_x, y = rdplot_y)) +
  theme_bw() +
  labs(x = "Vote margin (t)", y = "Vote share (t+2)",
       title = "Senate: regression function fit")

## ----rdplot-from-coef---------------------------------------------------------
eval_fit <- function(xs, coef, c)
  sapply(xs, function(xi) sum(coef * (xi - c) ^ (0:(length(coef) - 1))))

xs <- seq(-50, 50, length.out = 5)
data.frame(
  x       = xs,
  y_left  = ifelse(xs <  0, eval_fit(xs, out$coef[, "Left"],  0), NA),
  y_right = ifelse(xs >= 0, eval_fit(xs, out$coef[, "Right"], 0), NA)
)

## ----broom-tidy, eval=requireNamespace("broom", quietly=TRUE)-----------------
library(broom)
tidy(r_cr3)

## ----broom-glance, eval=requireNamespace("broom", quietly=TRUE)---------------
glance(r_cr3)

## ----broom-collect, eval=requireNamespace("broom", quietly=TRUE)--------------
specs <- list(CR1 = r_cr1, CR2 = r_cr2, CR3 = r_cr3)

# Robust BC estimates side-by-side
res <- do.call(rbind, lapply(names(specs), function(nm) {
  cbind(spec = nm, tidy(specs[[nm]])[3, ])  # row 3 = robust
}))
res[, c("spec", "estimate", "std.error", "conf.low", "conf.high")]

## ----broom-glance-compare, eval=requireNamespace("broom", quietly=TRUE)-------
gl <- do.call(rbind, lapply(names(specs), function(nm) {
  cbind(spec = nm, glance(specs[[nm]]))
}))
gl[, c("spec", "nobs.effective.left", "nobs.effective.right",
       "h.left", "vce", "clusters.left", "clusters.right")]

## ----broom-hc-vs-cl, eval=requireNamespace("broom", quietly=TRUE)-------------
r_hc3 <- rdrobust(y = vote, x = margin, vce = "hc3", data = df)

all_specs <- list("HC3"  = r_hc3,
                  "CR1"  = r_cr1,
                  "CR2"  = r_cr2,
                  "CR3"  = r_cr3)

comparison <- do.call(rbind, lapply(names(all_specs), function(nm) {
  g <- glance(all_specs[[nm]])
  t <- tidy(all_specs[[nm]])[3, ]   # robust row
  data.frame(spec      = nm,
             estimate  = round(t$estimate, 3),
             std.error = round(t$std.error, 3),
             ci        = paste0("[", round(t$conf.low, 3), ", ", round(t$conf.high, 3), "]"),
             h         = round(g$h.left, 3),
             clusters  = g$clusters.left)
}))
print(comparison, row.names = FALSE)

## ----table-fits---------------------------------------------------------------
covs_form <- ~ class + termshouse + termssenate

m1 <- rdrobust(y = vote, x = margin, data = df)
m2 <- rdrobust(y = vote, x = margin, covs = covs_form, data = df)
m3 <- rdrobust(y = vote, x = margin, vce = "cr1", cluster = state, data = df)
m4 <- rdrobust(y = vote, x = margin, vce = "cr1", cluster = state,
               covs = covs_form, data = df)

## ----table-row-builder--------------------------------------------------------
rdr_row <- function(fit, label) {
  data.frame(
    spec      = label,
    estimate  = fit$coef[1, 1],                                 # tau_conv
    rbc_ci    = sprintf("[%.3f, %.3f]", fit$ci[3, 1], fit$ci[3, 2]),
    rbc_pval  = fit$pv[3, 1],
    h         = fit$bws[1, 1],                                  # h_left
    N_eff     = fit$N_h[1] + fit$N_h[2],
    clusters  = if (!is.null(fit$n_clust)) sum(fit$n_clust) else NA_integer_
  )
}

tab <- rbind(
  rdr_row(m1, "sharp"),
  rdr_row(m2, "sharp + covs"),
  rdr_row(m3, "cluster"),
  rdr_row(m4, "cluster + covs")
)

## ----table-kable--------------------------------------------------------------
knitr::kable(
  tab,
  digits   = 3,
  col.names = c("Specification", "Estimate", "Robust 95% CI",
                "RBC p-value", "h", "Eff. N", "Clusters"),
  caption  = "RD estimates: conventional point estimate with RBC confidence interval."
)

## ----table-paper--------------------------------------------------------------
paper <- t(tab[, c("estimate", "rbc_ci", "h", "N_eff")])
colnames(paper) <- tab$spec
rownames(paper) <- c("RD effect (conv.)", "Robust 95% CI", "h", "Eff. N")

knitr::kable(
  paper,
  caption = "Senate incumbency: RD estimates across specifications."
)

## ----table-latex, eval=FALSE--------------------------------------------------
# tex <- knitr::kable(
#   paper,
#   format    = "latex",
#   booktabs  = TRUE,
#   caption   = "RD estimates for Senate incumbency.",
#   label     = "tab:rdrobust"
# )
# writeLines(tex, "rdrobust_table.tex")

## ----session------------------------------------------------------------------
sessionInfo()

# Normalize the `covs` argument into a numeric vector/matrix.
#
# Accepted forms:
#   - one-sided formula (e.g. `~ z1 + z2 + I(z3^2)`):
#       processed with model.matrix; factors expand to contrasts; the
#       intercept column is dropped. Symbols not in `data` fall through
#       to `caller`.
#   - character vector of column names (requires `data = `):
#       selected as `data[, covs]` and coerced to a matrix.
#   - anything else (numeric vector, matrix, data.frame, NULL):
#       returned unchanged.
.rdrobust_resolve_covs <- function(covs, data, caller) {
  if (is.null(covs)) return(NULL)
  if (is.character(covs) && length(covs) == 0)
    stop("`covs` is empty. Supply at least one column name, or omit `covs`.",
         call. = FALSE)
  if ((is.matrix(covs) || is.data.frame(covs)) && ncol(covs) == 0)
    stop("`covs` has zero columns. Supply at least one covariate, or omit `covs`.",
         call. = FALSE)
  if (inherits(covs, "formula")) {
    environment(covs) <- caller
    mf <- stats::model.frame(covs,
                             data = if (!is.null(data)) data else caller,
                             na.action = stats::na.pass)
    out <- stats::model.matrix(covs, data = mf)
    if (ncol(out) >= 1 && colnames(out)[1] == "(Intercept)")
      out <- out[, -1, drop = FALSE]
    return(out)
  }
  if (is.character(covs)) {
    if (is.null(data))
      stop("Character `covs` requires `data = ` to resolve column names. ",
           "Pass a matrix or use `covs = df[, cols]` instead.", call. = FALSE)
    missing_cols <- setdiff(covs, colnames(data))
    if (length(missing_cols))
      stop("Column(s) not found in `data`: ",
           paste(sQuote(missing_cols), collapse = ", "), call. = FALSE)
    return(data.matrix(data[, covs, drop = FALSE]))
  }
  if (!(is.numeric(covs) || is.data.frame(covs)))
    stop("'covs' must be a one-sided formula, character vector of column names, ",
         "numeric matrix/vector, or data frame.", call. = FALSE)
  covs
}

# Shared input-validation helpers used by rdrobust / rdbwselect / rdplot.
# Each entry-point validates auxiliary-vector lengths against length(x)
# BEFORE subset filtering, so wrong-length inputs error explicitly instead
# of being silently recycled by `[subset]` / `[na.ok]` indexing.
.rdrobust_check_length <- function(arg, name, n) {
  if (is.null(arg)) return(invisible())
  m <- if (is.matrix(arg) || is.data.frame(arg)) nrow(arg) else length(arg)
  if (m != n)
    stop(sprintf("'%s' must have %s equal to length(x) (got %d, expected %d).",
                 name,
                 if (is.matrix(arg) || is.data.frame(arg)) "nrow" else "length",
                 m, n),
         call. = FALSE)
}

.rdrobust_check_subset <- function(subset, n) {
  if (is.null(subset)) return(invisible())
  if (is.logical(subset)) {
    if (length(subset) != n)
      stop(sprintf("Logical 'subset' must have length equal to length(x) (got %d, expected %d).",
                   length(subset), n),
           call. = FALSE)
  } else if (is.numeric(subset)) {
    if (any(!is.finite(subset)) || any(subset < 1) || any(subset > n) || any(subset != round(subset)))
      stop(sprintf("Numeric 'subset' must contain integer indices in 1..%d.", n),
           call. = FALSE)
  } else {
    stop("'subset' must be logical or integer.", call. = FALSE)
  }
}


# Fast equivalent of `split(seq_along(C), C)` for an atomic cluster vector.
# R's `split.default` calls `as.factor(C)` every invocation (~12ms for a
# 50k integer vector), which dominates the CR1/CR2/CR3 variance loop across
# the ~14 BW-pilot calls per rdrobust. This avoids that by sorting once
# and slicing on consecutive-equal runs; works for integer, numeric, and
# character clusters alike. Returns an UNNAMED list -- rdrobust_vce only
# iterates by index, so the level names are not needed.
.rdrobust_vander <- function(u, p) {
  # Q1: build the Vandermonde matrix [1, u, u^2, ..., u^p] by successive
  # multiplication. 3-5x faster than outer(u, 0:p, `^`) for small p; same
  # allocation cost. Called from rdrobust_bw + rdrobust.R hot paths.
  n <- length(u)
  if (p < 1) return(matrix(1, n, 1))
  out <- matrix(1, n, p + 1)
  for (j in 2:(p + 1)) out[, j] <- out[, j - 1] * u
  out
}

.rdrobust_cluster_idx <- function(C) {
  if (is.factor(C)) return(split(seq_along(C), C))
  n <- length(C)
  if (n == 0) return(list())
  ord <- order(C)
  C_s <- C[ord]
  chg <- c(TRUE, C_s[-1L] != C_s[-n])
  brk <- c(which(chg), n + 1L)
  g <- length(brk) - 1L
  out <- vector("list", g)
  for (i in seq_len(g)) out[[i]] <- ord[brk[i]:(brk[i + 1L] - 1L)]
  out
}


# Resolve a set of bare-name args against `data`, returning a named list.
# Used by rdrobust / rdbwselect / rdplot to enable `y = vote, x = margin,
# data = df` style calls. `mc` and `caller` must be supplied by the caller
# (they're frame-sensitive). Only the args actually present in `mc` are
# resolved; the rest are returned as NULL so callers can keep their
# default. The function deliberately does NOT touch the entry-point
# argument promises, so `y = vote` works even when `vote` is a column of
# `data` that doesn't exist in the caller environment.
.rdrobust_resolve_data <- function(mc, data, caller, args) {
  out <- setNames(vector("list", length(args)), args)
  for (a in args) {
    if (a %in% names(mc) && !is.null(mc[[a]]))
      out[[a]] <- eval(mc[[a]], envir = data, enclos = caller)
  }
  out
}

qrXXinv = function(x, ...) {
  G <- crossprod(x)
  R <- try(chol(G), silent = TRUE)
  if (inherits(R, "try-error")) ginv(G) else chol2inv(R)
}

rdrobust_kweight = function(X, c,  h,  kernel){
  u = (X-c)/h
  if (kernel %in% c("epanechnikov", "epa")) {
    w = (0.75*(1-u^2)*(abs(u)<=1))/h
  } else if (kernel %in% c("uniform", "uni")) {
    w = (0.5*(abs(u)<=1))/h
  } else {
    w = ((1-abs(u))*(abs(u)<=1))/h
  }
  return(w)	
}

rdrobust_res = function(X, y, T, Z, m, hii, vce, matches, dups, dupsid, d, crv3=FALSE, crv2=FALSE, has_cluster=FALSE) {
  n = length(y)
  dT=dZ=0
  if (!is.null(T)) dT = 1
  if (!is.null(Z)) dZ = ncol(Z)
  res = matrix(NA,n,1+dT+dZ)

  if (vce=="nn") {
    for (pos in 1:n) {
      rpos = dups[pos] - dupsid[pos]
      lpos = dupsid[pos] - 1
      while (lpos+rpos < min(c(matches,n-1))) {
        if (pos-lpos-1 <= 0) rpos = rpos + dups[pos+rpos+1]
        else if (pos+rpos+1>n) lpos = lpos + dups[pos-lpos-1]
        else if ((X[pos]-X[pos-lpos-1]) > (X[pos+rpos+1]-X[pos])) rpos = rpos + dups[pos+rpos+1]
        else if ((X[pos]-X[pos-lpos-1]) < (X[pos+rpos+1]-X[pos])) lpos = lpos + dups[pos-lpos-1]
        else {
          rpos = rpos + dups[pos+rpos+1]
          lpos = lpos + dups[pos-lpos-1]
        }
      }
      ind_J = max(c(0,(pos-lpos))):min(c(n,(pos+rpos)))
      y_J   = sum(y[ind_J])-y[pos]
      Ji = length(ind_J)-1
      res[pos,1] = sqrt(Ji/(Ji+1))*(y[pos] - y_J/Ji)
      if (!is.null(T)) {
        T_J = sum(T[ind_J])-T[pos]
        res[pos,2] = sqrt(Ji/(Ji+1))*(T[pos] - T_J/Ji)
      }
      if (!is.null(Z)) {
        for (i in 1:dZ) {
          Z_J = sum(Z[ind_J,i])-Z[pos,i]
          res[pos,1+dT+i] = sqrt(Ji/(Ji+1))*(Z[pos,i] - Z_J/Ji)
        }
      }
    }
  }
  else if (crv3 | crv2) {
    # CRV2/CRV3 mode: return raw (unweighted) residuals.
    # Cluster-level hat-matrix adjustment is handled in rdrobust_vce.
    res[,1] = y - m[,1]
    if (dT==1) res[,2] = T - m[,2]
    if (dZ>0) {
      for (i in 1:dZ) {
        res[,1+dT+i] = Z[,i] - m[,1+dT+i]
      }
    }
  }
  else {
    if (vce=="hc0") w = 1
    else if (vce=="hc1") w = if (has_cluster) 1 else sqrt(n/(n-d))
    else if (vce=="hc2") w = sqrt(1/pmax(1-hii, 1e-8))
    else                 w =      1/pmax(1-hii, 1e-8)
    res[,1] = w*(y-m[,1])
    if (dT==1) res[,2] = w*(T-m[,2])
    if (dZ>0) res[,(2+dT):(1+dT+dZ)] = w*(Z-m[,(2+dT):(1+dT+dZ)])
  }
  return(res)
}


rdrobust_bw = function(Y, X, T, Z, C, W, c, o, nu, o_B, h_V, h_B, scale, vce, nnmatch, kernel, dups, dupsid, covs_drop_coll, ginv.tol, vcache = NULL){
  # T1 (2026-05-12): the V-fit depends only on (o, nu) when h_V is held
  # fixed across calls on the same side. Callers (rdbwselect.R and
  # rdrobust.R) pass a per-side environment via `vcache` to share results
  # across the 6-18 pilot calls per rdrobust invocation. The B-fit
  # below still runs each call (its h_B differs per stage).
  dT = dZ = dC = 0
  crv3 = (vce=="crv3") & !is.null(C)
  crv2 = (vce=="crv2") & !is.null(C)
  key <- paste0(o, "_", nu)
  hit <- !is.null(vcache) && exists(key, envir = vcache, inherits = FALSE)
  if (hit) {
    cached <- get(key, envir = vcache, inherits = FALSE)
    V_V    <- cached$V_V
    BConst <- cached$BConst
    s      <- cached$s
    if (!is.null(T)) dT <- 1
    if (!is.null(Z)) dZ <- ncol(Z)
  } else {
  w = rdrobust_kweight(X, c, h_V, kernel)
  if (!is.null(W)) w = W*w

  ind_V = w> 0; eY = Y[ind_V];eX = X[ind_V];eW = w[ind_V]
  n_V = sum(ind_V)
  D_V = eY
  R_V = .rdrobust_vander(as.numeric(eX - c), o)
  invG_V = qrXXinv(R_V*sqrt(eW))
  s = 1
  eT=eC=eZ=NULL
  if (!is.null(T)) {
    dT = 1
    eT = T[ind_V]
    D_V = cbind(D_V,eT)
  }
  if (!is.null(Z)) {
    dZ = ncol(Z)
    eZ = Z[ind_V,,drop=FALSE]
    D_V = cbind(D_V,eZ)
    U = crossprod(R_V*eW,D_V)
    ZWD  = crossprod(eZ*eW,D_V)
    colsZ = (2+dT):max(c(2+dT+dZ-1,(2+dT)))
    UiGU =  crossprod(matrix(U[,colsZ],nrow=o+1),invG_V%*%U)
    ZWZ = ZWD[,colsZ] - UiGU[,colsZ]
    ZWY = ZWD[,1:(1+dT)] - UiGU[,1:(1+dT)]
    if (covs_drop_coll==1) {
      gamma = ginv(ZWZ, tol=ginv.tol)%*%ZWY
      }
    else {
      gamma = chol2inv(chol(ZWZ))%*%ZWY
    }
    s = c(1 , -gamma[,1])
  }
  if (!is.null(C)) {
    dC = 1
    eC =  C[ind_V]
  }
  beta_V = invG_V%*%crossprod(R_V*eW,D_V)
  if (is.null(Z) & !is.null(T)) {
    tau_Y = c(factorial(nu)*beta_V[nu+1,1])
    tau_T = c(factorial(nu)*beta_V[nu+1,2])
    s = c(1/tau_T , -(tau_Y/tau_T^2))
  }
  if (!is.null(Z) & !is.null(T)) {
    s_T = c(1 , -gamma[,2])
    tau_Y = c(factorial(nu)*t(s)%*%  c(beta_V[nu+1,1],beta_V[nu+1,colsZ]))
    tau_T = c(factorial(nu)*t(s_T)%*%c(beta_V[nu+1,2],beta_V[nu+1,colsZ]))
    s = c(1/tau_T , -(tau_Y/tau_T^2) , -(1/tau_T)*gamma[,1] + (tau_Y/tau_T^2)*gamma[,2])
  }
  dups_V=dupsid_V=predicts_V=hii=0

  if (vce=="nn") {
    dups_V   = dups[ind_V]
    dupsid_V = dupsid[ind_V]
  }

  if ((vce=="hc0" | vce=="hc1" | vce=="hc2" | vce=="hc3") & !crv3) {
    predicts_V = R_V%*%beta_V
    if (vce=="hc2" | vce=="hc3") {
      hii = rowSums((R_V%*%invG_V)*(R_V*eW))
    }
  } else if (crv3 | crv2) {
    predicts_V = R_V%*%beta_V
  }

        sqrtRX_V  = if (crv3 | crv2) R_V * sqrt(eW) else NULL
        invG_V_c  = if (crv3 | crv2) invG_V else NULL
        res_V = rdrobust_res(eX, eY, eT, eZ, predicts_V, hii, vce, nnmatch, dups_V, dupsid_V, o+1, crv3=crv3, crv2=crv2, has_cluster=!is.null(C))
        aux = rdrobust_vce(dT+dZ, s, R_V*eW, res_V, eC, invG=invG_V_c, sqrtRX=sqrtRX_V, crv2=crv2)
        V_V = (invG_V%*%aux%*%invG_V)[nu+1,nu+1]
        v = crossprod(R_V*eW,((eX-c)/h_V)^(o+1))
        Hp = 0
        for (j in 1:(o+1)) Hp[j] = h_V^((j-1))
        BConst = (Hp*(invG_V%*%v))[nu+1]
        if (!is.null(vcache)) {
          assign(key, list(V_V=V_V, BConst=BConst, s=s),
                 envir = vcache, inherits = FALSE)
        }
  }
        
        w = rdrobust_kweight(X, c, h_B, kernel)
        if (!is.null(W)) w = W*w
        ind = w> 0 
        n_B = sum(ind)
        eY = Y[ind];eX = X[ind];eW = w[ind]
        D_B = eY
        R_B = .rdrobust_vander(as.numeric(eX - c), o_B)
        invG_B = qrXXinv(R_B*sqrt(eW))
        eT=eC=eZ=NULL
        if (!is.null(T)) {
          eT = T[ind]
          D_B = cbind(D_B,eT)
        }
        if (!is.null(Z)) {
          eZ = Z[ind,,drop=FALSE]
          D_B = cbind(D_B,eZ)
        }
        if (!is.null(C)) {
          eC=C[ind]
        }	
        beta_B = invG_B%*%crossprod(R_B*eW,D_B)	
        BWreg=0
        if (scale>0) {
        e_B = matrix(0,(o_B+1),1); e_B[o+2]=1
        dups_B=dupsid_B=hii=predicts_B=0
        if (vce=="nn") {
          dups_B   = dups[ind]
          dupsid_B = dupsid[ind]
        }
        if ((vce=="hc0" | vce=="hc1" | vce=="hc2" | vce=="hc3") & !crv3) {
          predicts_B = R_B%*%beta_B
          if (vce=="hc2" | vce=="hc3") {
            hii = rowSums((R_B%*%invG_B)*(R_B*eW))
          }
        } else if (crv3 | crv2) {
          predicts_B = R_B%*%beta_B
        }
        sqrtRX_B = if (crv3 | crv2) R_B * sqrt(eW) else NULL
        invG_B_c = if (crv3 | crv2) invG_B else NULL
        res_B = rdrobust_res(eX, eY, eT, eZ, predicts_B, hii, vce, nnmatch, dups_B, dupsid_B, o_B+1, crv3=crv3, crv2=crv2, has_cluster=!is.null(C))
        V_B = (invG_B%*%rdrobust_vce(dT+dZ, s, R_B*eW, res_B, eC, invG=invG_B_c, sqrtRX=sqrtRX_B, crv2=crv2)%*%invG_B)[o+2,o+2]
		BWreg = 3*BConst^2*V_B
	}
	B =  sqrt(2*(o+1-nu))*BConst%*%(t(s)%*%(beta_B[o+2,]))
	V = (2*nu+1)*h_V^(2*nu+1)*V_V
	R = scale*(2*(o+1-nu))*BWreg
	rate = 1/(2*o+3)
  output = list(V=V,B=B,R=R,rate=rate)
  return(output)
}

rdrobust_vce = function(d, s, RX, res, C, invG=NULL, sqrtRX=NULL, crv2=FALSE,
                        k_override=NULL, cluster_idx=NULL) {
  # k_override: when supplied, overrides the (n-1)/(n-k) df correction's k
  # for CR1. Used when RX is a "score-like" matrix (e.g. Q_q) whose ncol is
  # smaller than the effective number of regressors estimated. Without it,
  # CR1 at h!=b under cluster used k=p+1 from ncol(Q_q), inconsistent with
  # the q-regression direct path at h=b which uses k=q+1.
  # cluster_idx: optional precomputed output of .rdrobust_cluster_idx(C);
  # callers that invoke rdrobust_vce multiple times with the same C should
  # compute it once and pass it through (T3 optimization, 2026-05-12).
  # Note: k is the matrix dimension; k_df is what enters the df correction.
  k    = ncol(as.matrix(RX))
  k_df = if (is.null(k_override)) k else k_override
  M = matrix(0,k,k)
  crv3_mode = !is.null(invG) && !is.null(C) && !crv2
  crv2_mode = !is.null(invG) && !is.null(C) &&  crv2

  if (is.null(C)) {
    w = 1
    if (d==0){
      M  = crossprod(c(res)*RX)
    }
    else {
      # T4: Σᵢⱼ sᵢsⱼ rᵢrⱼ factors as (Σₗ sₗ rₗ)². One weighted
      # crossprod instead of an (d+1)² R loop. With r_comb = res %*% s,
      # crossprod(RX * r_comb) returns Σₙ r_comb[n]² · RX[n,]ᵀ RX[n,].
      r_comb <- as.vector(res %*% s)
      M = crossprod(c(r_comb) * RX)
    }
  }
  else if (crv3_mode) {
    # CRV3: cluster-level hat-matrix adjustment (Pustejovsky & Tipton 2018).
    #
    # By the Sherman-Morrison-Woodbury identity:
    #   A_g = (I_{n_g} - sqrtRX_g %*% invG %*% t(sqrtRX_g))^{-1}
    #       = I + sqrtRX_g %*% (G - L_g)^{-1} %*% t(sqrtRX_g)
    # where G = solve(invG) and L_g = crossprod(sqrtRX_g).
    #
    # The cluster score collapses to a k-vector without forming A_g:
    #   u_g     = t(sqrtRX_g) %*% sqrte_g          (k x 1)
    #   score_g = u_g + L_g %*% solve(G - L_g) %*% u_g
    #           = (I_k + L_g %*% M_g) %*% u_g       (k x 1)
    #
    # This is O(k^3) per cluster rather than O(n_g^3).
    #
    # For V_rb (sqrtRX=NULL, RX=Q_q used as approximate hat design):
    # same formula but without sqrt-weight scaling; fallback to M_g=0 (CR1)
    # when G - L_g is near-singular.
    if (is.null(cluster_idx)) cluster_idx = .rdrobust_cluster_idx(C)
    g = length(cluster_idx)
    w = 1  # CRV3 is approximately unbiased; no df correction needed
    has_sqrtRX = !is.null(sqrtRX)
    G = tryCatch(solve(invG), error=function(e) ginv(invG))  # k x k Gram matrix

    if (d==0) {
      for (i in seq_along(cluster_idx)) {
        ind   = cluster_idx[[i]]
        if (has_sqrtRX) {
          sRX_g = sqrtRX[ind,,drop=FALSE]
          sw_g  = sRX_g[,1,drop=FALSE]
          e_g   = res[ind,1,drop=FALSE]
          L_g   = crossprod(sRX_g)
          u_g   = crossprod(sRX_g, sw_g * e_g)
          M_g   = tryCatch(solve(G - L_g), error=function(e) ginv(G - L_g))
          score_g = u_g + L_g %*% (M_g %*% u_g)
        } else {
          RX_g  = RX[ind,,drop=FALSE]
          e_g   = res[ind,1,drop=FALSE]
          L_g   = crossprod(RX_g)
          M_g   = tryCatch(solve(G - L_g), error=function(e) matrix(0, k, k))
          u_g   = crossprod(RX_g, e_g)
          score_g = u_g + L_g %*% (M_g %*% u_g)
        }
        M = M + tcrossprod(score_g)
      }
    }
    else {
      for (i in seq_along(cluster_idx)) {
        ind = cluster_idx[[i]]
        ri  = res[ind,,drop=FALSE]
        sv  = numeric(k)
        if (has_sqrtRX) {
          sRX_g = sqrtRX[ind,,drop=FALSE]
          sw_g  = sRX_g[,1,drop=FALSE]
          L_g   = crossprod(sRX_g)
          M_g   = tryCatch(solve(G - L_g), error=function(e) ginv(G - L_g))
          for (l in 1:(1+d)) {
            u_g_l   = crossprod(sRX_g, sw_g * ri[,l])
            score_l = u_g_l + L_g %*% (M_g %*% u_g_l)
            sv      = sv + as.vector(score_l) * s[l]
          }
        } else {
          RX_g = RX[ind,,drop=FALSE]
          L_g  = crossprod(RX_g)
          M_g  = tryCatch(solve(G - L_g), error=function(e) matrix(0, k, k))
          for (l in 1:(1+d)) {
            u_g_l   = crossprod(RX_g, ri[,l])
            score_l = u_g_l + L_g %*% (M_g %*% u_g_l)
            sv      = sv + as.vector(score_l) * s[l]
          }
        }
        M = M + tcrossprod(matrix(sv, k, 1))
      }
    }
  }
  else if (crv2_mode) {
    # CRV2: cluster-level half-inverse hat-matrix adjustment (Bell & McCaffrey 2002).
    #
    # Uses A_g = (I_{n_g} - H_g)^{-1/2} where H_g = sqrtRX_g invG sqrtRX_g'.
    # Since H_g has rank <= k, we work entirely in k-dimensional space:
    #
    #   F_g'F_g = chol(invG) %*% L_g %*% t(chol(invG))   (k x k, symmetric)
    #
    # whose eigenvalues sigma2_j are the squared singular values of F_g
    # (= eigenvalues of H_g).  The per-eigenvalue CRV2 coefficient is:
    #
    #   c_j = [(1 - sigma2_j)^{-1/2} - 1] / sigma2_j
    #
    # and the cluster score collapses to:
    #
    #   score_g = u_g + L_g t(R) V diag(c) V' R u_g      (k x 1)
    #
    # where R = chol(invG), V = eigenvectors of F_g'F_g, u_g = X_g'W_g e_g.
    # This is O(k^3) per cluster - no n_g x k matrix needed.
    if (is.null(cluster_idx)) cluster_idx = .rdrobust_cluster_idx(C)
    g         = length(cluster_idx)
    w         = 1   # CRV2 is approximately unbiased; no df correction needed
    has_sqrtRX = !is.null(sqrtRX)
    C_half    = tryCatch(chol(invG), error=function(e) NULL)
    tC_half   = if (!is.null(C_half)) t(C_half) else NULL

    crv2_ok   = !is.null(C_half)

    if (d==0) {
      for (i in seq_along(cluster_idx)) {
        ind = cluster_idx[[i]]
        if (has_sqrtRX) {
          sRX_g = sqrtRX[ind,,drop=FALSE]
          sw_g  = sRX_g[,1,drop=FALSE]
          e_g   = res[ind,1,drop=FALSE]
          L_g   = crossprod(sRX_g)
          u_g   = crossprod(sRX_g, sw_g * e_g)
        } else {
          RX_g  = RX[ind,,drop=FALSE]
          e_g   = res[ind,1,drop=FALSE]
          L_g   = crossprod(RX_g)
          u_g   = crossprod(RX_g, e_g)
        }
        if (crv2_ok) {
          F_sq   = C_half %*% L_g %*% tC_half
          eig_g  = eigen(F_sq, symmetric=TRUE)
          sigma2 = pmax(eig_g$values, 0)
          V_g    = eig_g$vectors
          c_coef = ifelse(sigma2 < 1e-14, 0,
                          (1/sqrt(pmax(1-sigma2, 1e-8)) - 1) / sigma2)
          Ru_g    = C_half %*% u_g
          adj     = V_g %*% (c_coef * crossprod(V_g, Ru_g))
          score_g = u_g + L_g %*% (tC_half %*% adj)
        } else {
          score_g = u_g
        }
        M = M + tcrossprod(score_g)
      }
    }
    else {
      for (i in seq_along(cluster_idx)) {
        ind = cluster_idx[[i]]
        ri  = res[ind,,drop=FALSE]
        sv  = numeric(k)
        if (has_sqrtRX) {
          sRX_g = sqrtRX[ind,,drop=FALSE]
          sw_g  = sRX_g[,1,drop=FALSE]
          L_g   = crossprod(sRX_g)
        } else {
          RX_g  = RX[ind,,drop=FALSE]
          L_g   = crossprod(RX_g)
        }
        if (crv2_ok) {
          F_sq     = C_half %*% L_g %*% tC_half
          eig_g    = eigen(F_sq, symmetric=TRUE)
          sigma2   = pmax(eig_g$values, 0)
          V_g      = eig_g$vectors
          c_coef   = ifelse(sigma2 < 1e-14, 0,
                            (1/sqrt(pmax(1-sigma2, 1e-8)) - 1) / sigma2)
          LtC_half = L_g %*% tC_half
          for (l in 1:(1+d)) {
            u_g_l   = if (has_sqrtRX) crossprod(sRX_g, sw_g * ri[,l]) else crossprod(RX_g, ri[,l])
            Ru_g_l  = C_half %*% u_g_l
            adj_l   = V_g %*% (c_coef * crossprod(V_g, Ru_g_l))
            score_l = u_g_l + LtC_half %*% adj_l
            sv      = sv + as.vector(score_l) * s[l]
          }
        } else {
          for (l in 1:(1+d)) {
            u_g_l = if (has_sqrtRX) crossprod(sRX_g, sw_g * ri[,l]) else crossprod(RX_g, ri[,l])
            sv    = sv + as.vector(u_g_l) * s[l]
          }
        }
        M = M + tcrossprod(matrix(sv, k, 1))
      }
    }
  }
  else {
    # Standard cluster-robust (CR1) with small-sample df correction
    if (is.null(cluster_idx)) cluster_idx = .rdrobust_cluster_idx(C)
    g = length(cluster_idx)
    n = length(C)
    w = ((n-1)/(n-k_df))*(g/(g-1))
    if (d==0){
      for (i in seq_along(cluster_idx)) {
        ind = cluster_idx[[i]]
        Xi  = RX[ind,,drop=FALSE]
        ri  = res[ind,,drop=FALSE]
        sv  = as.vector(crossprod(Xi, ri))
        M   = M + tcrossprod(matrix(sv, k, 1))
      }
    }
    else {
      for (i in seq_along(cluster_idx)) {
        ind = cluster_idx[[i]]
        Xi  = RX[ind,,drop=FALSE]
        ri  = res[ind,,drop=FALSE]
        sv  = numeric(k)
        for (l in 1:(1+d)) sv = sv + as.vector(crossprod(Xi, s[l]*ri[,l]))
        M = M + tcrossprod(matrix(sv, k, 1))
      }
    }
  }
  return(w*M)
}

J.fun = function(B,V,n) {ceiling((((2*B)/V)*n)^(1/3))}


# Decoupled cluster-robust variance for the bias-corrected estimator when h != b.
# The sandwich uses Q_q (mixed-bandwidth score matrix) but the leverage comes
# from the q-regression (bandwidth b), whose cluster hat matrix has eigvals in
# [0, 1) — matrix analog of using hii_q for HC2/HC3.
#
# Returns the meat matrix M such that V_rb = invG_p · M · invG_p.
# Back-conversion from transformed (sqrt(W_b)) to original space leaves:
#   CRV3: e_adj_g = e_g + R_q_g · (G_q - L_g)^{-1} · u_q_g
#   CRV2: e_adj_g = e_g + R_q_g · Lambda_g · u_q_g
# where u_q_g = R_q_g' W_b e_g and L_g = R_q_g' W_b R_q_g.
.rdrobust_vce_qq_cluster <- function(Q, R_q, W_b, invG_q, res, C,
                                     crv2 = FALSE, s = 1, d = 0,
                                     cluster_idx = NULL) {
  Q      <- as.matrix(Q)
  R_q    <- as.matrix(R_q)
  res    <- as.matrix(res)
  k      <- ncol(Q)
  k_R    <- ncol(R_q)
  M      <- matrix(0, k, k)
  if (is.null(cluster_idx)) cluster_idx <- .rdrobust_cluster_idx(C)
  G_q    <- tryCatch(solve(invG_q), error = function(e) ginv(invG_q))

  for (g in seq_along(cluster_idx)) {
    idx <- cluster_idx[[g]]
    Q_g    <- Q[idx, , drop = FALSE]
    R_g    <- R_q[idx, , drop = FALSE]
    Wb_g   <- W_b[idx]
    sR_g   <- R_g * sqrt(Wb_g)                      # n_g × k_R, transformed design
    L_g    <- crossprod(sR_g)                       # k_R × k_R  = R_q_g' W_b R_q_g
    P_g    <- crossprod(Q_g, R_g)                   # k × k_R     = Q_q_g' R_q_g

    # T2: per-cluster Lambda (CRV2) or M_g (CRV3) depends only on
    # L_g / G_q / invG_q — not on the residual column `l`. Compute
    # once per cluster, reuse across the l-loop.
    Lambda <- NULL
    M_g    <- NULL
    use_simple <- FALSE
    if (crv2) {
      eig_L <- tryCatch(eigen(L_g, symmetric = TRUE),
                        error = function(e) NULL)
      if (is.null(eig_L)) {
        use_simple <- TRUE
      } else {
        sigma2 <- pmax(eig_L$values, 0)
        sigma  <- sqrt(sigma2)
        V_sv   <- eig_L$vectors
        keep   <- sigma > max(sigma) * 1e-10
        r      <- sum(keep)
        if (r == 0) {
          use_simple <- TRUE
        } else {
          V_r   <- V_sv[, keep, drop = FALSE]
          sig   <- sigma[keep]
          Dsig  <- diag(sig, nrow = r, ncol = r)
          T_mat <- Dsig %*% crossprod(V_r, invG_q %*% V_r) %*% Dsig
          T_mat <- 0.5 * (T_mat + t(T_mat))
          eigT  <- eigen(T_mat, symmetric = TRUE)
          tau   <- pmin(pmax(eigT$values, 0), 1 - 1e-10)
          W_g   <- eigT$vectors
          gamma <- 1/sqrt(1 - tau) - 1
          Dinv  <- diag(1/sig, nrow = r, ncol = r)
          A     <- V_r %*% Dinv %*% W_g
          Lambda <- A %*% (gamma * t(A))
        }
      }
    } else {
      GmL <- G_q - L_g
      M_g <- tryCatch(solve(GmL), error = function(e) matrix(0, k_R, k_R))
    }

    sv <- numeric(k)
    for (l in seq_len(1 + d)) {
      e_gl <- res[idx, l]
      u_Qq <- crossprod(Q_g, e_gl)               # k × 1
      if (use_simple) {
        score_l <- as.vector(u_Qq)
      } else {
        u_q     <- crossprod(R_g, Wb_g * e_gl)   # k_R × 1
        adj     <- if (crv2) Lambda %*% u_q else M_g %*% u_q
        score_l <- as.vector(u_Qq + P_g %*% adj)
      }
      sv <- sv + score_l * s[l]
    }
    M <- M + tcrossprod(matrix(sv, k, 1))
  }
  M
}





covs_drop_fun <- function(z) {
  z    <- as.matrix(z)
  qr_z <- qr(z, tol = 1e-7)
  keep <- sort(qr_z$pivot[seq_len(qr_z$rank)])
  list(covs = z[, keep, drop = FALSE], ncovs = qr_z$rank)
}

rdrobust = function(y, x, c = NULL, fuzzy = NULL, deriv = NULL,
                    p = NULL, q = NULL, h = NULL, b = NULL, rho = NULL,
                    covs = NULL, covs_drop = TRUE, ginv.tol = 1e-20,
                    kernel = "tri", weights = NULL, bwselect = "mserd",
                    vce = "nn", cluster = NULL, nnmatch = 3, level = 95,
                    scalepar = 1, scaleregul = 1, sharpbw = FALSE,
                    subset = NULL, masspoints = "adjust",
                    bwcheck = NULL, bwrestrict = TRUE, stdvars = FALSE,
                    data = NULL) {

  if (!is.null(data)) {
    .vars <- .rdrobust_resolve_data(match.call(), data, parent.frame(),
              c("y","x","covs","cluster","fuzzy","weights","subset"))
    y <- .vars$y; x <- .vars$x; covs <- .vars$covs
    cluster <- .vars$cluster; fuzzy <- .vars$fuzzy
    weights <- .vars$weights; subset <- .vars$subset
  }

  covs <- .rdrobust_resolve_covs(covs, data, parent.frame())

  # Validate auxiliary-vector lengths against length(x) BEFORE subset filtering,
  # so wrong-length inputs error explicitly instead of being silently recycled
  # by `[subset]` / `[na.ok]` indexing.
  n_orig <- length(x)
  .rdrobust_check_length(y,       "y",       n_orig)
  .rdrobust_check_length(weights, "weights", n_orig)
  .rdrobust_check_length(fuzzy,   "fuzzy",   n_orig)
  .rdrobust_check_length(covs,    "covs",    n_orig)
  .rdrobust_check_length(cluster, "cluster", n_orig)
  .rdrobust_check_subset(subset, n_orig)

  if (!is.null(subset)) {
    x <- x[subset]
    y <- y[subset]
    if (length(x) == 0L)
      stop("'subset' removed all observations.", call. = FALSE)
  }

  if (is.null(c)) c <- 0
  if (length(c) != 1 || !is.numeric(c) || !is.finite(c))
    stop(sprintf("Cutoff 'c' must be a single finite numeric value (received: %s).", toString(c)), call. = FALSE)
  if (is.null(p) & !is.null(deriv)) {p = deriv+1}
  
  if (length(p) == 0) {
    flag_no_p <- TRUE
    p <- 1
  } else if ((length(p) != 1) | !(p[1]%in%0:20)) {
    stop(sprintf("Polynomial order 'p' must be a single integer in 0:20 (received: %s).", toString(p)))
  } else {
    flag_no_p <- FALSE
  }

  if (length(q) == 0) {
    flag_no_q <- TRUE
    q <- p + 1
  } else if ((length(q) != 1) | !(q[1]%in%0:20) | (q[1]<=p)) {
    stop(sprintf("Polynomial order 'q' (for bias correction) must be a single integer in 0:20 with q > p (received q=%s, p=%s).", toString(q), toString(p)))
  } else {
    flag_no_q <- FALSE
  }

  if (length(deriv) == 0) {
    flag_no_deriv <- TRUE
    deriv <- 0
  } else if ((length(deriv) != 1) | !(deriv[1]%in%0:20) | (deriv[1]>p)) {
    stop(sprintf("Derivative order must be a single integer in 0:20 with deriv <= p (received deriv=%s, p=%s).", toString(deriv), toString(p)))
  } else {
    flag_no_deriv <- FALSE
  }
  
  na.ok <- complete.cases(x) & complete.cases(y)
  
  if (!is.null(cluster)){
    if (!is.null(subset))  cluster <- cluster[subset]
    if (length(cluster) != length(x))
      stop("cluster must have the same length as y and x.")
    na.ok <- na.ok & complete.cases(cluster)
  }
  
  if (!is.null(covs)){
    #if (!is.null(subset))  covs <- subset(covs,subset)
    if (!is.null(subset)) covs <- covs[subset, ,drop=FALSE]
    na.ok <- na.ok & complete.cases(covs)
    } 
  
  if (!is.null(fuzzy)){
    if (!is.null(subset)) fuzzy <- fuzzy[subset]
    na.ok <- na.ok & complete.cases(fuzzy)
  } 

  if (!is.null(weights)){
    if (!is.null(subset)) weights <- weights[subset]
    na.ok <- na.ok & complete.cases(weights) & weights>=0
  } 
  
  x = as.matrix(x[na.ok])
  y = as.matrix(y[na.ok])
  
  if (!is.null(covs))    covs    = as.matrix(covs)[na.ok, , drop = FALSE]
  if (!is.null(fuzzy))   fuzzy   = as.matrix(  fuzzy[na.ok])
  # Keep cluster as matrix (not factor): the cluster vector gets sliced into
  # left / right sides, and factor subsetting preserves unused levels, which
  # inflates `g = length(split(seq_along(C), C))` inside rdrobust_vce (empty
  # groups contribute 0 to M but change the CR1 df correction g/(g-1)).
  if (!is.null(cluster)) cluster = as.matrix(cluster[na.ok])
  if (!is.null(weights)) weights = as.matrix(weights[na.ok])
  
  if (is.null(masspoints)) masspoints <- FALSE

  if (vce=="nn" | masspoints=="check" |masspoints=="adjust") {
    order_x <- order(x)
    x <- x[order_x,,drop=FALSE]
    y <- y[order_x,,drop=FALSE]
    if (!is.null(covs))    covs    =  as.matrix(covs)[order_x,,drop=FALSE]
    if (!is.null(fuzzy))   fuzzy   =   fuzzy[order_x,,drop=FALSE]
    if (!is.null(cluster)) cluster = cluster[order_x,,drop=FALSE]
    if (!is.null(weights)) weights = weights[order_x,,drop=FALSE]
  }

  ############## COLLINEARITY
  covs_drop_coll=dZ=0
  if (isTRUE(covs_drop)) covs_drop_coll = 1
  if (!is.null(covs)) dZ = ncol(covs)
  
  if (!is.null(covs) & isTRUE(covs_drop)) {
    covs.names = colnames(covs)
    if (is.null(covs.names)) {
      covs.names = paste("z",1:ncol(covs),sep="")
      colnames(covs) = covs.names
    }
    covs = covs[,order(nchar(covs.names))]
    covs = as.matrix(covs)
    dZ = length(covs.names)
    covs.check = covs_drop_fun(covs)
    if (covs.check$ncovs < dZ) {
      covs  <- as.matrix(covs.check$covs)
      dZ    <- covs.check$ncovs
      warning("Multicollinearity issue detected in covs. Redundant covariates dropped.")  
    }
  }
  
  kernel   <- tolower(kernel)
  bwselect <- tolower(bwselect)
  vce      <- tolower(vce)
  
  x_sd = y_sd = 1
  if (is.null(h) & isTRUE(stdvars)) {
    y_sd = sd(y)
    x_sd = sd(x)
    y = y/y_sd
    x = x/x_sd
    c = c/x_sd
  }

  ind_l = x<c;  ind_r = x>=c
  X_l = x[ind_l,,drop=FALSE];  X_r = x[ind_r,,drop=FALSE]
  Y_l = y[ind_l,,drop=FALSE];  Y_r = y[ind_r,,drop=FALSE]
  x_min = min(x);  x_max = max(x)
  range_l = abs(c-x_min);  range_r = abs(c-x_max)
  N_l = length(X_l);   N_r = length(X_r)
  N = N_r + N_l
  quant = -qnorm(abs((1-(level/100))/2))
  
  dT = 0
  T_l = T_r = NULL
  perf_comp = FALSE
  if (!is.null(fuzzy)) {
    dT = 1
    T_l  = fuzzy[ind_l,,drop=FALSE];  T_r  = fuzzy[ind_r,,drop=FALSE]

    # Reject only the *fully* degenerate case: T constant across the whole
    # sample (zero variance on both sides AND no jump at the cutoff). The
    # one-sided cases (e.g. T=0 left and varying right, or T=0 left and T=1
    # right) are legitimate and handled by the perf_comp branch below.
    if (var(T_l) == 0 && var(T_r) == 0 && abs(mean(T_l) - mean(T_r)) < .Machine$double.eps^0.5) {
      stop("Fuzzy RD: first-stage variable has no variation and no jump at the cutoff. The fuzzy estimator is not identified.", call. = FALSE)
    }

    if (var(T_l)==0 | var(T_r)==0) perf_comp=TRUE

  if (isTRUE(perf_comp) | isTRUE(sharpbw)) {
      dT = 0
      T_l = T_r = NULL
    }
  }
  

  
  
  Z_l = Z_r = NULL
  if (!is.null(covs)) {
    Z_l  = covs[ind_l,,drop=FALSE];   Z_r  = covs[ind_r,,drop=FALSE]
  }
  
  g_l = g_r = 0       
  C_l = C_r = NULL
  if (!is.null(cluster)) {
    C_l = cluster[ind_l,,drop=FALSE]; g_l = length(unique(C_l))
    C_r = cluster[ind_r,,drop=FALSE]; g_r = length(unique(C_r))
  }
  
  fw_l = fw_r = 0
  if (!is.null(weights)) {
    fw_l = weights[ind_l,,drop=FALSE];  fw_r = weights[ind_r,,drop=FALSE]
  }

  if (vce=="nn") {
    runs_l   = rle(as.vector(X_l))
    dups_l   = rep(runs_l$lengths, runs_l$lengths)
    dupsid_l = sequence(runs_l$lengths)
    runs_r   = rle(as.vector(X_r))
    dups_r   = rep(runs_r$lengths, runs_r$lengths)
    dupsid_r = sequence(runs_r$lengths)
  }
  
  #####################################################   CHECK ERRORS
  exit=0
  valid_kernels  <- c("uni","uniform","tri","triangular","epa","epanechnikov","")
  valid_bwselect <- c("mserd","msetwo","msesum","msecomb1","msecomb2",
                      "cerrd","certwo","cersum","cercomb1","cercomb2","")
  valid_vce      <- c("nn","hc0","hc1","hc2","hc3","cr1","cr2","cr3","")

  if (!kernel %in% valid_kernels) {
    warning(sprintf("kernel incorrectly specified (received '%s'); allowed: tri, epa, uni.", kernel))
    exit = 1
  }

  if (!bwselect %in% valid_bwselect) {
    if (bwselect %in% c("cct","ik","cv")) {
      warning("bwselect options IK, CCT and CV have been deprecated. Please see help for new options.")
    } else if (bwselect %in% c("manual")) {
      warning("bwselect='manual' is internal-only; pass `h = ` directly to specify a manual bandwidth.")
    } else {
      warning(sprintf("bwselect incorrectly specified (received '%s').", bwselect))
    }
    exit = 1
  }

  if (!vce %in% valid_vce) {
    warning(sprintf("vce incorrectly specified (received '%s'); allowed: nn, hc0, hc1, hc2, hc3, cr1, cr2, cr3.", vce))
    exit = 1
  }
    
  if (c<=x_min | c>=x_max){
    warning("c should be set within the range of x")
    exit = 1
  }
    
  if (length(level) != 1 || !is.finite(level) || level >= 100 || level <= 0){
    warning("level should be a single number in (0, 100)")
    exit = 1
  }
    
  if (!is.null(rho)){
     if (length(rho) != 1 || !is.finite(rho) || rho<=0){
        warning("rho should be a single positive number")
        exit = 1
      }
  }

  if (!is.null(bwcheck)) {
    if (length(bwcheck) != 1 || !is.finite(bwcheck) || bwcheck < 1 || bwcheck != round(bwcheck)) {
      warning("bwcheck must be a single positive integer")
      exit = 1
    }
  }

  if (!is.null(masspoints) && !isFALSE(masspoints) &&
      !(is.character(masspoints) && length(masspoints) == 1L &&
        masspoints %in% c("check","adjust","off"))) {
    warning("masspoints must be one of 'check', 'adjust', 'off', or FALSE")
    exit = 1
  }

  if (!is.null(h)) {
    if (length(h) > 2 || any(!is.finite(h)) || any(h <= 0)) {
      warning("h must be a positive scalar or a length-2 positive vector")
      exit = 1
    }
  }

  if (!is.null(b)) {
    if (length(b) > 2 || any(!is.finite(b)) || any(b <= 0)) {
      warning("b must be a positive scalar or a length-2 positive vector")
      exit = 1
    }
  }

  if (!is.null(weights) && sum(weights, na.rm = TRUE) <= 0) {
    warning("weights must include at least one positive value")
    exit = 1
  }

  if (exit>0) stop("rdrobust: invalid input (see warnings above).")
  if (!is.null(h)) bwselect = "Manual"
  if (!is.null(h) & is.null(rho) & is.null(b)) {
    rho = 1
    b = h
  }
  if (!is.null(h) & !is.null(rho) ) b = h/rho
    
  
  if (N<20){
			warning("Not enough observations to perform bandwidth calculations. Estimates computed using entire sample")
      h = b = max(range_l,range_r)
			bwselect = "Manual"
		}
  
  if (kernel %in% c("epanechnikov", "epa")) {
    kernel_type = "Epanechnikov"
  } else if (kernel %in% c("uniform", "uni")) {
    kernel_type = "Uniform"
  } else {
    kernel_type = "Triangular"
  }
  
  # Cluster vce validation: with cluster, only cr1/cr2/cr3 are valid.
  # Without cluster, cr1/cr2/cr3 fall back to hc1/hc2/hc3.
  if (!is.null(cluster)) {
    if (vce %in% c("nn","")) {
      vce = "cr1"  # silent default
    } else if (vce %in% c("hc0","hc1")) {
      warning(paste0("vce='", vce, "' is not a cluster option. Switching to vce='cr1'."))
      vce = "cr1"
    } else if (vce == "hc2") {
      warning("vce='hc2' is not a cluster option. Switching to vce='cr2'.")
      vce = "cr2"
    } else if (vce == "hc3") {
      warning("vce='hc3' is not a cluster option. Switching to vce='cr3'.")
      vce = "cr3"
    } else if (!vce %in% c("cr1","cr2","cr3")) {
      warning(paste0("vce='", vce, "' is not a valid cluster option. Switching to vce='cr1'."))
      vce = "cr1"
    }
  } else {
    if (vce=="cr1") {
      warning("vce='cr1' requires a cluster variable. Falling back to vce='hc1'.")
      vce = "hc1"
    } else if (vce=="cr2") {
      warning("vce='cr2' requires a cluster variable. Falling back to vce='hc2'.")
      vce = "hc2"
    } else if (vce=="cr3") {
      warning("vce='cr3' requires a cluster variable. Falling back to vce='hc3'.")
      vce = "hc3"
    }
  }

  # Map user-facing cr1/cr2/cr3 to internal vce values
  vce_type = "NN"
  if (vce=="hc0")  vce_type = "HC0"
  if (vce=="hc1")  vce_type = "HC1"
  if (vce=="hc2")  vce_type = "HC2"
  if (vce=="hc3")  vce_type = "HC3"
  if (vce=="cr1") { vce_type = "CR1"; vce = "hc1" }
  if (vce=="cr2") { vce_type = "CR2"; vce = "crv2" }
  if (vce=="cr3") { vce_type = "CR3"; vce = "crv3" }

  crv3 = (vce=="crv3") & !is.null(cluster)
  crv2 = (vce=="crv2") & !is.null(cluster)

  mN = N;  M_l = N_l;  M_r = N_r

  # X_uniq_l/r and M_l/M_r are descriptive output fields and are also used
  # by the bandwidth-selection masspoints and bwcheck logic below.
  if (masspoints=="check" | masspoints=="adjust" | !is.null(bwcheck)) {
    X_uniq_l = sort(unique(X_l), decreasing=TRUE)
    X_uniq_r = unique(X_r)
    M_l = length(X_uniq_l)
    M_r = length(X_uniq_r)
    M = M_l + M_r
  }
  if (masspoints=="check" | masspoints=="adjust") {
    mass_l = 1-M_l/N_l
    mass_r = 1-M_r/N_r
    if (mass_l>=0.2 | mass_r>=0.2) {
      warning("Mass points detected in the running variable.")
      if (masspoints=="check") warning("Try using option masspoints=adjust.")
      if (is.null(bwcheck) & masspoints=="adjust") bwcheck <- 10
    }
  }

  if (is.null(h)) {
    x_iq = quantile(x,.75,type=2) - quantile(x,.25,type=2)
    BWp  = min(c(sd(x),x_iq/1.349))

    C_c = 2.576
    if (kernel %in% c("epanechnikov", "epa")) C_c = 2.34
    if (kernel %in% c("uniform",      "uni")) C_c = 1.843

    c_bw = C_c*BWp*N^(-1/5)
    if (masspoints=="adjust") c_bw = C_c*BWp*M^(-1/5)

    if (isTRUE(bwrestrict)) {
      bw_max_l = abs(c-x_min);  bw_max_r = abs(c-x_max)
      bw_max   = max(bw_max_l, bw_max_r)
      c_bw <- min(c_bw, bw_max)
    }

    bw.adj <- 0
    if (!is.null(bwcheck)) {
      bwcheck_l = min(bwcheck, M_l);  bwcheck_r = min(bwcheck, M_r)
      bw_min_l = abs(X_uniq_l-c)[bwcheck_l] + 1e-8
      bw_min_r = abs(X_uniq_r-c)[bwcheck_r] + 1e-8
      c_bw = max(c_bw, bw_min_l, bw_min_r)
      bw.adj <- 1
    }

    bw_fw_l = if (!is.null(weights)) fw_l else NULL
    bw_fw_r = if (!is.null(weights)) fw_r else NULL

    ### T1: per-side V-fit caches reused across all pilot calls.
    vcache_l <- new.env(parent = emptyenv())
    vcache_r <- new.env(parent = emptyenv())

    ### Step 1: d_bw
    C_d_l = rdrobust_bw(Y_l, X_l, T_l, Z_l, C_l, bw_fw_l, c=c, o=q+1, nu=q+1, o_B=q+2, h_V=c_bw, h_B=range_l, 0,         vce, nnmatch, kernel, dups_l, dupsid_l, covs_drop_coll, ginv.tol, vcache = vcache_l)
    C_d_r = rdrobust_bw(Y_r, X_r, T_r, Z_r, C_r, bw_fw_r, c=c, o=q+1, nu=q+1, o_B=q+2, h_V=c_bw, h_B=range_r, 0,         vce, nnmatch, kernel, dups_r, dupsid_r, covs_drop_coll, ginv.tol, vcache = vcache_r)

    ### TWO
    if (bwselect=="msetwo" | bwselect=="certwo" | bwselect=="msecomb2" | bwselect=="cercomb2") {
      d_bw_l = c((C_d_l$V / C_d_l$B^2)^C_d_l$rate)
      d_bw_r = c((C_d_r$V / C_d_r$B^2)^C_d_r$rate)
      if (isTRUE(bwrestrict)) { d_bw_l <- min(d_bw_l, bw_max_l); d_bw_r <- min(d_bw_r, bw_max_r) }
      if (!is.null(bwcheck))  { d_bw_l <- max(d_bw_l, bw_min_l); d_bw_r <- max(d_bw_r, bw_min_r) }
      C_b_l  = rdrobust_bw(Y_l, X_l, T_l, Z_l, C_l, bw_fw_l, c=c, o=q, nu=p+1, o_B=q+1, h_V=c_bw, h_B=d_bw_l, scaleregul, vce, nnmatch, kernel, dups_l, dupsid_l, covs_drop_coll, ginv.tol, vcache = vcache_l)
      b_bw_l = c((C_b_l$V / (C_b_l$B^2 + scaleregul*C_b_l$R))^C_b_l$rate)
      C_b_r  = rdrobust_bw(Y_r, X_r, T_r, Z_r, C_r, bw_fw_r, c=c, o=q, nu=p+1, o_B=q+1, h_V=c_bw, h_B=d_bw_r, scaleregul, vce, nnmatch, kernel, dups_r, dupsid_r, covs_drop_coll, ginv.tol, vcache = vcache_r)
      b_bw_r = c((C_b_r$V / (C_b_r$B^2 + scaleregul*C_b_r$R))^C_b_r$rate)
      if (isTRUE(bwrestrict)) { b_bw_l <- min(b_bw_l, bw_max_l); b_bw_r <- min(b_bw_r, bw_max_r) }
      C_h_l  = rdrobust_bw(Y_l, X_l, T_l, Z_l, C_l, bw_fw_l, c=c, o=p, nu=deriv, o_B=q, h_V=c_bw, h_B=b_bw_l, scaleregul, vce, nnmatch, kernel, dups_l, dupsid_l, covs_drop_coll, ginv.tol, vcache = vcache_l)
      h_bw_l = c((C_h_l$V / (C_h_l$B^2 + scaleregul*C_h_l$R))^C_h_l$rate)
      C_h_r  = rdrobust_bw(Y_r, X_r, T_r, Z_r, C_r, bw_fw_r, c=c, o=p, nu=deriv, o_B=q, h_V=c_bw, h_B=b_bw_r, scaleregul, vce, nnmatch, kernel, dups_r, dupsid_r, covs_drop_coll, ginv.tol, vcache = vcache_r)
      h_bw_r = c((C_h_r$V / (C_h_r$B^2 + scaleregul*C_h_r$R))^C_h_r$rate)
      if (isTRUE(bwrestrict)) { h_bw_l <- min(h_bw_l, bw_max_l); h_bw_r <- min(h_bw_r, bw_max_r) }
    }

    ### SUM
    if (bwselect=="msesum" | bwselect=="cersum" | bwselect=="msecomb1" | bwselect=="msecomb2" | bwselect=="cercomb1" | bwselect=="cercomb2") {
      d_bw_s = c(((C_d_l$V+C_d_r$V) / (C_d_r$B+C_d_l$B)^2)^C_d_l$rate)
      if (isTRUE(bwrestrict)) d_bw_s <- min(d_bw_s, bw_max)
      if (!is.null(bwcheck))  d_bw_s <- max(d_bw_s, bw_min_l, bw_min_r)
      C_b_l  = rdrobust_bw(Y_l, X_l, T_l, Z_l, C_l, bw_fw_l, c=c, o=q, nu=p+1, o_B=q+1, h_V=c_bw, h_B=d_bw_s, scaleregul, vce, nnmatch, kernel, dups_l, dupsid_l, covs_drop_coll, ginv.tol, vcache = vcache_l)
      C_b_r  = rdrobust_bw(Y_r, X_r, T_r, Z_r, C_r, bw_fw_r, c=c, o=q, nu=p+1, o_B=q+1, h_V=c_bw, h_B=d_bw_s, scaleregul, vce, nnmatch, kernel, dups_r, dupsid_r, covs_drop_coll, ginv.tol, vcache = vcache_r)
      b_bw_s = c(((C_b_l$V+C_b_r$V) / ((C_b_r$B+C_b_l$B)^2 + scaleregul*(C_b_r$R+C_b_l$R)))^C_b_l$rate)
      if (isTRUE(bwrestrict)) b_bw_s <- min(b_bw_s, bw_max)
      C_h_l  = rdrobust_bw(Y_l, X_l, T_l, Z_l, C_l, bw_fw_l, c=c, o=p, nu=deriv, o_B=q, h_V=c_bw, h_B=b_bw_s, scaleregul, vce, nnmatch, kernel, dups_l, dupsid_l, covs_drop_coll, ginv.tol, vcache = vcache_l)
      C_h_r  = rdrobust_bw(Y_r, X_r, T_r, Z_r, C_r, bw_fw_r, c=c, o=p, nu=deriv, o_B=q, h_V=c_bw, h_B=b_bw_s, scaleregul, vce, nnmatch, kernel, dups_r, dupsid_r, covs_drop_coll, ginv.tol, vcache = vcache_r)
      h_bw_s = c(((C_h_l$V+C_h_r$V) / ((C_h_r$B+C_h_l$B)^2 + scaleregul*(C_h_r$R+C_h_l$R)))^C_h_l$rate)
      if (isTRUE(bwrestrict)) h_bw_s <- min(h_bw_s, bw_max)
    }

    ### RD
    if (bwselect=="mserd" | bwselect=="cerrd" | bwselect=="msecomb1" | bwselect=="msecomb2" | bwselect=="cercomb1" | bwselect=="cercomb2" | bwselect=="") {
      d_bw_d = c(((C_d_l$V+C_d_r$V) / (C_d_r$B-C_d_l$B)^2)^C_d_l$rate)
      if (isTRUE(bwrestrict)) d_bw_d <- min(d_bw_d, bw_max)
      if (!is.null(bwcheck))  d_bw_d <- max(d_bw_d, bw_min_l, bw_min_r)
      C_b_l  = rdrobust_bw(Y_l, X_l, T_l, Z_l, C_l, bw_fw_l, c=c, o=q, nu=p+1, o_B=q+1, h_V=c_bw, h_B=d_bw_d, scaleregul, vce, nnmatch, kernel, dups_l, dupsid_l, covs_drop_coll, ginv.tol, vcache = vcache_l)
      C_b_r  = rdrobust_bw(Y_r, X_r, T_r, Z_r, C_r, bw_fw_r, c=c, o=q, nu=p+1, o_B=q+1, h_V=c_bw, h_B=d_bw_d, scaleregul, vce, nnmatch, kernel, dups_r, dupsid_r, covs_drop_coll, ginv.tol, vcache = vcache_r)
      b_bw_d = c(((C_b_l$V+C_b_r$V) / ((C_b_r$B-C_b_l$B)^2 + scaleregul*(C_b_r$R+C_b_l$R)))^C_b_l$rate)
      if (isTRUE(bwrestrict)) b_bw_d <- min(b_bw_d, bw_max)
      C_h_l  = rdrobust_bw(Y_l, X_l, T_l, Z_l, C_l, bw_fw_l, c=c, o=p, nu=deriv, o_B=q, h_V=c_bw, h_B=b_bw_d, scaleregul, vce, nnmatch, kernel, dups_l, dupsid_l, covs_drop_coll, ginv.tol, vcache = vcache_l)
      C_h_r  = rdrobust_bw(Y_r, X_r, T_r, Z_r, C_r, bw_fw_r, c=c, o=p, nu=deriv, o_B=q, h_V=c_bw, h_B=b_bw_d, scaleregul, vce, nnmatch, kernel, dups_r, dupsid_r, covs_drop_coll, ginv.tol, vcache = vcache_r)
      h_bw_d = c(((C_h_l$V+C_h_r$V) / ((C_h_r$B-C_h_l$B)^2 + scaleregul*(C_h_r$R+C_h_l$R)))^C_h_l$rate)
      if (isTRUE(bwrestrict)) h_bw_d <- min(h_bw_d, bw_max)
    }

    if (bwselect=="mserd" | bwselect=="cerrd" | bwselect=="msecomb1" | bwselect=="msecomb2" | bwselect=="cercomb1" | bwselect=="cercomb2" | bwselect=="") { h_mserd = x_sd*h_bw_d; b_mserd = x_sd*b_bw_d }
    if (bwselect=="msesum" | bwselect=="cersum" | bwselect=="msecomb1" | bwselect=="msecomb2" | bwselect=="cercomb1" | bwselect=="cercomb2")                { h_msesum = x_sd*h_bw_s; b_msesum = x_sd*b_bw_s }
    if (bwselect=="msetwo" | bwselect=="certwo" | bwselect=="msecomb2" | bwselect=="cercomb2") { h_msetwo_l=x_sd*h_bw_l; h_msetwo_r=x_sd*h_bw_r; b_msetwo_l=x_sd*b_bw_l; b_msetwo_r=x_sd*b_bw_r }
    if (bwselect=="msecomb1" | bwselect=="cercomb1") { h_msecomb1=min(c(h_mserd,h_msesum)); b_msecomb1=min(c(b_mserd,b_msesum)) }
    if (bwselect=="msecomb2" | bwselect=="cercomb2") {
      h_msecomb2_l=median(c(h_mserd,h_msesum,h_msetwo_l)); h_msecomb2_r=median(c(h_mserd,h_msesum,h_msetwo_r))
      b_msecomb2_l=median(c(b_mserd,b_msesum,b_msetwo_l)); b_msecomb2_r=median(c(b_mserd,b_msesum,b_msetwo_r))
    }

    cer_h = N^(-(p/((3+p)*(3+2*p))))
    if (!is.null(cluster)) cer_h = (g_l+g_r)^(-(p/((3+p)*(3+2*p))))
    cer_b = 1
    if (bwselect=="cerrd")    { h_cerrd=h_mserd*cer_h;       b_cerrd=b_mserd*cer_b }
    if (bwselect=="cersum")   { h_cersum=h_msesum*cer_h;     b_cersum=b_msesum*cer_b }
    if (bwselect=="certwo")   { h_certwo_l=h_msetwo_l*cer_h; h_certwo_r=h_msetwo_r*cer_h; b_certwo_l=b_msetwo_l*cer_b; b_certwo_r=b_msetwo_r*cer_b }
    if (bwselect=="cercomb1") { h_cercomb1=h_msecomb1*cer_h; b_cercomb1=b_msecomb1*cer_b }
    if (bwselect=="cercomb2") { h_cercomb2_l=h_msecomb2_l*cer_h; h_cercomb2_r=h_msecomb2_r*cer_h; b_cercomb2_l=b_msecomb2_l*cer_b; b_cercomb2_r=b_msecomb2_r*cer_b }

    bws = matrix(NA,1,4)
    colnames(bws) = c("h (left)","h (right)","b (left)","b (right)")
    rownames(bws) = bwselect
    if (bwselect=="mserd"   | bwselect=="") bws[1,] = c(h_mserd,      h_mserd,      b_mserd,      b_mserd)
    if (bwselect=="msetwo")                 bws[1,] = c(h_msetwo_l,   h_msetwo_r,   b_msetwo_l,   b_msetwo_r)
    if (bwselect=="msesum")                 bws[1,] = c(h_msesum,     h_msesum,     b_msesum,     b_msesum)
    if (bwselect=="msecomb1")               bws[1,] = c(h_msecomb1,   h_msecomb1,   b_msecomb1,   b_msecomb1)
    if (bwselect=="msecomb2")               bws[1,] = c(h_msecomb2_l, h_msecomb2_r, b_msecomb2_l, b_msecomb2_r)
    if (bwselect=="cerrd")                  bws[1,] = c(h_cerrd,      h_cerrd,      b_cerrd,      b_cerrd)
    if (bwselect=="certwo")                 bws[1,] = c(h_certwo_l,   h_certwo_r,   b_certwo_l,   b_certwo_r)
    if (bwselect=="cersum")                 bws[1,] = c(h_cersum,     h_cersum,     b_cersum,     b_cersum)
    if (bwselect=="cercomb1")               bws[1,] = c(h_cercomb1,   h_cercomb1,   b_cercomb1,   b_cercomb1)
    if (bwselect=="cercomb2")               bws[1,] = c(h_cercomb2_l, h_cercomb2_r, b_cercomb2_l, b_cercomb2_r)

    h_l = bws[1,1]; h_r = bws[1,2]
    b_l = bws[1,3]; b_r = bws[1,4]
    if (!is.null(rho)) {
      b_l = h_l/rho
      b_r = h_r/rho
    }
    } else{
      if (length(h)==1) h_l = h_r = h
      if (length(h)==2) {
        h_l = h[1]
        h_r = h[2]
      }
      if (is.null(b)) {
        b_l = h_l
        b_r = h_r
      } else {
        if (length(b)==1) b_l = b_r = b
        if (length(b)==2) {
          b_l = b[1]
          b_r = b[2]
        }  
      }  
    }

  if (isTRUE(stdvars)) { 
    c = c*x_sd
  	X_l = X_l*x_sd;	X_r = X_r*x_sd
  	Y_l = Y_l*y_sd;	Y_r = Y_r*y_sd
  }
  
  ### end BW selection
  
  
  ### Estimation
  w_h_l <- rdrobust_kweight(X_l,c,h_l,kernel);	w_h_r <- rdrobust_kweight(X_r,c,h_r,kernel)
  w_b_l <- rdrobust_kweight(X_l,c,b_l,kernel);	w_b_r <- rdrobust_kweight(X_r,c,b_r,kernel)
  
  if (!is.null(weights)) {
    w_h_l <- fw_l*w_h_l;	w_h_r <- fw_r*w_h_r
    w_b_l <- fw_l*w_b_l;	w_b_r <- fw_r*w_b_r			
  }

  ind_h_l <- w_h_l> 0;		ind_h_r <- w_h_r> 0
  ind_b_l <- w_b_l> 0;		ind_b_r <- w_b_r> 0
  N_h_l <- sum(ind_h_l); N_b_l <- sum(ind_b_l)
  N_h_r <- sum(ind_h_r); N_b_r <- sum(ind_b_r)
  
  ind_l = ind_b_l; ind_r = ind_b_r
  if (h_l>b_l) ind_l = ind_h_l   
  if (h_r>b_r) ind_r = ind_h_r   
  
  eN_l = sum(ind_l); eN_r = sum(ind_r)
  eY_l  = Y_l[ind_l,,drop=FALSE];	eY_r  = Y_r[ind_r,,drop=FALSE]
  eX_l  = X_l[ind_l,,drop=FALSE];	eX_r  = X_r[ind_r,,drop=FALSE]
  W_h_l = w_h_l[ind_l];	W_h_r = w_h_r[ind_r]
  W_b_l = w_b_l[ind_l];	W_b_r = w_b_r[ind_r]
  
  edups_l = edupsid_l = edups_r = edupsid_r = 0
 
  if (vce=="nn") {
    edups_l   = dups_l[ind_l] 
    edups_r   = dups_r[ind_r]
    edupsid_l = dupsid_l[ind_l]
    edupsid_r = dupsid_r[ind_r]
  }          
          
  u_l <- (eX_l-c)/h_l;	u_r <-(eX_r-c)/h_r
  R_q_l = .rdrobust_vander(as.numeric(eX_l - c), q)
  R_q_r = .rdrobust_vander(as.numeric(eX_r - c), q)
  R_p_l = R_q_l[, 1:(p+1), drop = FALSE]
  R_p_r = R_q_r[, 1:(p+1), drop = FALSE]

  
  L_l = crossprod(R_p_l*W_h_l,u_l^(p+1)); L_r = crossprod(R_p_r*W_h_r,u_r^(p+1)) 
  invG_q_l  = qrXXinv((sqrt(W_b_l)*R_q_l));	invG_q_r  = qrXXinv((sqrt(W_b_r)*R_q_r))
  invG_p_l  = qrXXinv((sqrt(W_h_l)*R_p_l));	invG_p_r  = qrXXinv((sqrt(W_h_r)*R_p_r))
  e_p1 = matrix(0,(q+1),1); e_p1[p+2]=1
  e_v  = matrix(0,(p+1),1); e_v[deriv+1]=1
  Q_q_l = t(t(R_p_l*W_h_l) - h_l^(p+1)*(L_l%*%t(e_p1))%*%t(t(invG_q_l%*%t(R_q_l))*W_b_l))
  Q_q_r = t(t(R_p_r*W_h_r) - h_r^(p+1)*(L_r%*%t(e_p1))%*%t(t(invG_q_r%*%t(R_q_r))*W_b_r))
  D_l = eY_l; D_r = eY_r

  eT_l = eT_r = NULL
  if (!is.null(fuzzy)) {
  
    if (isTRUE(perf_comp) | isTRUE(sharpbw)) {
      dT = 1
      T_l  = fuzzy[x<c,,drop=FALSE];  T_r  = fuzzy[x>=c,,drop=FALSE]
    }
    
    eT_l = T_l[ind_l,,drop=FALSE]; D_l  = cbind(D_l,eT_l)
    eT_r = T_r[ind_r,,drop=FALSE]; D_r  = cbind(D_r,eT_r)
  }
  
  eZ_l = eZ_r = NULL
  if (!is.null(covs)) {
    eZ_l  = Z_l[ind_l,,drop=FALSE]; D_l   = cbind(D_l,eZ_l)
    eZ_r  = Z_r[ind_r,,drop=FALSE]; D_r   = cbind(D_r,eZ_r)
    U_p_l = crossprod(R_p_l*W_h_l,D_l); U_p_r = crossprod(R_p_r*W_h_r,D_r)
  }
    
  eC_l = eC_r = NULL
  cidx_l = cidx_r = NULL
  if (!is.null(cluster)) {
    eC_l  = C_l[ind_l]; eC_r  = C_r[ind_r]
    # T3: precompute per-side cluster index once for all V_Y/V_T vce calls.
    cidx_l = .rdrobust_cluster_idx(eC_l)
    cidx_r = .rdrobust_cluster_idx(eC_r)
  }
  
  beta_p_l  = invG_p_l%*%crossprod(R_p_l*W_h_l,D_l) 
  beta_p_r  = invG_p_r%*%crossprod(R_p_r*W_h_r,D_r) 
  beta_q_l  = invG_q_l%*%crossprod(R_q_l*W_b_l,D_l)
  beta_q_r  = invG_q_r%*%crossprod(R_q_r*W_b_r,D_r)
  beta_bc_l = invG_p_l%*%crossprod(Q_q_l,D_l) 
  beta_bc_r = invG_p_r%*%crossprod(Q_q_r,D_r)
  beta_p    = beta_p_r  - beta_p_l
  beta_q    = beta_q_r  - beta_q_l
  beta_bc   = beta_bc_r - beta_bc_l

  gamma_p = NULL
  if (is.null(covs)) {	
    tau_cl = tau_Y_cl = scalepar*factorial(deriv)*beta_p[(deriv+1),1]
    tau_bc = tau_Y_bc = scalepar*factorial(deriv)*beta_bc[(deriv+1),1]
    s_Y = 1
    
    tau_Y_cl_l = scalepar*factorial(deriv)*beta_p_l[(deriv+1),1]
    tau_Y_cl_r = scalepar*factorial(deriv)*beta_p_r[(deriv+1),1]
    tau_Y_bc_l = scalepar*factorial(deriv)*beta_bc_l[(deriv+1),1]
    tau_Y_bc_r = scalepar*factorial(deriv)*beta_bc_r[(deriv+1),1]
    bias_l = tau_Y_cl_l-tau_Y_bc_l
    bias_r = tau_Y_cl_r-tau_Y_bc_r 
    
    beta_Y_p_l = scalepar*factorial(deriv)*beta_p_l[,1]
    beta_Y_p_r = scalepar*factorial(deriv)*beta_p_r[,1]
    
    
    if (!is.null(fuzzy)) {
       tau_T_cl = factorial(deriv)*beta_p[(deriv+1),2]
       tau_T_bc = factorial(deriv)*beta_bc[(deriv+1),2]
       tau_cl   = tau_Y_cl/tau_T_cl
       s_Y      = c(1/tau_T_cl , -(tau_Y_cl/tau_T_cl^2))
       B_F      = c(tau_Y_cl-tau_Y_bc , tau_T_cl-tau_T_bc)
       tau_bc   = tau_cl - t(s_Y)%*%B_F
       sV_T     = c(0 , 1)
       
       tau_T_cl_l = factorial(deriv)*beta_p_l[(deriv+1),2]
       tau_T_cl_r = factorial(deriv)*beta_p_r[(deriv+1),2]
       tau_T_bc_l = factorial(deriv)*beta_bc_l[(deriv+1),2]
       tau_T_bc_r = factorial(deriv)*beta_bc_r[(deriv+1),2]
       B_F_l = c(tau_Y_cl_l-tau_Y_bc_l, tau_T_cl_l-tau_T_bc_l)
       B_F_r = c(tau_Y_cl_r-tau_Y_bc_r, tau_T_cl_r-tau_T_bc_r)
       bias_l = t(s_Y)%*%B_F_l
  		 bias_r = t(s_Y)%*%B_F_r
  		 
  		 beta_T_p_l = scalepar*factorial(deriv)*beta_p_l[,2]
  		 beta_T_p_r = scalepar*factorial(deriv)*beta_p_r[,2]
  		 
  }	
  } else {	
    ZWD_p_l  = crossprod(eZ_l*W_h_l,D_l)
    ZWD_p_r  = crossprod(eZ_r*W_h_r,D_r)
    colsZ = (2+dT):max(c(2+dT+dZ-1,(2+dT)))
    UiGU_p_l =  crossprod(matrix(U_p_l[,colsZ],nrow=p+1),invG_p_l%*%U_p_l) 
    UiGU_p_r =  crossprod(matrix(U_p_r[,colsZ],nrow=p+1),invG_p_r%*%U_p_r) 
    ZWZ_p_l = ZWD_p_l[,colsZ] - UiGU_p_l[,colsZ] 
    ZWZ_p_r = ZWD_p_r[,colsZ] - UiGU_p_r[,colsZ]     
    ZWY_p_l = ZWD_p_l[,1:(1+dT)] - UiGU_p_l[,1:(1+dT)] 
    ZWY_p_r = ZWD_p_r[,1:(1+dT)] - UiGU_p_r[,1:(1+dT)]     
    ZWZ_p = ZWZ_p_r + ZWZ_p_l
    ZWY_p = ZWY_p_r + ZWY_p_l
    if (covs_drop_coll == 0) gamma_p = chol2inv(chol(ZWZ_p))%*%ZWY_p
    if (covs_drop_coll == 1) gamma_p = ginv(ZWZ_p, tol = ginv.tol)%*%ZWY_p
    s_Y = c(1 ,  -gamma_p[,1])
    
    if (is.null(fuzzy)) {
        tau_cl = scalepar*t(s_Y)%*%beta_p[(deriv+1),]
        tau_bc = scalepar*t(s_Y)%*%beta_bc[(deriv+1),]
        
        tau_Y_cl_l = scalepar*t(s_Y)%*%beta_p_l[(deriv+1),]
        tau_Y_cl_r = scalepar*t(s_Y)%*%beta_p_r[(deriv+1),]
        tau_Y_bc_l = scalepar*t(s_Y)%*%beta_bc_l[(deriv+1),]
        tau_Y_bc_r = scalepar*t(s_Y)%*%beta_bc_r[(deriv+1),]
        bias_l = tau_Y_cl_l-tau_Y_bc_l
        bias_r = tau_Y_cl_r-tau_Y_bc_r 
        
        beta_Y_p_l = scalepar*tcrossprod(s_Y,beta_p_l)
        beta_Y_p_r = scalepar*tcrossprod(s_Y,beta_p_r)

    } else {
      s_T  = c(1,    -gamma_p[,2])
      sV_T = c(0, 1, -gamma_p[,2])
      tau_Y_cl = c(scalepar*factorial(deriv)*t(s_Y)%*%c(beta_p[(deriv+1),1], beta_p[(deriv+1),colsZ]))
      tau_Y_bc = c(scalepar*factorial(deriv)*t(s_Y)%*%c(beta_bc[(deriv+1),1],beta_bc[(deriv+1),colsZ]))
      tau_T_cl = c(         factorial(deriv)*t(s_T)%*%c(beta_p[(deriv+1),2], beta_p[(deriv+1),colsZ]))
      tau_T_bc = c(         factorial(deriv)*t(s_T)%*%c(beta_bc[(deriv+1),2],beta_bc[(deriv+1),colsZ]))

      tau_Y_cl_l = c(scalepar*factorial(deriv)*t(s_Y)%*%c(beta_p_l[(deriv+1),1], beta_p_l[(deriv+1),colsZ]))
      tau_Y_cl_r = c(scalepar*factorial(deriv)*t(s_Y)%*%c(beta_p_r[(deriv+1),1], beta_p_r[(deriv+1),colsZ]))
      tau_Y_bc_l = c(scalepar*factorial(deriv)*t(s_Y)%*%c(beta_bc_l[(deriv+1),1],beta_bc_l[(deriv+1),colsZ]))
      tau_Y_bc_r = c(scalepar*factorial(deriv)*t(s_Y)%*%c(beta_bc_r[(deriv+1),1],beta_bc_r[(deriv+1),colsZ]))

      tau_T_cl_l = c(factorial(deriv)*t(s_T)%*%c(beta_p_l[(deriv+1),2], beta_p_l[(deriv+1), colsZ]))
      tau_T_cl_r = c(factorial(deriv)*t(s_T)%*%c(beta_p_r[(deriv+1),2], beta_p_r[(deriv+1), colsZ]))
      tau_T_bc_l = c(factorial(deriv)*t(s_T)%*%c(beta_bc_l[(deriv+1),2],beta_bc_l[(deriv+1),colsZ]))
      tau_T_bc_r = c(factorial(deriv)*t(s_T)%*%c(beta_bc_r[(deriv+1),2],beta_bc_r[(deriv+1),colsZ]))

      beta_Y_p_l = scalepar*factorial(deriv)*t(s_Y)%*%t(cbind(beta_p_l[,1], beta_p_l[,colsZ]))
      beta_Y_p_r = scalepar*factorial(deriv)*t(s_Y)%*%t(cbind(beta_p_r[,1], beta_p_r[,colsZ]))
      beta_T_p_l =          factorial(deriv)*t(s_T)%*%t(cbind(beta_p_l[,2], beta_p_l[,colsZ]))
      beta_T_p_r =          factorial(deriv)*t(s_T)%*%t(cbind(beta_p_r[,2], beta_p_r[,colsZ]))
      
      tau_cl = tau_Y_cl/tau_T_cl
      B_F   = c(tau_Y_cl-tau_Y_bc,     tau_T_cl-tau_T_bc)
      B_F_l = c(tau_Y_cl_l-tau_Y_bc_l, tau_T_cl_l-tau_T_bc_l)
      B_F_r = c(tau_Y_cl_r-tau_Y_bc_r, tau_T_cl_r-tau_T_bc_r)
      
      s_Y = c(1/tau_T_cl , -(tau_Y_cl/tau_T_cl^2))
      tau_bc = tau_cl - t(s_Y)%*%B_F
      
      bias_l = t(s_Y)%*%B_F_l
      bias_r = t(s_Y)%*%B_F_r

      s_Y = c(1/tau_T_cl , -(tau_Y_cl/tau_T_cl^2) , -(1/tau_T_cl)*gamma_p[,1] + (tau_Y_cl/tau_T_cl^2)*gamma_p[,2])
    }
  }



  hii_p_l = hii_p_r = hii_q_l = hii_q_r = predicts_p_l = predicts_p_r = predicts_q_l = predicts_q_r = 0
  if ((vce=="hc0" | vce=="hc1" | vce=="hc2" | vce=="hc3") & !crv3) {
    predicts_p_l = R_p_l%*%beta_p_l
    predicts_p_r = R_p_r%*%beta_p_r
    predicts_q_l = R_q_l%*%beta_q_l
    predicts_q_r = R_q_r%*%beta_q_r

    if (vce=="hc2" | vce=="hc3") {
      hii_p_l = rowSums((R_p_l%*%invG_p_l)*(R_p_l*W_h_l))
      hii_p_r = rowSums((R_p_r%*%invG_p_r)*(R_p_r*W_h_r))

      hii_q_l = rowSums((R_q_l%*%invG_q_l)*(R_q_l*W_b_l))
      hii_q_r = rowSums((R_q_r%*%invG_q_r)*(R_q_r*W_b_r))
    }
  } else if (crv3 | crv2) {
    # CRV2/CRV3: compute predictions for raw residuals; leverage adjustment done cluster-wise
    predicts_p_l = R_p_l%*%beta_p_l
    predicts_p_r = R_p_r%*%beta_p_r
    predicts_q_l = R_q_l%*%beta_q_l
    predicts_q_r = R_q_r%*%beta_q_r
  }

	res_h_l = rdrobust_res(eX_l, eY_l, eT_l, eZ_l, predicts_p_l, hii_p_l, vce, nnmatch, edups_l, edupsid_l, p+1, crv3=crv3, crv2=crv2, has_cluster=!is.null(cluster))
	res_h_r = rdrobust_res(eX_r, eY_r, eT_r, eZ_r, predicts_p_r, hii_p_r, vce, nnmatch, edups_r, edupsid_r, p+1, crv3=crv3, crv2=crv2, has_cluster=!is.null(cluster))

		if (vce=="nn") {
			res_b_l = res_h_l;	res_b_r = res_h_r
	} 	else {
			res_b_l = rdrobust_res(eX_l, eY_l, eT_l, eZ_l, predicts_q_l, hii_q_l, vce, nnmatch, edups_l, edupsid_l, q+1, crv3=crv3, crv2=crv2, has_cluster=!is.null(cluster))
			res_b_r = rdrobust_res(eX_r, eY_r, eT_r, eZ_r, predicts_q_r, hii_q_r, vce, nnmatch, edups_r, edupsid_r, q+1, crv3=crv3, crv2=crv2, has_cluster=!is.null(cluster))
  }

  # CRV2/CRV3: sqrt-kernel-weighted design matrices for symmetric hat-matrix blocks
  sqrtRX_p_l = sqrtRX_p_r = NULL
  if (crv3 | crv2) {
    sqrtRX_p_l = R_p_l * sqrt(W_h_l)
    sqrtRX_p_r = R_p_r * sqrt(W_h_r)
  }
  crv_invG_l = if (crv3 | crv2) invG_p_l else NULL
  crv_invG_r = if (crv3 | crv2) invG_p_r else NULL

	V_Y_cl_l = invG_p_l%*%rdrobust_vce(dT+dZ, s_Y, as.matrix(R_p_l*W_h_l), res_h_l, eC_l, cluster_idx=cidx_l, invG=crv_invG_l, sqrtRX=sqrtRX_p_l, crv2=crv2)%*%invG_p_l
	V_Y_cl_r = invG_p_r%*%rdrobust_vce(dT+dZ, s_Y, as.matrix(R_p_r*W_h_r), res_h_r, eC_r, cluster_idx=cidx_r, invG=crv_invG_r, sqrtRX=sqrtRX_p_r, crv2=crv2)%*%invG_p_r

  # V_rb: when h == b, use q-regression formulation directly (clean CRV2/CRV3).
  # When h != b with CRV2/CRV3, use Q_q sandwich + q-regression cluster leverage
  # (matrix analog of hii_q for HC2/HC3). Otherwise CR1.
  hb_match = (h_l == b_l) & (h_r == b_r)
  if (hb_match & (crv3 | crv2)) {
    sqrtRX_q_l = R_q_l * sqrt(W_h_l)
    sqrtRX_q_r = R_q_r * sqrt(W_h_r)
    V_Y_rb_l = invG_q_l%*%rdrobust_vce(dT+dZ, s_Y, as.matrix(R_q_l*W_h_l), res_b_l, eC_l, cluster_idx=cidx_l, invG=invG_q_l, sqrtRX=sqrtRX_q_l, crv2=crv2)%*%invG_q_l
    V_Y_rb_r = invG_q_r%*%rdrobust_vce(dT+dZ, s_Y, as.matrix(R_q_r*W_h_r), res_b_r, eC_r, cluster_idx=cidx_r, invG=invG_q_r, sqrtRX=sqrtRX_q_r, crv2=crv2)%*%invG_q_r
  } else if (hb_match & !is.null(cluster)) {
    V_Y_rb_l = invG_q_l%*%rdrobust_vce(dT+dZ, s_Y, as.matrix(R_q_l*W_h_l), res_b_l, eC_l, cluster_idx=cidx_l, invG=NULL, sqrtRX=NULL, crv2=FALSE)%*%invG_q_l
    V_Y_rb_r = invG_q_r%*%rdrobust_vce(dT+dZ, s_Y, as.matrix(R_q_r*W_h_r), res_b_r, eC_r, cluster_idx=cidx_r, invG=NULL, sqrtRX=NULL, crv2=FALSE)%*%invG_q_r
  } else if (crv3 | crv2) {
    M_rb_l = .rdrobust_vce_qq_cluster(Q_q_l, R_q_l, W_b_l, invG_q_l, res_b_l, eC_l, cluster_idx=cidx_l, crv2=crv2, s=s_Y, d=dT+dZ)
    M_rb_r = .rdrobust_vce_qq_cluster(Q_q_r, R_q_r, W_b_r, invG_q_r, res_b_r, eC_r, cluster_idx=cidx_r, crv2=crv2, s=s_Y, d=dT+dZ)
    V_Y_rb_l = invG_p_l %*% M_rb_l %*% invG_p_l
    V_Y_rb_r = invG_p_r %*% M_rb_r %*% invG_p_r
  } else {
    # k_override=q+1 aligns the CR1 df correction with the q-regression
    # path used at h=b, so SE is continuous across the h=b boundary.
    # Non-cluster paths ignore k (df adjustment is in rdrobust_res's w factor).
    V_Y_rb_l = invG_p_l%*%rdrobust_vce(dT+dZ, s_Y, as.matrix(Q_q_l),       res_b_l, eC_l, cluster_idx=cidx_l, invG=NULL, sqrtRX=NULL, crv2=FALSE, k_override=q+1)%*%invG_p_l
    V_Y_rb_r = invG_p_r%*%rdrobust_vce(dT+dZ, s_Y, as.matrix(Q_q_r),       res_b_r, eC_r, cluster_idx=cidx_r, invG=NULL, sqrtRX=NULL, crv2=FALSE, k_override=q+1)%*%invG_p_r
  }
	V_tau_cl = scalepar^2*factorial(deriv)^2*(V_Y_cl_l+V_Y_cl_r)[deriv+1,deriv+1]
	V_tau_rb = scalepar^2*factorial(deriv)^2*(V_Y_rb_l+V_Y_rb_r)[deriv+1,deriv+1]
	se_tau_cl = sqrt(V_tau_cl);	se_tau_rb = sqrt(V_tau_rb)

	if (!is.null(fuzzy)) {
		V_T_cl_l = invG_p_l%*%rdrobust_vce(dT+dZ, sV_T, as.matrix(R_p_l*W_h_l), res_h_l, eC_l, cluster_idx=cidx_l, invG=crv_invG_l, sqrtRX=sqrtRX_p_l, crv2=crv2)%*%invG_p_l
		V_T_cl_r = invG_p_r%*%rdrobust_vce(dT+dZ, sV_T, as.matrix(R_p_r*W_h_r), res_h_r, eC_r, cluster_idx=cidx_r, invG=crv_invG_r, sqrtRX=sqrtRX_p_r, crv2=crv2)%*%invG_p_r
    if (hb_match & (crv3 | crv2)) {
      V_T_rb_l = invG_q_l%*%rdrobust_vce(dT+dZ, sV_T, as.matrix(R_q_l*W_h_l), res_b_l, eC_l, cluster_idx=cidx_l, invG=invG_q_l, sqrtRX=sqrtRX_q_l, crv2=crv2)%*%invG_q_l
      V_T_rb_r = invG_q_r%*%rdrobust_vce(dT+dZ, sV_T, as.matrix(R_q_r*W_h_r), res_b_r, eC_r, cluster_idx=cidx_r, invG=invG_q_r, sqrtRX=sqrtRX_q_r, crv2=crv2)%*%invG_q_r
    } else if (hb_match & !is.null(cluster)) {
      V_T_rb_l = invG_q_l%*%rdrobust_vce(dT+dZ, sV_T, as.matrix(R_q_l*W_h_l), res_b_l, eC_l, cluster_idx=cidx_l, invG=NULL, sqrtRX=NULL, crv2=FALSE)%*%invG_q_l
      V_T_rb_r = invG_q_r%*%rdrobust_vce(dT+dZ, sV_T, as.matrix(R_q_r*W_h_r), res_b_r, eC_r, cluster_idx=cidx_r, invG=NULL, sqrtRX=NULL, crv2=FALSE)%*%invG_q_r
    } else if (crv3 | crv2) {
      M_T_rb_l = .rdrobust_vce_qq_cluster(Q_q_l, R_q_l, W_b_l, invG_q_l, res_b_l, eC_l, cluster_idx=cidx_l, crv2=crv2, s=sV_T, d=dT+dZ)
      M_T_rb_r = .rdrobust_vce_qq_cluster(Q_q_r, R_q_r, W_b_r, invG_q_r, res_b_r, eC_r, cluster_idx=cidx_r, crv2=crv2, s=sV_T, d=dT+dZ)
      V_T_rb_l = invG_p_l %*% M_T_rb_l %*% invG_p_l
      V_T_rb_r = invG_p_r %*% M_T_rb_r %*% invG_p_r
    } else {
      # See V_Y_rb path-B comment above (k_override aligns CR1 df with h=b).
      V_T_rb_l = invG_p_l%*%rdrobust_vce(dT+dZ, sV_T, as.matrix(Q_q_l), res_b_l, eC_l, cluster_idx=cidx_l, invG=NULL, sqrtRX=NULL, crv2=FALSE, k_override=q+1)%*%invG_p_l
      V_T_rb_r = invG_p_r%*%rdrobust_vce(dT+dZ, sV_T, as.matrix(Q_q_r), res_b_r, eC_r, cluster_idx=cidx_r, invG=NULL, sqrtRX=NULL, crv2=FALSE, k_override=q+1)%*%invG_p_r
    }
		V_T_cl = factorial(deriv)^2*(V_T_cl_l+V_T_cl_r)[deriv+1,deriv+1]
		V_T_rb = factorial(deriv)^2*(V_T_rb_l+V_T_rb_r)[deriv+1,deriv+1]
		se_tau_T_cl = sqrt(V_T_cl);	se_tau_T_rb = sqrt(V_T_rb)
	}
  
	if (is.null(fuzzy)) {
	  if (is.null(covs)) {
	    if      (deriv==0) rdmodel = "Sharp RD estimates using local polynomial regression." 
	    else if (deriv==1) rdmodel = "Sharp Kink RD estimates using local polynomial regression."	
	    else               rdmodel = "Sharp RD estimates using local polynomial regression. Derivative of order d" 
		}
		else {
			if      (deriv==0) rdmodel = "Covariate-adjusted Sharp RD estimates using local polynomial regression." 
			else if (deriv==1) rdmodel = "Covariate-adjusted Sharp Kink RD estimates using local polynomial regression."	
			else               rdmodel = paste("Covariate-adjusted Sharp RD estimates using local polynomial regression. Derivative of order ", deriv, ".")	
	  }
	} else {
	  if (is.null(covs)) {
	    if      (deriv==0) rdmodel = "Fuzzy RD estimates using local polynomial regression." 
	    else if (deriv==1) rdmodel = "Fuzzy Kink RD estimates using local polynomial regression."	
	    else               rdmodel = paste("Fuzzy RD estimates using local polynomial regression. Derivative of order ", deriv, ".")	
		}
		else {
			if      (deriv==0) rdmodel = "Covariate-adjusted Fuzzy RD estimates using local polynomial regression." 
			else if (deriv==1) rdmodel = "Covariate-adjusted Fuzzy Kink RD estimates using local polynomial regression."	
			else               rdmodel = paste("Covariate-adjusted Fuzzy RD estimates using local polynomial regression. Derivative of order ", deriv, ".")			
	  }
	}

  n_clust <- NULL
  if (!is.null(cluster)) {
    n_clust <- c(g_l, g_r)
    rdmodel <- paste0(rdmodel, " Std. errors are clustered (", g_l + g_r, " clusters).")
  }

  tau = c(tau_cl, tau_bc, tau_bc)
  se  = c(se_tau_cl,se_tau_cl,se_tau_rb)
  t   =  tau/se
  pv  = 2*pnorm(-abs(t))
  ci  = matrix(NA,nrow=3,ncol=2)
  rownames(ci) = c("Conventional","Bias-Corrected","Robust")
  colnames(ci) = c("Lower","Upper")
  ci[1,] = c(tau[1] - quant*se[1], tau[1] + quant*se[1])
  ci[2,] = c(tau[2] - quant*se[2], tau[2] + quant*se[2])
  ci[3,] = c(tau[3] - quant*se[3], tau[3] + quant*se[3])
    
  if (!is.null(fuzzy)) {
      tau_T = c(tau_T_cl, tau_T_bc, tau_T_bc)
      se_T  = c(se_tau_T_cl, se_tau_T_cl, se_tau_T_rb)
      z_T   = tau_T/se_T
      pv_T  = 2*pnorm(-abs(z_T))
      ci_T  = matrix(NA,nrow=3,ncol=2)
      ci_T[1,] = c(tau_T[1] - quant*se_T[1], tau_T[1] + quant*se_T[1])
      ci_T[2,] = c(tau_T[2] - quant*se_T[2], tau_T[2] + quant*se_T[2])
      ci_T[3,] = c(tau_T[3] - quant*se_T[3], tau_T[3] + quant*se_T[3])
  }

    coef = matrix(tau,3,1)
    se   = matrix(se, 3,1)
    z    = matrix(t,  3,1)
    pv   = matrix(pv, 3,1)
    ci   = ci

  bws=matrix(c(h_l,b_l,h_r,b_r),2,2)
  colnames(bws)=c("left","right")
  rownames(bws)=c("h","b")
  
  rownames(coef)=rownames(se)=rownames(se)=rownames(z)=rownames(pv)=c("Conventional","Bias-Corrected","Robust")
  colnames(coef)="Coeff"
  colnames(se)="Std. Err."
  colnames(z)="z"
  colnames(pv)="P>|z|"
  colnames(bws)=c("left","right")
  rownames(ci)=c("Conventional","Bias-Corrected","Robust")
  colnames(ci)=c("CI Lower","CI Upper")
    
  Estimate=matrix(NA,1,4)
  colnames(Estimate)=c("tau.us","tau.bc","se.us","se.rb")
  Estimate[1,] <- c(tau_cl,tau_bc, se_tau_cl, se_tau_rb) 

  if (is.null(fuzzy)) {
  out <- list(Estimate = Estimate, bws = bws, coef = coef, se = se, z = z, pv = pv, ci = ci,
           beta_Y_p_l = beta_Y_p_l, beta_Y_p_r = beta_Y_p_r,
           V_cl_l=V_Y_cl_l, V_cl_r=V_Y_cl_r, V_rb_l=V_Y_rb_l, V_rb_r=V_Y_rb_r,
           N=c(N_l,N_r), N_h=c(N_h_l,N_h_r), N_b=c(N_b_l,N_b_r), M=c(M_l,M_r),
           tau_cl=c(tau_Y_cl_l,tau_Y_cl_r), tau_bc=c(tau_Y_bc_l,tau_Y_bc_r),
           c=c, p=p, q=q, bias=c(bias_l,bias_r), kernel=kernel_type,
           vce=vce_type, bwselect=bwselect, level=level, masspoints=masspoints,
           rdmodel=rdmodel, n_clust=n_clust, coef_covs=gamma_p)
  } else {
    out <- list(Estimate = Estimate, bws = bws, coef = coef, se = se, z = z, pv = pv, ci = ci,
             beta_Y_p_l = beta_Y_p_l, beta_Y_p_r = beta_Y_p_r,
             beta_T_p_l = beta_T_p_l, beta_T_p_r = beta_T_p_r,
             tau_T = tau_T, se_T = se_T, z_T = z_T, pv_T = pv_T, ci_T = ci_T,
             V_cl_l=V_Y_cl_l, V_cl_r=V_Y_cl_r, V_rb_l=V_Y_rb_l, V_rb_r=V_Y_rb_r,
             N=c(N_l,N_r), N_h=c(N_h_l,N_h_r), N_b=c(N_b_l,N_b_r), M=c(M_l,M_r),
             tau_cl=c(tau_Y_cl_l,tau_Y_cl_r), tau_bc=c(tau_Y_bc_l,tau_Y_bc_r),
             c=c, p=p, q=q, bias=c(bias_l,bias_r), kernel=kernel_type,
             vce=vce_type, bwselect=bwselect, level=level, masspoints=masspoints,
             rdmodel=rdmodel, n_clust=n_clust, coef_covs=gamma_p)
    }
  
  out$call <- match.call()
  class(out) <- "rdrobust"
  return(out)


}


## Shared header used by print.rdrobust + print.summary.rdrobust.
.print_rdrobust_header <- function(x) {
  cat("Call: rdrobust\n\n")
  cat(paste(x$rdmodel,"\n", sep=""))
  cat("\n")
  cat(paste("Number of Obs.           ",  format(x$N[1]+x$N[2], width=10, justify="right"),"\n", sep=""))
  cat(paste("BW type                  ",  format(x$bwselect, width=10, justify="right"),"\n", sep=""))
  cat(paste("Kernel                   ",  format(x$kernel,   width=10, justify="right"),"\n", sep=""))
  cat(paste("VCE method               ",  format(x$vce,      width=10, justify="right"),"\n", sep=""))
  cat("\n")
  cat(paste(format("", width=25), format("Left",  width=10, justify="right"), "   ", format("Right", width=10, justify="right"), "\n", sep=""))
  cat(paste("Number of Obs.           ",  format(x$N[1],  width=10, justify="right"),  "   ", format(x$N[2],   width=10, justify="right"), "\n", sep=""))
  cat(paste("Eff. Number of Obs.      ",  format(x$N_h[1],width=10, justify="right"),  "   ", format(x$N_h[2], width=10, justify="right"), "\n", sep=""))
  cat(paste("Order est. (p)           ",  format(x$p,     width=10, justify="right"),  "   ", format(x$p,      width=10, justify="right"), "\n", sep=""))
  cat(paste("Order bias (q)           ",  format(x$q,     width=10, justify="right"),  "   ", format(x$q,      width=10, justify="right"), "\n", sep=""))
  cat(paste("BW est. (h)              ",  format(sprintf("%10.3f",x$bws[1,1])),        "   ", format(sprintf("%10.3f",x$bws[1,2])),        "\n", sep=""))
  cat(paste("BW bias (b)              ",  format(sprintf("%10.3f",x$bws[2,1])),        "   ", format(sprintf("%10.3f",x$bws[2,2])),        "\n", sep=""))
  cat(paste("rho (h/b)                ",  format(sprintf("%10.3f",x$bws[1,1]/x$bws[2,1])), "   ", format(sprintf("%10.3f",x$bws[1,2]/x$bws[2,2])), "\n", sep=""))
  if (!is.null(x$n_clust)) cat(paste("Clusters (g)             ",  format(x$n_clust[1], width=10, justify="right"), "   ", format(x$n_clust[2], width=10, justify="right"), "\n", sep=""))
  if (isTRUE(x$masspoints=="adjust") || isTRUE(x$masspoints=="check"))
    cat(paste("Unique Obs.              ",  format(x$M[1], width=10, justify="right"), "   ", format(x$M[2], width=10, justify="right"), "\n", sep=""))
  cat("\n")
  invisible()
}

print.rdrobust <- function(x, ...) {
  .print_rdrobust_header(x)
  invisible(x)
}

summary.rdrobust <- function(object, ..., detail = FALSE, all = FALSE) {
  out <- object
  out$.detail <- isTRUE(detail)
  out$.all    <- isTRUE(all)
  class(out)  <- c("summary.rdrobust", class(object))
  out
}

print.summary.rdrobust <- function(x, ...) {
  .print_rdrobust_header(x)

  detail <- isTRUE(x$.detail)
  all    <- isTRUE(x$.all)

  ### read stored CIs, z-statistics and p-values
  CI_us_l <- x$ci[1, 1]; CI_us_r <- x$ci[1, 2]
  CI_bc_l <- x$ci[2, 1]; CI_bc_r <- x$ci[2, 2]
  CI_rb_l <- x$ci[3, 1]; CI_rb_r <- x$ci[3, 2]

  t_us <- x$z[1, 1]; t_bc <- x$z[2, 1]; t_rb <- x$z[3, 1]
  pv_us <- x$pv[1, 1]; pv_bc <- x$pv[2, 1]; pv_rb <- x$pv[3, 1]
  
  
  if (!is.null(x$tau_T)) {
    
    cat(paste("First-stage estimates.","\n", sep=""))
    cat(paste("","\n", sep=""))    
    
    ### print output

    if (isTRUE(detail) | isTRUE(all)) {
    
    llength = 14 + 10 + 8 + 10 + 10 +  25
    
    cat(paste(rep("=", llength), collapse="")); cat("\n")
    
    cat(format("Method"          , width=14, justify="right"))
    cat(format("Coef."           , width=10, justify="right"))
    cat(format("Std. Err."       , width=10 , justify="right"))
    cat(format("z"               , width=10, justify="right"))
    cat(format("P>|z|"           , width=10, justify="right"))
    cat(format(paste("[ ", x$level, "%", " C.I. ]", sep=""), width=25, justify="centre"))
    cat("\n")
    
    cat(paste(rep("=", llength), collapse="")); cat("\n")
    } else {
      
      llength = 14 + 10  + 10 + 10 + 25
      
      cat(paste(rep("=", llength), collapse="")); cat("\n")
      
      cat(format(""          , width=14, justify="right"))
      cat(format("Point"           , width=10, justify="right"))
      cat(format("Robust Inference"               , width=20, justify="right"))
      cat("\n")
      
      cat(format(""          , width=14, justify="right"))
      cat(format("Estimate"           , width=10, justify="right"))
      cat(format("z"               , width=10, justify="right"))
      cat(format("P>|z|"           , width=10, justify="right"))
      cat(format(paste("[ ", x$level, "%", " C.I. ]", sep=""), width=25, justify="centre"))
      cat("\n")
      
      cat(paste(rep("=", llength), collapse="")); cat("\n")
    
    
      cat(format("Rd Effect", width=14, justify="right"))
      cat(format(      sprintf("%3.3f", x$tau_T[1]) , width=10, justify="right"))
      cat(format(sprintf("%3.3f", x$z_T[3]) , width=10, justify="right"))
      cat(format(sprintf("%3.3f", x$pv_T[3]), width=10, justify="right"))
      cat(format(paste("[", sprintf("%3.3f", x$ci_T[3,1]), " , ", sep="")  , width=14, justify="right"))
      cat(format(paste(sprintf("%3.3f", x$ci_T[3,2]), "]", sep=""), width=11, justify="left"))
     cat("\n")
    }
    
    if (isTRUE(detail)) {
        
    cat(format("Conventional", width=14, justify="right"))
    cat(format(      sprintf("%3.3f", x$tau_T[1]) , width=10, justify="right"))
    cat(format(paste(sprintf("%3.3f", x$se_T[1]), sep=""), width=10, justify="right"))
    cat(format(      sprintf("%3.3f", x$z_T[1]) , width=10, justify="right"))
    cat(format(      sprintf("%3.3f", x$pv_T[1]), width=10, justify="right"))
    cat(format(paste("[", sprintf("%3.3f", x$ci_T[1,1]), " , ", sep="")  , width=14, justify="right"))
    cat(format(paste(     sprintf("%3.3f", x$ci_T[1,2]), "]", sep=""), width=11, justify="left"))
    cat("\n")

        
      cat(format("Robust", width=14, justify="right"))
      cat(format("-", width=10, justify="right"))
      cat(format("-", width=10, justify="right"))
       cat(format(sprintf("%3.3f", x$z_T[3]) , width=10, justify="right"))
      cat(format(sprintf("%3.3f", x$pv_T[3]), width=10, justify="right"))
      cat(format(paste("[", sprintf("%3.3f", x$ci_T[3,1]), " , ", sep="")  , width=14, justify="right"))
      cat(format(paste(sprintf("%3.3f", x$ci_T[3,2]), "]", sep=""), width=11, justify="left"))
      cat("\n") 
      }
      
    if (isTRUE(all)) {
      
      cat(format("Conventional", width=14, justify="right"))
      cat(format(      sprintf("%3.3f", x$tau_T[1]) , width=10, justify="right"))
      cat(format(paste(sprintf("%3.3f", x$se_T[1]), sep=""), width=10, justify="right"))
      cat(format(      sprintf("%3.3f", x$z_T[1]) , width=10, justify="right"))
      cat(format(      sprintf("%3.3f", x$pv_T[1]), width=10, justify="right"))
      cat(format(paste("[", sprintf("%3.3f", x$ci_T[1,1]), " , ", sep="")  , width=14, justify="right"))
      cat(format(paste(     sprintf("%3.3f", x$ci_T[1,2]), "]", sep=""), width=11, justify="left"))
      cat("\n")
      
      cat(format("Bias-Corrected", width=14, justify="right"))
      cat(format(sprintf("%3.3f", x$tau_T[2]) , width=10, justify="right"))
      cat(format(paste(sprintf("%3.3f", x$se_T[2]), sep=""), width=10, justify="right"))
      cat(format(sprintf("%3.3f", x$z_T[2]) , width=10, justify="right"))
      cat(format(sprintf("%3.3f", x$pv_T[2]), width=10, justify="right"))
      cat(format(paste("[", sprintf("%3.3f", x$ci_T[2,1]), " , ", sep="")  , width=14, justify="right"))
      cat(format(paste(sprintf("%3.3f", x$ci_T[2,2]), "]", sep=""), width=11, justify="left"))
      cat("\n")
      
      cat(format("Robust", width=14, justify="right"))
      cat(format(sprintf("%3.3f", x$tau_T[3]) , width=10, justify="right"))
      cat(format(paste(sprintf("%3.3f", x$se_T[3]), sep=""), width=10, justify="right"))
      cat(format(sprintf("%3.3f", x$z_T[3]) , width=10, justify="right"))
      cat(format(sprintf("%3.3f", x$pv_T[3]), width=10, justify="right"))
      cat(format(paste("[", sprintf("%3.3f", x$ci_T[3,1]), " , ", sep="")  , width=14, justify="right"))
      cat(format(paste(sprintf("%3.3f", x$ci_T[3,2]), "]", sep=""), width=11, justify="left"))
      cat("\n")
    }
    
    cat(paste(rep("=", llength), collapse="")); cat("\n")
    
    cat(paste("","\n", sep="")) 
    cat(paste("Treatment effect estimates.","\n", sep=""))
    cat(paste("","\n", sep=""))   
  }
  
  ### print output
  if (isTRUE(detail) | isTRUE(all)) {
    
    llength = 14 + 10 + 8 + 10 + 10 + 10 + 25
    cat(paste(rep("=", llength), collapse="")); cat("\n")
    
    cat(format("Method"          , width=14, justify="right"))
    cat(format("Coef."           , width=10, justify="right"))
    cat(format("Std. Err."       , width=10 , justify="right"))
    cat(format("z"               , width=10, justify="right"))
    cat(format("P>|z|"           , width=10, justify="right"))
    cat(format(paste("[ ", x$level, "%", " C.I. ]", sep=""), width=25, justify="centre"))
    cat("\n")
    
    cat(paste(rep("=", llength), collapse="")); cat("\n")
  } else {
    
    llength = 14 + 10 + 10+ 10+ 25
    cat(paste(rep("=", llength), collapse="")); cat("\n")
    
    
    cat(format(""          , width=14, justify="right"))
    cat(format("Point"           , width=10, justify="right"))
    cat(format("Robust Inference"               , width=20, justify="right"))
    cat("\n")
    
    
    cat(format(""               , width=14, justify="right"))
    cat(format("Estimate"       , width=10, justify="right"))
    cat(format("z"              , width=10, justify="right"))
    cat(format("P>|z|"          , width=10, justify="right"))
    cat(format(paste("[ ", x$level, "%", " C.I. ]", sep=""), width=25, justify="centre"))
    cat("\n")
    
    cat(paste(rep("-", llength), collapse="")); cat("\n")
  
    cat(format("RD Effect", width=14, justify="right"))
    cat(format(sprintf("%3.3f", x$Estimate[1, "tau.us"]) , width=10, justify="right"))
    cat(format(sprintf("%3.3f", t_rb) , width=10, justify="right"))
    cat(format(sprintf("%3.3f", pv_rb), width=10, justify="right"))
    cat(format(paste("[", sprintf("%3.3f", CI_rb_l[1]), " , ", sep="")  , width=14, justify="right"))
    cat(format(paste(sprintf("%3.3f", CI_rb_r[1]), "]", sep=""), width=11, justify="left"))
    cat("\n") 
    
  }
    
    if (isTRUE(detail)) {
      
    cat(format("Conventional", width=14, justify="right"))
    cat(format(sprintf("%3.3f", x$Estimate[1, "tau.us"]) , width=10, justify="right"))
    cat(format(paste(sprintf("%3.3f", x$Estimate[1, "se.us"]), sep=""), width=10, justify="right"))
    cat(format(sprintf("%3.3f", t_us) , width=10, justify="right"))
    cat(format(sprintf("%3.3f", pv_us), width=10, justify="right"))
    cat(format(paste("[", sprintf("%3.3f", CI_us_l[1]), " , ", sep="")  , width=14, justify="right"))
    cat(format(paste(sprintf("%3.3f", CI_us_r[1]), "]", sep=""), width=11, justify="left"))
    cat("\n")
  
      cat(format("Robust", width=14, justify="right"))
      cat(format("-", width=10, justify="right"))
      cat(format("-", width=10, justify="right"))
      cat(format(sprintf("%3.3f", t_rb) , width=10, justify="right"))
      cat(format(sprintf("%3.3f", pv_rb), width=10, justify="right"))
      cat(format(paste("[", sprintf("%3.3f", CI_rb_l[1]), " , ", sep="")  , width=14, justify="right"))
      cat(format(paste(sprintf("%3.3f", CI_rb_r[1]), "]", sep=""), width=11, justify="left"))
      cat("\n") 
  
    }      
      
      if (isTRUE(all)) {
        
        cat(format("Conventional", width=14, justify="right"))
        cat(format(sprintf("%3.3f", x$Estimate[1, "tau.us"]) , width=10, justify="right"))
        cat(format(paste(sprintf("%3.3f", x$Estimate[1, "se.us"]), sep=""), width=10, justify="right"))
        cat(format(sprintf("%3.3f", t_us) , width=10, justify="right"))
        cat(format(sprintf("%3.3f", pv_us), width=10, justify="right"))
        cat(format(paste("[", sprintf("%3.3f", CI_us_l[1]), " , ", sep="")  , width=14, justify="right"))
        cat(format(paste(sprintf("%3.3f", CI_us_r[1]), "]", sep=""), width=11, justify="left"))
        cat("\n")
        
      cat(format("Bias-Corrected", width=14, justify="right"))
      cat(format(sprintf("%3.3f", x$Estimate[1, "tau.bc"]) , width=10, justify="right"))
      cat(format(paste(sprintf("%3.3f", x$Estimate[1, "se.us"]), sep=""), width=10, justify="right"))
      cat(format(sprintf("%3.3f", t_bc) , width=10, justify="right"))
      cat(format(sprintf("%3.3f", pv_bc), width=10, justify="right"))
      cat(format(paste("[", sprintf("%3.3f", CI_bc_l[1]), " , ", sep="")  , width=14, justify="right"))
      cat(format(paste(sprintf("%3.3f", CI_bc_r[1]), "]", sep=""), width=11, justify="left"))
      cat("\n")
  
      cat(format("Robust", width=14, justify="right"))
      cat(format(sprintf("%3.3f", x$Estimate[1, "tau.bc"]) , width=10, justify="right"))
      cat(format(paste(sprintf("%3.3f", x$Estimate[1, "se.rb"]), sep=""), width=10, justify="right"))
      cat(format(sprintf("%3.3f", t_rb) , width=10, justify="right"))
      cat(format(sprintf("%3.3f", pv_rb), width=10, justify="right"))
      cat(format(paste("[", sprintf("%3.3f", CI_rb_l[1]), " , ", sep="")  , width=14, justify="right"))
      cat(format(paste(sprintf("%3.3f", CI_rb_r[1]), "]", sep=""), width=11, justify="left"))
      cat("\n")
    }

  cat(paste(rep("=", llength), collapse="")); cat("\n")

  invisible(x)
}


coef.rdrobust <- function(object, ...) {
  out <- as.vector(object$coef)
  names(out) <- rownames(object$coef)
  out
}

vcov.rdrobust <- function(object, ...) {
  se <- as.vector(object$se)
  V  <- diag(se^2, nrow = length(se), ncol = length(se))
  dimnames(V) <- list(rownames(object$se), rownames(object$se))
  V
}


## broom-compatible tidy()/glance() methods.
##
## Registered in NAMESPACE via `S3method(broom::tidy, <class>)` so they
## dispatch when broom is loaded, without requiring broom at install time.

tidy.rdrobust <- function(object, ...){
  ret <- data.frame(term = row.names(object$coef), 
                    estimate  = object$coef[, 1], 
                    std.error = object$se[, 1], 
                    statistic = object$z[, 1],
                    p.value   = object$pv[, 1], 
                    conf.low  = object$ci[,1],
                    conf.high = object$ci[, 2])
  row.names(ret) <- NULL
  ret
}

glance.rdrobust <- function(object, ...){
  ## glance() returns a single-row data.frame; split bilateral quantities
  ## into .left/.right columns so cluster counts (which are per-side) don't
  ## recycle into two rows.
  has_clust <- !is.null(object$n_clust)
  data.frame(nobs.left  = object$N[1],
             nobs.right = object$N[2],
             nobs.effective.left  = object$N_h[1],
             nobs.effective.right = object$N_h[2],
             h.left  = object$bws[1, 1],
             h.right = object$bws[1, 2],
             b.left  = object$bws[2, 1],
             b.right = object$bws[2, 2],
             cutoff = object$c,
             order.regression = object$p,
             order.bias = object$q,
             kernel  = object$kernel,
             vce     = object$vce,
             bwselect = object$bwselect,
             covariates = !is.null(object$coef_covs),
             clusters.left  = if (has_clust) object$n_clust[1] else NA_integer_,
             clusters.right = if (has_clust) object$n_clust[2] else NA_integer_,
             stringsAsFactors = FALSE)
}

utils::globalVariables(c("xmid", "ymid", "sz", "x", "ylo", "yhi",
                          "est", "label", "lo", "hi"))

plot.rdrobust = function(x, y, x_run, nbins = 20,
                         ci          = TRUE,
                         show_effect = FALSE,
                         title       = NULL,
                         x.label     = "Running Variable",
                         y.label     = "Outcome",
                         x.lim       = NULL,
                         y.lim       = NULL,
                         col.l       = "#3B7DD8",
                         col.r       = "#D95F3B",
                         base_size   = 14,
                         ...) {

  obj <- x   # rdrobust object; rename for clarity

  # --- extract key quantities --------------------------------------------------
  c_val   <- obj$c
  h_l     <- obj$bws["h", "left"]
  h_r     <- obj$bws["h", "right"]
  coefs_l <- as.numeric(obj$beta_Y_p_l)
  coefs_r <- as.numeric(obj$beta_Y_p_r)
  V_l     <- obj$V_cl_l
  V_r     <- obj$V_cl_r
  kernel  <- obj$kernel

  tau_cl   <- obj$coef["Conventional",  1]
  ci_rb    <- obj$ci["Robust", ]
  lev      <- obj$level
  z_crit   <- qnorm(1 - (1 - lev / 100) / 2)
  is_fuzzy <- !is.null(obj$tau_T)

  # ggplot text sizes: annotate/geom_text use mm, ~1pt = 0.353mm
  txt_main <- base_size * 0.353        # body annotations
  txt_sub  <- (base_size - 2) * 0.353  # footnote / secondary

  # --- Healy-inspired theme ----------------------------------------------------
  theme_rd <- function() {
    ggplot2::theme_bw(base_size = base_size) +
      ggplot2::theme(
        panel.border        = ggplot2::element_blank(),
        axis.line           = ggplot2::element_line(color = "grey30",
                                                    linewidth = 0.4),
        panel.grid.major    = ggplot2::element_line(color = "grey92",
                                                    linewidth = 0.4),
        panel.grid.minor    = ggplot2::element_blank(),
        plot.title          = ggplot2::element_text(
                                face = "bold",
                                size = base_size + 2,
                                margin = ggplot2::margin(b = 6)),
        axis.title          = ggplot2::element_text(size = base_size),
        axis.text           = ggplot2::element_text(size = base_size - 1,
                                                    color = "grey30"),
        plot.margin         = ggplot2::margin(12, 16, 8, 8),
        plot.caption        = ggplot2::element_text(size = base_size - 3,
                                                    color = "grey50",
                                                    hjust = 0,
                                                    margin = ggplot2::margin(t = 6))
      )
  }

  # --- kernel weight function --------------------------------------------------
  kern_w <- function(u) {
    u <- abs(u)
    switch(tolower(substr(kernel, 1, 3)),
      "tri" = pmax(1 - u,            0),
      "epa" = pmax(0.75 * (1 - u^2), 0),
      "uni" = as.numeric(u <= 1) * 0.5,
      pmax(1 - u, 0)
    )
  }

  # --- filter to bandwidth window + drop NAs -----------------------------------
  keep   <- !is.na(x_run) & !is.na(y) &
            (x_run >= c_val - h_l) & (x_run <= c_val + h_r)
  x_w    <- x_run[keep]
  y_w    <- y[keep]
  side   <- ifelse(x_w < c_val, "left", "right")

  kw <- ifelse(side == "left",
               kern_w((x_w - c_val) / h_l),
               kern_w((x_w - c_val) / h_r))

  # --- binned means + average kernel weight per bin ----------------------------
  make_bins <- function(xv, yv, wv, nb) {
    brks <- seq(min(xv), max(xv), length.out = nb + 1)
    grp  <- cut(xv, brks, include.lowest = TRUE)
    out  <- data.frame(xmid = as.numeric(tapply(xv, grp, mean, na.rm = TRUE)),
                       ymid = as.numeric(tapply(yv, grp, mean, na.rm = TRUE)),
                       wavg = as.numeric(tapply(wv, grp, mean, na.rm = TRUE)))
    out[stats::complete.cases(out), ]
  }

  idx_l  <- side == "left"
  idx_r  <- side == "right"
  bins_l <- make_bins(x_w[idx_l], y_w[idx_l], kw[idx_l], nbins)
  bins_r <- make_bins(x_w[idx_r], y_w[idx_r], kw[idx_r], nbins)

  # rescale kernel weights to point size range [1.5, 5]
  all_w  <- c(bins_l$wavg, bins_r$wavg)
  w_min  <- min(all_w, na.rm = TRUE)
  w_rng  <- max(all_w, na.rm = TRUE) - w_min
  if (w_rng == 0) w_rng <- 1
  rescale_sz <- function(w) 1.5 + 3.5 * (w - w_min) / w_rng
  bins_l$sz  <- rescale_sz(bins_l$wavg)
  bins_r$sz  <- rescale_sz(bins_r$wavg)

  # --- polynomial fit + pointwise CI bands -------------------------------------
  poly_fit <- function(xv, coefs, c0) {
    xc  <- xv - c0
    mat <- outer(xc, 0:length(coefs[-1]), "^")
    as.numeric(mat %*% coefs)
  }

  poly_se <- function(xv, V, c0) {
    xc   <- xv - c0
    mat  <- outer(xc, 0:(ncol(V) - 1), "^")
    vars <- rowSums((mat %*% V) * mat)
    sqrt(pmax(vars, 0))
  }

  xseq_l <- seq(c_val - h_l, c_val,       length.out = 200)
  xseq_r <- seq(c_val,       c_val + h_r, length.out = 200)
  yhat_l <- poly_fit(xseq_l, coefs_l, c_val)
  yhat_r <- poly_fit(xseq_r, coefs_r, c_val)
  fit_l  <- data.frame(x = xseq_l, y = yhat_l)
  fit_r  <- data.frame(x = xseq_r, y = yhat_r)

  if (ci) {
    se_l   <- poly_se(xseq_l, V_l, c_val)
    se_r   <- poly_se(xseq_r, V_r, c_val)
    band_l <- data.frame(x   = xseq_l,
                         ylo = yhat_l - z_crit * se_l,
                         yhi = yhat_l + z_crit * se_l)
    band_r <- data.frame(x   = xseq_r,
                         ylo = yhat_r - z_crit * se_r,
                         yhi = yhat_r + z_crit * se_r)
  }

  y0_l <- coefs_l[1]
  y0_r <- coefs_r[1]

  # --- RD plot annotation ------------------------------------------------------
  lbl <- sprintf("RD = %.3f\n%d%% RBC CI: [%.3f, %.3f]%s",
                 round(tau_cl,     3), lev,
                 round(ci_rb[[1]], 3),
                 round(ci_rb[[2]], 3),
                 if (is_fuzzy) "\n(Fuzzy)" else "")

  # --- build RD plot -----------------------------------------------------------
  p_rd <- ggplot2::ggplot()

  if (ci) {
    p_rd <- p_rd +
      ggplot2::geom_ribbon(data = band_l,
                           ggplot2::aes(x = x, ymin = ylo, ymax = yhi),
                           fill = col.l, alpha = 0.12) +
      ggplot2::geom_ribbon(data = band_r,
                           ggplot2::aes(x = x, ymin = ylo, ymax = yhi),
                           fill = col.r, alpha = 0.12)
  }

  p_rd <- p_rd +
    ggplot2::geom_point(data = bins_l,
                        ggplot2::aes(x = xmid, y = ymid, size = sz),
                        color = col.l, alpha = 0.85, show.legend = FALSE) +
    ggplot2::geom_point(data = bins_r,
                        ggplot2::aes(x = xmid, y = ymid, size = sz),
                        color = col.r, alpha = 0.85, show.legend = FALSE) +
    ggplot2::geom_line(data = fit_l,
                       ggplot2::aes(x = x, y = y),
                       color = col.l, linewidth = 1) +
    ggplot2::geom_line(data = fit_r,
                       ggplot2::aes(x = x, y = y),
                       color = col.r, linewidth = 1) +
    ggplot2::geom_vline(xintercept = c_val,
                        linetype = "dashed", color = "grey50", linewidth = 0.5) +
    ggplot2::geom_point(ggplot2::aes(x = c_val, y = y0_l),
                        shape = 21, size = 3.5,
                        color = col.l, fill = "white", stroke = 1.5) +
    ggplot2::geom_point(ggplot2::aes(x = c_val, y = y0_r),
                        shape = 21, size = 3.5,
                        color = col.r, fill = "white", stroke = 1.5) +
    ggplot2::annotate("text",
                      x = c_val + 0.04 * (h_l + h_r), y = Inf,
                      label = lbl,
                      hjust = 0, vjust = 1.4,
                      size = txt_main, color = "grey20",
                      lineheight = 1.3) +
    ggplot2::scale_size_identity() +
    ggplot2::labs(title = title, x = x.label, y = y.label) +
    theme_rd()

  if (!is.null(x.lim)) p_rd <- p_rd + ggplot2::xlim(x.lim)
  if (!is.null(y.lim)) p_rd <- p_rd + ggplot2::ylim(y.lim)

  # --- effect plot: conventional estimate + RBC CI with significance stars -----
  if (show_effect) {
    pv_rb <- obj$pv["Robust", 1]
    stars <- ifelse(pv_rb < 0.01, "***",
             ifelse(pv_rb < 0.05, "**",
             ifelse(pv_rb < 0.10, "*", "")))

    eff_lbl <- sprintf("%.3f%s  [%.3f, %.3f]",
                       tau_cl, stars, ci_rb[[1]], ci_rb[[2]])

    eff_df <- data.frame(label = "RD Effect",
                         est   = tau_cl,
                         lo    = ci_rb[[1]],
                         hi    = ci_rb[[2]])

    eff_size <- base_size - 4
    eff_txt  <- eff_size * 0.353

    p_eff <- ggplot2::ggplot(eff_df,
                             ggplot2::aes(x = est, y = label)) +
      ggplot2::geom_vline(xintercept = 0,
                          linetype = "dashed", color = "grey60",
                          linewidth = 0.4) +
      ggplot2::geom_errorbar(ggplot2::aes(xmin = lo, xmax = hi),
                             width = 0, linewidth = 0.8, color = "grey30",
                             orientation = "y") +
      ggplot2::geom_point(size = 3, color = "grey20") +
      ggplot2::annotate("text",
                        x = tau_cl, y = 1.35,
                        label = eff_lbl,
                        size  = eff_txt,
                        color = "grey20") +
      ggplot2::labs(
        x       = sprintf("Estimate (%d%% RBC CI)", lev),
        y       = NULL,
        caption = "Point estimate: conventional. CI: robust bias-corrected. * p<0.10  ** p<0.05  *** p<0.01"
      ) +
      ggplot2::scale_y_discrete(expand = ggplot2::expansion(add = c(0.4, 0.6))) +
      ggplot2::theme_bw(base_size = eff_size) +
      ggplot2::theme(
        panel.border        = ggplot2::element_blank(),
        axis.line.x         = ggplot2::element_line(color = "grey30",
                                                     linewidth = 0.4),
        panel.grid.major.x  = ggplot2::element_line(color = "grey92",
                                                     linewidth = 0.3),
        panel.grid.minor    = ggplot2::element_blank(),
        panel.grid.major.y  = ggplot2::element_blank(),
        axis.text.y         = ggplot2::element_text(size = eff_size,
                                                     color = "grey20"),
        axis.text.x         = ggplot2::element_text(size = eff_size - 1,
                                                     color = "grey30"),
        axis.title.x        = ggplot2::element_text(size = eff_size),
        axis.line.y         = ggplot2::element_blank(),
        axis.ticks.y        = ggplot2::element_blank(),
        plot.caption        = ggplot2::element_text(size = eff_size - 2,
                                                     color = "grey50",
                                                     hjust = 0),
        plot.margin         = ggplot2::margin(4, 16, 4, 8)
      )

    if (requireNamespace("gridExtra", quietly = TRUE)) {
      gridExtra::grid.arrange(p_rd, p_eff, ncol = 1, heights = c(3, 1))
    } else {
      print(p_rd)
      print(p_eff)
    }

    return(invisible(list(rd_plot = p_rd, effect_plot = p_eff)))
  }

  print(p_rd)
  invisible(p_rd)
}

rdbwselect = function(y, x, c = NULL, fuzzy = NULL, deriv = NULL, p = NULL, q = NULL,
                      covs = NULL,  covs_drop = TRUE, ginv.tol = 1e-20,
                      kernel = "tri", weights = NULL, bwselect = "mserd",
                      vce = "nn", cluster = NULL,
                      nnmatch = 3,  scaleregul = 1, sharpbw = FALSE,
                      all = NULL, subset = NULL, masspoints = "adjust",
                      bwcheck = NULL, bwrestrict = TRUE, stdvars = FALSE,
                      data = NULL){

  if (!is.null(data)) {
    .vars <- .rdrobust_resolve_data(match.call(), data, parent.frame(),
              c("y","x","covs","cluster","fuzzy","weights","subset"))
    y <- .vars$y; x <- .vars$x; covs <- .vars$covs
    cluster <- .vars$cluster; fuzzy <- .vars$fuzzy
    weights <- .vars$weights; subset <- .vars$subset
  }

  covs <- .rdrobust_resolve_covs(covs, data, parent.frame())

  # Validate auxiliary-vector lengths against length(x) BEFORE subset filtering.
  n_orig <- length(x)
  .rdrobust_check_length(y,       "y",       n_orig)
  .rdrobust_check_length(weights, "weights", n_orig)
  .rdrobust_check_length(fuzzy,   "fuzzy",   n_orig)
  .rdrobust_check_length(covs,    "covs",    n_orig)
  .rdrobust_check_length(cluster, "cluster", n_orig)
  .rdrobust_check_subset(subset, n_orig)

  if (!is.null(subset)) {
    x <- x[subset]
    y <- y[subset]
    if (length(x) == 0L)
      stop("'subset' removed all observations.", call. = FALSE)
  }

  if (is.null(all)) all<-FALSE
  if (is.null(c)) c <- 0
  if (length(c) != 1 || !is.numeric(c) || !is.finite(c))
    stop(sprintf("Cutoff 'c' must be a single finite numeric value (received: %s).", toString(c)), call. = FALSE)
    
  # p
  if (is.null(p) & !is.null(deriv)) {p = deriv+1}
  if (length(p) == 0) {
    flag_no_p <- TRUE
    p <- 1
  } else if ((length(p) != 1) | !(p[1]%in%0:20)) {
    stop(sprintf("Polynomial order 'p' must be a single integer in 0:20 (received: %s).", toString(p)))
  } else {
    flag_no_p <- FALSE
  }

  # q
  if (length(q) == 0) {
    flag_no_q <- TRUE
    q <- p + 1
  } else if ((length(q) != 1) | !(q[1]%in%0:20) | (q[1]<=p)) {
    stop(sprintf("Polynomial order 'q' (for bias correction) must be a single integer in 0:20 with q > p (received q=%s, p=%s).", toString(q), toString(p)))
  } else {
    flag_no_q <- FALSE
  }

  # deriv
  if (length(deriv) == 0) {
    flag_no_deriv <- TRUE
    deriv <- 0
  } else if ((length(deriv) != 1) | !(deriv[1]%in%0:20) | (deriv[1]>p)) {
    stop(sprintf("Derivative order must be a single integer in 0:20 with deriv <= p (received deriv=%s, p=%s).", toString(deriv), toString(p)))
  } else {
    flag_no_deriv <- FALSE
  }
  
  na.ok <- complete.cases(x) & complete.cases(y)
  
  if (!is.null(cluster)){
    if (!is.null(subset))  cluster <- cluster[subset]
    if (length(cluster) != length(x))
      stop("cluster must have the same length as y and x.")
    na.ok <- na.ok & complete.cases(cluster)
  }
  
  if (!is.null(covs)){
    #if (!is.null(subset))  covs <- subset(covs,subset)
    if (!is.null(subset)) covs <- covs[subset, ,drop=FALSE]
    na.ok <- na.ok & complete.cases(covs)
  } 
  
  if (!is.null(fuzzy)){
    if (!is.null(subset)) fuzzy <- fuzzy[subset]
    na.ok <- na.ok & complete.cases(fuzzy)
  } 
  
  if (!is.null(weights)){
    if (!is.null(subset)) weights <- weights[subset]
    na.ok <- na.ok & complete.cases(weights) & weights>=0
  } 
  
  x = as.matrix(x[na.ok])
  y = as.matrix(y[na.ok])
  if (!is.null(covs))    covs    = as.matrix(covs)[na.ok, , drop = FALSE]
  if (!is.null(fuzzy))   fuzzy   = as.matrix(fuzzy[na.ok])
  if (!is.null(cluster)) cluster = as.matrix(cluster[na.ok])
  if (!is.null(weights)) weights = as.matrix(weights[na.ok])
  
  if (is.null(masspoints)) masspoints=FALSE
  
  if (vce=="nn" | masspoints=="check" | masspoints=="adjust") {
    order_x = order(x)
    x = x[order_x,,drop=FALSE]
    y = y[order_x,,drop=FALSE]
    if (!is.null(covs))    covs    =  as.matrix(covs)[order_x,,drop=FALSE]
    if (!is.null(fuzzy))   fuzzy   =   fuzzy[order_x,,drop=FALSE]
    if (!is.null(cluster)) cluster = cluster[order_x,,drop=FALSE]
    if (!is.null(weights)) weights = weights[order_x,,drop=FALSE]
  }
  
  ### reescaling
  x_iq = quantile(x,.75,type=2) - quantile(x,.25,type=2)
  BWp = min(c(sd(x),x_iq/1.349))
  
  x_sd = y_sd = 1
  c_orig = c
  if (isTRUE(stdvars)) { 
    y_sd = sd(y)
    x_sd = sd(x)
    y = y/y_sd
	  x = x/x_sd
	  c = c/x_sd
	  BWp = min(c(1,(x_iq/x_sd)/1.349))
  }
  ###############################################
  ind_l = x<c; ind_r = x>=c
  X_l = x[ind_l,,drop=FALSE];   X_r = x[ind_r,,drop=FALSE]
  Y_l = y[ind_l,,drop=FALSE];   Y_r = y[ind_r,,drop=FALSE]
  N_l = nrow(X_l);   N_r = nrow(X_r)
  x_min=min(x);  x_max=max(x)
  range_l = abs(c-x_min);  range_r = abs(c-x_max)
  N = N_l + N_r

  M_l = N_l;  M_r = N_r

  # X_uniq_l/r and M_l/M_r are needed by both the masspoints inspection
  # branch AND the bwcheck branch. Hoist so bwcheck works even when
  # masspoints='off'.
  if (masspoints=="check" | masspoints=="adjust" | !is.null(bwcheck)) {
    X_uniq_l = sort(unique(X_l), decreasing=TRUE)
    X_uniq_r = unique(X_r)
    M_l = length(X_uniq_l)
    M_r = length(X_uniq_r)
    M = M_l + M_r
  }
  if (masspoints=="check" | masspoints=="adjust") {
    mass_l = 1-M_l/N_l
    mass_r = 1-M_r/N_r
    if (mass_l>=0.2 | mass_r>=0.2){
      warning("Mass points detected in the running variable.")
      if (masspoints=="check") warning("Try using option masspoints=adjust.")
      if (is.null(bwcheck) & masspoints=="adjust") bwcheck <- 10
    }
  }
  


  ############## COLLINEARITY
  covs_drop_coll=dZ=0
  if (!is.null(covs)) dZ = ncol(covs)
  if (isTRUE(covs_drop)) covs_drop_coll = 1
  
  if (!is.null(covs) & isTRUE(covs_drop)) {
    covs.names = colnames(covs)
    if (is.null(covs.names)) {
      covs.names = paste("z",1:ncol(covs),sep="")
      colnames(covs) = covs.names
    }
    covs = covs[,order(nchar(covs.names))]
    covs = as.matrix(covs)
    dZ = length(covs.names)
    covs.check = covs_drop_fun(covs)
    if (covs.check$ncovs < dZ) {
      covs  <- as.matrix(covs.check$covs)
      dZ    <- covs.check$ncovs
      warning("Multicollinearity issue detected in covs. Redundant covariates dropped.")  
    }
  }
  
    exit=0
    #################  ERRORS
    if (kernel!="uni" & kernel!="uniform" & kernel!="tri" & kernel!="triangular" & kernel!="epa" & kernel!="epanechnikov" & kernel!="" ){
      warning("kernel incorrectly specified")
      exit = 1
    }
    
    valid_bwselect <- c("mserd","msetwo","msesum","msecomb1","msecomb2",
                        "cerrd","certwo","cersum","cercomb1","cercomb2","")
    valid_vce      <- c("nn","hc0","hc1","hc2","hc3","cr1","cr2","cr3","")

    if (!bwselect %in% valid_bwselect) {
      if (bwselect %in% c("cct","ik","cv","CCT","IK","CV")) {
        warning("bwselect options IK, CCT and CV have been deprecated. Please see help for new options")
      } else {
        warning("bwselect incorrectly specified")
      }
      exit = 1
    }

    if (!vce %in% valid_vce) {
      warning("vce incorrectly specified")
      exit = 1
    }

    if (c<=x_min | c>=x_max){
      warning("c should be set within the range of x")
      exit = 1
    }
    
    if (p<0 | q<0 | deriv<0 | nnmatch<=0 ){
      warning("p, q, deriv and matches should be positive integers")
      exit = 1
    }
    
    if (p>=q){
      warning("q must be strictly greater than p")
      exit = 1
    }
    
    if (deriv>p){
      warning("deriv can only be equal or lower p")
      exit = 1
    }
    
    p_round = round(p)/p;    q_round = round(q)/q;    d_round = round(deriv+1)/(deriv+1);    m_round = round(nnmatch)/nnmatch
        
    if ((p_round!=1 &p>0) | (q_round!=1&q>0) | d_round!=1 | m_round!=1 ){
      warning("p,q,deriv and matches should be integer numbers")
      exit = 1
    }
    
    if (N<20){
      warning("Not enough observations to perform bandwidth calculations")
      exit = 1
    }

    if (!is.null(weights) && sum(weights, na.rm = TRUE) <= 0) {
      warning("weights must include at least one positive value")
      exit = 1
    }

    if (!is.null(bwcheck)) {
      if (length(bwcheck) != 1 || !is.finite(bwcheck) || bwcheck < 1 || bwcheck != round(bwcheck)) {
        warning("bwcheck must be a single positive integer")
        exit = 1
      }
    }

    if (!is.null(masspoints) && !isFALSE(masspoints) &&
        !(is.character(masspoints) && masspoints %in% c("check","adjust","off",""))) {
      warning("masspoints must be one of 'check', 'adjust', 'off', or FALSE")
      exit = 1
    }

    if (exit>0) stop("rdbwselect: invalid input (see warnings above).")
    
  
  
  
  if (kernel %in% c("epanechnikov", "epa")) {
    kernel_type = "Epanechnikov"
    C_c=2.34
  }  else if (kernel %in% c("uniform", "uni")) {
    kernel_type = "Uniform"
    C_c=1.843
  }   else  {
    kernel_type = "Triangular"
    C_c=2.576
  }
  
  # Cluster vce validation: with cluster, only cr1/cr2/cr3 are valid.
  # Without cluster, cr1/cr2/cr3 fall back to hc1/hc2/hc3.
  if (!is.null(cluster)) {
    if (vce %in% c("nn","")) {
      vce = "cr1"  # silent default
    } else if (vce %in% c("hc0","hc1")) {
      warning(paste0("vce='", vce, "' is not a cluster option. Switching to vce='cr1'."))
      vce = "cr1"
    } else if (vce == "hc2") {
      warning("vce='hc2' is not a cluster option. Switching to vce='cr2'.")
      vce = "cr2"
    } else if (vce == "hc3") {
      warning("vce='hc3' is not a cluster option. Switching to vce='cr3'.")
      vce = "cr3"
    } else if (!vce %in% c("cr1","cr2","cr3")) {
      warning(paste0("vce='", vce, "' is not a valid cluster option. Switching to vce='cr1'."))
      vce = "cr1"
    }
  } else {
    if (vce=="cr1") {
      warning("vce='cr1' requires a cluster variable. Falling back to vce='hc1'.")
      vce = "hc1"
    } else if (vce=="cr2") {
      warning("vce='cr2' requires a cluster variable. Falling back to vce='hc2'.")
      vce = "hc2"
    } else if (vce=="cr3") {
      warning("vce='cr3' requires a cluster variable. Falling back to vce='hc3'.")
      vce = "hc3"
    }
  }

  # Map user-facing cr1/cr2/cr3 to internal vce values
  vce_type = "NN"
  if (vce=="hc0")  vce_type = "HC0"
  if (vce=="hc1")  vce_type = "HC1"
  if (vce=="hc2")  vce_type = "HC2"
  if (vce=="hc3")  vce_type = "HC3"
  if (vce=="cr1") { vce_type = "CR1"; vce = "hc1" }
  if (vce=="cr2") { vce_type = "CR2"; vce = "crv2" }
  if (vce=="cr3") { vce_type = "CR3"; vce = "crv3" }
    
  #***********************************************************************
  Z_l=Z_r=T_l=T_r=C_l=C_r=NULL
  g_l=g_r=0
  
  if (vce=="nn") {
    runs_l   = rle(as.vector(X_l))
    dups_l   = rep(runs_l$lengths, runs_l$lengths)
    dupsid_l = sequence(runs_l$lengths)
    runs_r   = rle(as.vector(X_r))
    dups_r   = rep(runs_r$lengths, runs_r$lengths)
    dupsid_r = sequence(runs_r$lengths)
  }
  
  if (!is.null(covs)) {
    Z_l  = covs[ind_l,,drop=FALSE];  Z_r  = covs[ind_r,,drop=FALSE]
  }
  
  perf_comp=FALSE
  if (!is.null(fuzzy)) {
    T_l  = fuzzy[ind_l,,drop=FALSE];  T_r  = fuzzy[ind_r,,drop=FALSE];
    # Reject fully degenerate first stage (no variation, no jump). One-sided
    # non-compliance falls through to the perf_comp branch.
    if (var(T_l) == 0 && var(T_r) == 0 && abs(mean(T_l) - mean(T_r)) < .Machine$double.eps^0.5) {
      stop("Fuzzy RD: first-stage variable has no variation and no jump at the cutoff. The fuzzy estimator is not identified.", call. = FALSE)
    }
    if (var(T_l)==0 | var(T_r)==0) perf_comp=TRUE
    if (isTRUE(perf_comp) | isTRUE(sharpbw)) {
      T_l = T_r = NULL
      }
    }
   
  if (!is.null(cluster)) {
    C_l  = cluster[ind_l,,drop=FALSE]; C_r= cluster[ind_r,,drop=FALSE]
    g_l = length(unique(C_l));	g_r = length(unique(C_r))
  }
  
  fw_l = fw_r = NULL
  if (!is.null(weights)) {
    fw_l=weights[ind_l];  fw_r=weights[ind_r]
  }                                                                           

  
  ######################################################################
    c_bw = C_c*BWp*N^(-1/5)
    if (masspoints=="adjust") c_bw = C_c*BWp*M^(-1/5)
    
    if (isTRUE(bwrestrict)) {
      bw_max_l = abs(c-x_min)
      bw_max_r = abs(c-x_max)
      bw_max = max(bw_max_l, bw_max_r)
      c_bw <- min(c_bw, bw_max)
    }
    
    bw.adj <- 0
    if (!is.null(bwcheck)) {
      bwcheck_l = min(bwcheck, M_l)
			bwcheck_r = min(bwcheck, M_r)
      bw_min_l = abs(X_uniq_l-c)[bwcheck_l] + 1e-8
      bw_min_r = abs(X_uniq_r-c)[bwcheck_r] + 1e-8
      c_bw = max(c_bw, bw_min_l, bw_min_r)
      bw.adj <- 1
    }
    
    
    ### T1: per-side V-fit caches reused across all pilot calls.
    vcache_l <- new.env(parent = emptyenv())
    vcache_r <- new.env(parent = emptyenv())

    ### Step 1: d_bw
    C_d_l = rdrobust_bw(Y_l, X_l, T_l, Z_l, C_l, fw_l, c=c, o=q+1, nu=q+1, o_B=q+2, h_V=c_bw, h_B=range_l, 0, vce, nnmatch, kernel, dups_l, dupsid_l, covs_drop_coll, ginv.tol, vcache = vcache_l)
    C_d_r = rdrobust_bw(Y_r, X_r, T_r, Z_r, C_r, fw_r, c=c, o=q+1, nu=q+1, o_B=q+2, h_V=c_bw, h_B=range_r, 0, vce, nnmatch, kernel, dups_r, dupsid_r, covs_drop_coll, ginv.tol, vcache = vcache_r)
    ### TWO
    if  (bwselect=="msetwo" |  bwselect=="certwo" | bwselect=="msecomb2" | bwselect=="cercomb2"  | isTRUE(all))  {		
      d_bw_l = c((  C_d_l$V              /   C_d_l$B^2             )^C_d_l$rate)
      d_bw_r = c((  C_d_r$V              /   C_d_r$B^2             )^C_d_r$rate)
      if (isTRUE(bwrestrict)) {
        d_bw_l <- min(d_bw_l, bw_max_l)
        d_bw_r <- min(d_bw_r, bw_max_r)
      }
      
      if (!is.null(bwcheck)) {
        d_bw_l  <- max(d_bw_l, bw_min_l)
        d_bw_r  <- max(d_bw_r, bw_min_r)
      }
      C_b_l  = rdrobust_bw(Y_l, X_l, T_l, Z_l, C_l, fw_l, c=c, o=q, nu=p+1, o_B=q+1, h_V=c_bw, h_B=d_bw_l, scaleregul, vce, nnmatch, kernel, dups_l, dupsid_l, covs_drop_coll, ginv.tol, vcache = vcache_l)
      b_bw_l = c((  C_b_l$V              /   (C_b_l$B^2 + scaleregul*C_b_l$R)        )^C_b_l$rate)
      C_b_r  = rdrobust_bw(Y_r, X_r, T_r, Z_r, C_r, fw_r, c=c, o=q, nu=p+1, o_B=q+1, h_V=c_bw, h_B=d_bw_r, scaleregul, vce, nnmatch, kernel, dups_r, dupsid_r, covs_drop_coll, ginv.tol, vcache = vcache_r)
      b_bw_r = c((  C_b_r$V              /   (C_b_r$B^2 + scaleregul*C_b_r$R)        )^C_b_r$rate)
      if (isTRUE(bwrestrict)) {
        b_bw_l <- min(b_bw_l, bw_max_l)
        b_bw_r <- min(b_bw_r, bw_max_r)
      }
      
      C_h_l  = rdrobust_bw(Y_l, X_l, T_l, Z_l, C_l, fw_l, c=c, o=p, nu=deriv, o_B=q, h_V=c_bw, h_B=b_bw_l, scaleregul, vce, nnmatch, kernel, dups_l, dupsid_l, covs_drop_coll, ginv.tol, vcache = vcache_l)
      h_bw_l = c((  C_h_l$V              /   (C_h_l$B^2 + scaleregul*C_h_l$R)         )^C_h_l$rate)
      C_h_r  = rdrobust_bw(Y_r, X_r, T_r, Z_r, C_r, fw_r, c=c, o=p, nu=deriv, o_B=q, h_V=c_bw, h_B=b_bw_r, scaleregul, vce, nnmatch, kernel, dups_r, dupsid_r, covs_drop_coll, ginv.tol, vcache = vcache_r)
      h_bw_r = c((  C_h_r$V              /   (C_h_r$B^2 + scaleregul*C_h_r$R)         )^C_h_r$rate)
      if (isTRUE(bwrestrict)) {
        h_bw_l <- min(h_bw_l, bw_max_l)
        h_bw_r <- min(h_bw_r, bw_max_r) 
      }
      
    }
  
  ### SUM
  if  (bwselect=="msesum" | bwselect=="cersum" |  bwselect=="msecomb1" | bwselect=="msecomb2" |  bwselect=="cercomb1" | bwselect=="cercomb2"  |  isTRUE(all))  {
    d_bw_s = c(( (C_d_l$V + C_d_r$V)  /  (C_d_r$B + C_d_l$B)^2 )^C_d_l$rate)
    if (isTRUE(bwrestrict)) {
    d_bw_s <- min(d_bw_s, bw_max)
    }
    
    if (!is.null(bwcheck)) d_bw_s  <-  max(d_bw_s, bw_min_l, bw_min_r)
    C_b_l  = rdrobust_bw(Y_l, X_l, T_l, Z_l, C_l, fw_l, c=c, o=q, nu=p+1, o_B=q+1, h_V=c_bw, h_B=d_bw_s, scaleregul, vce, nnmatch, kernel, dups_l, dupsid_l, covs_drop_coll, ginv.tol, vcache = vcache_l)
    C_b_r  = rdrobust_bw(Y_r, X_r, T_r, Z_r, C_r, fw_r, c=c, o=q, nu=p+1, o_B=q+1, h_V=c_bw, h_B=d_bw_s, scaleregul, vce, nnmatch, kernel, dups_r, dupsid_r, covs_drop_coll, ginv.tol, vcache = vcache_r)
    b_bw_s = c(( (C_b_l$V + C_b_r$V)  /  ((C_b_r$B + C_b_l$B)^2 + scaleregul*(C_b_r$R+C_b_l$R)) )^C_b_l$rate)
    if (isTRUE(bwrestrict)) {
    b_bw_s <- min(b_bw_s, bw_max)
    }
    C_h_l  = rdrobust_bw(Y_l, X_l, T_l, Z_l, C_l, fw_l, c=c, o=p, nu=deriv, o_B=q, h_V=c_bw, h_B=b_bw_s, scaleregul, vce, nnmatch, kernel, dups_l, dupsid_l, covs_drop_coll, ginv.tol, vcache = vcache_l)
    C_h_r  = rdrobust_bw(Y_r, X_r, T_r, Z_r, C_r, fw_r, c=c, o=p, nu=deriv, o_B=q, h_V=c_bw, h_B=b_bw_s, scaleregul, vce, nnmatch, kernel, dups_r, dupsid_r, covs_drop_coll, ginv.tol, vcache = vcache_r)
    h_bw_s = c(( (C_h_l$V + C_h_r$V)  /  ((C_h_r$B + C_h_l$B)^2 + scaleregul*(C_h_r$R + C_h_l$R)) )^C_h_l$rate)
    if (isTRUE(bwrestrict)) {
    h_bw_s <- min(h_bw_s, bw_max)
    }
}

    ### RD
if  (bwselect=="mserd" | bwselect=="cerrd" | bwselect=="msecomb1" | bwselect=="msecomb2" | bwselect=="cercomb1" | bwselect=="cercomb2" | bwselect=="" | isTRUE(all) ) {
  d_bw_d = c(( (C_d_l$V + C_d_r$V)  /  (C_d_r$B - C_d_l$B)^2 )^C_d_l$rate)
  if (isTRUE(bwrestrict)) {
    d_bw_d <- min(d_bw_d, bw_max)
  }
  if (!is.null(bwcheck)) d_bw_d  <- max(d_bw_d, bw_min_l, bw_min_r)
  C_b_l  = rdrobust_bw(Y_l, X_l, T_l, Z_l, C_l, fw_l, c=c, o=q, nu=p+1, o_B=q+1, h_V=c_bw, h_B=d_bw_d, scaleregul, vce, nnmatch, kernel, dups_l, dupsid_l, covs_drop_coll, ginv.tol, vcache = vcache_l)
  C_b_r  = rdrobust_bw(Y_r, X_r, T_r, Z_r, C_r, fw_r, c=c, o=q, nu=p+1, o_B=q+1, h_V=c_bw, h_B=d_bw_d, scaleregul, vce, nnmatch, kernel, dups_r, dupsid_r, covs_drop_coll, ginv.tol, vcache = vcache_r)
  b_bw_d = c(( (C_b_l$V + C_b_r$V)  /  ((C_b_r$B - C_b_l$B)^2 + scaleregul*(C_b_r$R + C_b_l$R)) )^C_b_l$rate)
  if (isTRUE(bwrestrict)) {
    b_bw_d <- min(b_bw_d, bw_max)
  }
  C_h_l  = rdrobust_bw(Y_l, X_l, T_l, Z_l, C_l, fw_l, c=c, o=p, nu=deriv, o_B=q, h_V=c_bw, h_B=b_bw_d, scaleregul, vce, nnmatch, kernel, dups_l, dupsid_l, covs_drop_coll, ginv.tol, vcache = vcache_l)
  C_h_r  = rdrobust_bw(Y_r, X_r, T_r, Z_r, C_r, fw_r, c=c, o=p, nu=deriv, o_B=q, h_V=c_bw, h_B=b_bw_d, scaleregul, vce, nnmatch, kernel, dups_r, dupsid_r, covs_drop_coll, ginv.tol, vcache = vcache_r)
  h_bw_d = c(( (C_h_l$V + C_h_r$V)  /  ((C_h_r$B - C_h_l$B)^2 + scaleregul*(C_h_r$R + C_h_l$R)) )^C_h_l$rate)
  if (isTRUE(bwrestrict)) {
  h_bw_d <- min(h_bw_d, bw_max)
  }
}	

if  (bwselect=="mserd" | bwselect=="cerrd" | bwselect=="msecomb1" | bwselect=="msecomb2" | bwselect=="cercomb1" | bwselect=="cercomb2" | bwselect=="" | isTRUE(all) ) {
  h_mserd = x_sd*h_bw_d
  b_mserd = x_sd*b_bw_d
}	
if  (bwselect=="msesum" | bwselect=="cersum" |  bwselect=="msecomb1" | bwselect=="msecomb2" |  bwselect=="cercomb1" | bwselect=="cercomb2"  |  isTRUE(all))  {
  h_msesum = x_sd*h_bw_s
  b_msesum = x_sd*b_bw_s
}
if  (bwselect=="msetwo" |  bwselect=="certwo" | bwselect=="msecomb2" | bwselect=="cercomb2"  | isTRUE(all))  {		
  h_msetwo_l = x_sd*h_bw_l
  h_msetwo_r = x_sd*h_bw_r
  b_msetwo_l = x_sd*b_bw_l
  b_msetwo_r = x_sd*b_bw_r
}
if  (bwselect=="msecomb1" | bwselect=="cercomb1" | isTRUE(all) ) {
  h_msecomb1 = min(c(h_mserd,h_msesum))
  b_msecomb1 = min(c(b_mserd,b_msesum))
}
if  (bwselect=="msecomb2" | bwselect=="cercomb2" |  isTRUE(all) ) {
  h_msecomb2_l = median(c(h_mserd,h_msesum,h_msetwo_l))
  h_msecomb2_r = median(c(h_mserd,h_msesum,h_msetwo_r))
  b_msecomb2_l = median(c(b_mserd,b_msesum,b_msetwo_l))
  b_msecomb2_r = median(c(b_mserd,b_msesum,b_msetwo_r))
}
cer_h = N^(-(p/((3+p)*(3+2*p))))

if (!is.null(cluster)) {
  cer_h = (g_l+g_r)^(-(p/((3+p)*(3+2*p))))
}

#cer_b = N^(-(q/((3+q)*(3+2*q))))
cer_b = 1
	if  (bwselect=="cerrd" | isTRUE(all) ){
		h_cerrd = h_mserd*cer_h
		b_cerrd = b_mserd*cer_b
	}
	if  (bwselect=="cersum" | isTRUE(all) ){
		h_cersum = h_msesum*cer_h
		b_cersum=  b_msesum*cer_b
		}
	if  (bwselect=="certwo" | isTRUE(all) ){
		h_certwo_l   = h_msetwo_l*cer_h
		h_certwo_r   = h_msetwo_r*cer_h
		b_certwo_l   = b_msetwo_l*cer_b
		b_certwo_r   = b_msetwo_r*cer_b
		}
	if  (bwselect=="cercomb1" | isTRUE(all) ){
		h_cercomb1 = h_msecomb1*cer_h
		b_cercomb1 = b_msecomb1*cer_b
		}
	if  (bwselect=="cercomb2" | isTRUE(all) ){
		h_cercomb2_l = h_msecomb2_l*cer_h
		h_cercomb2_r = h_msecomb2_r*cer_h
		b_cercomb2_l = b_msecomb2_l*cer_b
		b_cercomb2_r = b_msecomb2_r*cer_b
	}

if (isFALSE(all)){
  bws = matrix(NA,1,4)
  colnames(bws)=c("h (left)","h (right)","b (left)","b (right)")
  rownames(bws)=bwselect
  if  (bwselect=="mserd" | bwselect=="") bws[1,] = c(h_mserd,      h_mserd,      b_mserd,      b_mserd)
  if  (bwselect=="msetwo")               bws[1,] = c(h_msetwo_l,   h_msetwo_r,   b_msetwo_l,   b_msetwo_r)
  if  (bwselect=="msesum")               bws[1,] = c(h_msesum,     h_msesum,     b_msesum,     b_msesum)
  if  (bwselect=="msecomb1")             bws[1,] = c(h_msecomb1,   h_msecomb1,   b_msecomb1,   b_msecomb1)
  if  (bwselect=="msecomb2")             bws[1,] = c(h_msecomb2_l, h_msecomb2_r, b_msecomb2_l, b_msecomb2_r) 
  if  (bwselect=="cerrd")                bws[1,] = c(h_cerrd,      h_cerrd,      b_cerrd,      b_cerrd)
  if  (bwselect=="certwo")               bws[1,] = c(h_certwo_l,   h_certwo_r,   b_certwo_l,   b_certwo_r)
  if  (bwselect=="cersum")               bws[1,] = c(h_cersum,     h_cersum,     b_cersum,     b_cersum)
  if  (bwselect=="cercomb1")             bws[1,] = c(h_cercomb1,   h_cercomb1,   b_cercomb1,   b_cercomb1)
  if  (bwselect=="cercomb2")             bws[1,] = c(h_cercomb2_l, h_cercomb2_r, b_cercomb2_l, b_cercomb2_r)
}

  if (isTRUE(all)){
    bwselect="All"
    bws = matrix(NA,10,4)
    colnames(bws)=c("h (left)","h (right)","b (left)","b (right)")
    rownames(bws)=c("mserd","msetwo","msesum","msecomb1","msecomb2","cerrd","certwo","cersum","cercomb1","cercomb2")
    bws[1,] =c(h_mserd,      h_mserd,      b_mserd,      b_mserd)
    bws[2,] =c(h_msetwo_l,   h_msetwo_r,   b_msetwo_l,   b_msetwo_r)
    bws[3,] =c(h_msesum,     h_msesum,     b_msesum,     b_msesum)
    bws[4,] =c(h_msecomb1,   h_msecomb1,   b_msecomb1,   b_msecomb1)
    bws[5,] =c(h_msecomb2_l, h_msecomb2_r, b_msecomb2_l, b_msecomb2_r) 
    bws[6,] =c(h_cerrd,      h_cerrd,      b_cerrd,      b_cerrd)
    bws[7,] =c(h_certwo_l,   h_certwo_r,   b_certwo_l,   b_certwo_r)
    bws[8,] =c(h_cersum,     h_cersum,     b_cersum,     b_cersum)
    bws[9,] =c(h_cercomb1,   h_cercomb1,   b_cercomb1,   b_cercomb1)
    bws[10,]=c(h_cercomb2_l, h_cercomb2_r, b_cercomb2_l, b_cercomb2_r)
  }



### Eff N
w_h_l <- rdrobust_kweight(X_l,c,bws[1,1],kernel)
w_h_r <- rdrobust_kweight(X_r,c,bws[1,2],kernel)
N_h_l <- sum(w_h_l> 0)
N_h_r <- sum(w_h_r> 0)

  out = list(bws=bws,
             bwselect=bwselect, kernel=kernel_type, p=p, q=q, c=c,
             N=c(N_l,N_r), N_h=c(N_h_l,N_h_r), M=c(M_l,M_r), vce=vce_type, masspoints=masspoints)
  out$call <- match.call()
  class(out) <- "rdbwselect"
  return(out)
}

.print_rdbwselect_header <- function(x) {
  cat("Call: rdbwselect\n\n")
  cat(paste("Number of Obs.           ",  format(x$N[1]+x$N[2], width=10, justify="right"),"\n", sep=""))
  cat(paste("BW type                  ",  format(x$bwselect, width=10, justify="right"),"\n", sep=""))
  cat(paste("Kernel                   ",  format(x$kernel,   width=10, justify="right"),"\n", sep=""))
  cat(paste("VCE method               ",  format(x$vce,      width=10, justify="right"),"\n", sep=""))
  cat("\n")
  cat(paste(format("", width=25), format("Left",  width=10, justify="right"), "   ", format("Right", width=10, justify="right"), "\n", sep=""))
  cat(paste("Number of Obs.           ",  format(x$N[1],   width=10, justify="right"),  "   ", format(x$N[2],   width=10, justify="right"), "\n", sep=""))
  cat(paste("Eff. Number of Obs.      ",  format(x$N_h[1], width=10, justify="right"),  "   ", format(x$N_h[2], width=10, justify="right"), "\n", sep=""))
  cat(paste("Order est. (p)           ",  format(x$p,      width=10, justify="right"),  "   ", format(x$p,      width=10, justify="right"), "\n", sep=""))
  cat(paste("Order bias (q)           ",  format(x$q,      width=10, justify="right"),  "   ", format(x$q,      width=10, justify="right"), "\n", sep=""))
  if (isTRUE(x$masspoints == "adjust") || isTRUE(x$masspoints == "check"))
    cat(paste("Unique Obs.              ", format(x$M[1], width=10, justify="right"), "   ", format(x$M[2], width=10, justify="right"), "\n", sep=""))
  cat("\n")
  invisible()
}

print.rdbwselect <- function(x, ...) {
  .print_rdbwselect_header(x)
  invisible(x)
}

summary.rdbwselect <- function(object, ...) {
  out <- object
  class(out) <- c("summary.rdbwselect", class(object))
  out
}

print.summary.rdbwselect <- function(x, ...) {
  .print_rdbwselect_header(x)

  col1.names = c("","BW est. (h)", "BW bias (b)")
  col2.names = c("","Left of c", "Right of c","Left of c", "Right of c")

  cat(paste(rep("=", 15 + 10*4), collapse="")); cat("\n")
  cat(format(col1.names, width=14, justify="right"))
  cat("\n")
  cat(format(col2.names, width=10, justify="right"))
  cat("\n")
  cat(paste(rep("=", 15 + 10*4), collapse="")); cat("\n")

  for (j in 1:nrow(x$bws)) {
    cat(format(rownames(x$bws)[j], width=10, justify="right"))
    cat(format(sprintf("%3.3f", x$bws[j,]), width=10, justify="right"))
    cat("\n")
  }
  cat(paste(rep("=", 15 + 10*ncol(x$bws)), collapse="")); cat("\n")

  invisible(x)
}

tidy.rdbwselect <- function(object, ...) {
  bws <- as.data.frame(object$bws)
  bws$type <- rownames(object$bws)
  rownames(bws) <- NULL
  bws[, c("type", setdiff(names(bws), "type")), drop = FALSE]
}

glance.rdbwselect <- function(object, ...) {
  data.frame(
    nobs.left  = object$N[1],
    nobs.right = object$N[2],
    nobs.effective.left  = object$N_h[1],
    nobs.effective.right = object$N_h[2],
    cutoff           = object$c,
    order.regression = object$p,
    order.bias       = object$q,
    kernel           = object$kernel,
    vce              = object$vce,
    bwselect         = object$bwselect,
    stringsAsFactors = FALSE
  )
}

rdplot = function(y, x, c = 0, p = 4, nbins = NULL, binselect = "esmv", scale = NULL,
                  kernel = "uni", weights = NULL, h = NULL,
                  covs = NULL,  covs_eval = "mean", covs_drop = TRUE, ginv.tol = 1e-20,
                  support = NULL, subset = NULL, masspoints = "adjust",
                  hide = FALSE, ci = NULL, shade = FALSE,
                  title = NULL, x.label = NULL, y.label = NULL, x.lim = NULL, y.lim = NULL,
                  col.dots = NULL, col.lines = NULL,
                  data = NULL) {

  if (!is.null(data)) {
    .vars <- .rdrobust_resolve_data(match.call(), data, parent.frame(),
              c("y","x","covs","weights","subset"))
    y <- .vars$y; x <- .vars$x; covs <- .vars$covs
    weights <- .vars$weights; subset <- .vars$subset
  }

  covs <- .rdrobust_resolve_covs(covs, data, parent.frame())

  # Validate auxiliary-vector lengths against length(x) BEFORE subset filtering.
  n_orig <- length(x)
  .rdrobust_check_length(y,       "y",       n_orig)
  .rdrobust_check_length(weights, "weights", n_orig)
  .rdrobust_check_length(covs,    "covs",    n_orig)
  .rdrobust_check_subset(subset, n_orig)

  if (!is.null(subset)) {
    x <- x[subset]
    y <- y[subset]
    if (length(x) == 0L)
      stop("'subset' removed all observations.", call. = FALSE)
  }

  if (length(c) != 1 || !is.numeric(c) || !is.finite(c))
    stop(sprintf("Cutoff 'c' must be a single finite numeric value (received: %s).", toString(c)), call. = FALSE)

  na.ok <- complete.cases(x) & complete.cases(y)
  
  if (!is.null(covs)){
    if (!is.null(subset))  covs <- covs[subset, , drop=FALSE]
    na.ok <- na.ok & complete.cases(covs)
    
    message("covs option is meant to be used when plotting RDROBUST estimates. If the option is used for global polynomial fitting, it may deliver results that are not visually compatible with the local binned means depicted due to the underlying assumptions used.")
    
  } 
  
  if (!is.null(weights)){
    if (!is.null(subset)) weights <- weights[subset]
    na.ok <- na.ok & complete.cases(weights) & weights>=0
  } 
  
  x <- x[na.ok]
  y <- y[na.ok]
  
  if (!is.null(covs))    covs    = as.matrix(covs)[na.ok, , drop = FALSE]
  if (!is.null(weights)) weights = as.matrix(weights[na.ok])  
  
  x_min = min(x);	x_max = max(x)
  ind_l = x<c;  ind_r = x>=c
  x_l = x[ind_l]; x_r = x[ind_r]	
  y_l = y[ind_l];	y_r = y[ind_r]
	
	if (!is.null(support)) {
	  support_l = support[1]
	  support_r = support[2]
	  if (support_l<x_min) x_min = support_l
	  if (support_r>x_max) x_max = support_r
	}
	
	range_l = c - x_min
	range_r = x_max - c
	
	n_l = length(x_l)
	n_r = length(x_r)
	n = n_l + n_r
  meth="es"
  
  if (is.null(scale)) {
    scale = scale_l = scale_r = 1  
  } else{
    if (length(scale)==1) scale_l = scale_r = scale
    if (length(scale)==2) {
      scale_l = scale[1]
      scale_r = scale[2]
    }
  }
  
  if (!is.null(nbins)) {
    if (length(nbins)==1) nbins_l = nbins_r = nbins
    if (length(nbins)==2) {
      nbins_l = nbins[1]
      nbins_r = nbins[2]
    }
  }
  
  if (is.null(h)) {
    h_l = range_l
		h_r = range_r
  } else{
    if (length(h)==1) h_l = h_r = h
    if (length(h)==2) {
      h_l = h[1]
      h_r = h[2]
    }
  }
  
  flag_no_ci <- FALSE
  if (is.null(ci)) {
    ci<- 95
    flag_no_ci <- TRUE
  }
  
  kernel_type = "Uniform"
  if (kernel %in% c("epanechnikov", "epa")) kernel_type = "Epanechnikov"
  if (kernel %in% c("triangular",   "tri")) kernel_type = "Triangular"
  
  ### Mass Points
  if (is.null(masspoints)) masspoints=FALSE
  mN = n;  M_l = n_l;  M_r = n_r
  if (masspoints=="check" | masspoints=="adjust") {
    X_uniq_l = sort(unique(x_l), decreasing=TRUE)
    X_uniq_r = unique(x_r)
    M_l = length(X_uniq_l)
    M_r = length(X_uniq_r)
    M = M_l + M_r
    mass_l = 1-M_l/n_l
    mass_r = 1-M_r/n_r				
    if (mass_l>=0.2 | mass_r>=0.2){
      warning("Mass points detected in the running variable.")
      if (masspoints=="check") message("Try using option masspoints=adjust.")
      if (masspoints=="adjust") {
        if (binselect=="es")    binselect="espr"
        if (binselect=="esmv")  binselect="esmvpr"
        if (binselect=="qs")    binselect="qspr"
        if (binselect=="qsmv")  binselect="qsmvpr"
      }
    }
  }				
  
  ############## COLLINEARITY
  covs_drop_coll=dZ=0
  if (!is.null(covs)) dZ = ncol(covs)
  if (isTRUE(covs_drop)) covs_drop_coll = 1
  
  if (!is.null(covs) & isTRUE(covs_drop)) {
    covs.names = colnames(covs)
    if (is.null(covs.names)) {
      covs.names = paste("z",1:ncol(covs),sep="")
      colnames(covs) = covs.names
    }
    covs = covs[,order(nchar(covs.names))]
    covs = as.matrix(covs)
    dZ = length(covs.names)
    covs.check = covs_drop_fun(covs)
    if (covs.check$ncovs < dZ) {
      covs  <- as.matrix(covs.check$covs)
      dZ    <- covs.check$ncovs
      warning("Multicollinearity issue detected in covs. Redundant covariates dropped.")  
    }
  }
  
  #####  ERRORS
  exit=0
	if (c<=x_min | c>=x_max){
		message("c should be set within the range of x")
		exit = 1
	}
  
  if (kernel!="uni" & kernel!="uniform" & kernel!="tri" & kernel!="triangular" & kernel!="epa" & kernel!="epanechnikov" & kernel!="" ){
    message("kernel incorrectly specified")
    exit = 1
  }

	if (p<0 ){
		message("p should be a positive number")
		exit = 1
	}

	if (scale<=0 |scale_l<=0 |scale_r<=0){
		message("scale should be a positive number")
		exit = 1
	}

	p_ceiling = ceiling(p)/p

	if (p_ceiling!=1 & p>0) {
  	message("p should be an integer number")
  	exit = 1
	}
	
	if (n<20){
	  message("Not enough observations to perform bin calculations")
	  exit = 1
	}

	if (!is.null(masspoints) && !isFALSE(masspoints) &&
	    !(is.character(masspoints) && masspoints %in% c("check","adjust","off",""))) {
	  message("masspoints must be one of 'check', 'adjust', 'off', or FALSE")
	  exit = 1
	}

	if (exit>0) stop()

	###################################################################
	### Polynomial curve (order = p) ##################################
	###################################################################
	W_h_l = rdrobust_kweight(x_l, c, h_l, kernel)
	W_h_r = rdrobust_kweight(x_r, c, h_r, kernel)
	
	n_h_l = sum(W_h_l>0)
	n_h_r = sum(W_h_r>0)
	
	if (!is.null(weights)) {
	  fw_l = weights[ind_l];  W_h_l = fw_l*W_h_l
	  fw_r = weights[ind_r];	W_h_r = fw_r*W_h_r
	}

	R_p_l = outer(x_l-c, Y=0:p, FUN = "^")
	R_p_r = outer(x_r-c, Y=0:p, FUN = "^")	
	invG_p_l  = qrXXinv((sqrt(W_h_l)*R_p_l))	
	invG_p_r  = qrXXinv((sqrt(W_h_r)*R_p_r))
	
	gamma_p = NULL
	if (is.null(covs)) {
	  gamma_p1_l = invG_p_l%*%crossprod(R_p_l*W_h_l, y_l)	
    gamma_p1_r = invG_p_r%*%crossprod(R_p_r*W_h_r, y_r)
  } else {
	  z_l  = covs[ind_l,,drop=FALSE]; z_r  = covs[ind_r,,drop=FALSE]
	  D_l  = cbind(y_l,z_l); D_r = cbind(y_r,z_r)
	  U_p_l = crossprod(R_p_l*W_h_l,D_l); U_p_r = crossprod(R_p_r*W_h_r,D_r)
	  beta_p_l = invG_p_l%*%crossprod(R_p_l*W_h_l,D_l); beta_p_r = invG_p_r%*%crossprod(R_p_r*W_h_r,D_r); 
	  
	  ZWD_p_l  = crossprod(z_l*W_h_l,D_l)
	  ZWD_p_r  = crossprod(z_r*W_h_r,D_r)
	  colsZ = 2:max(c(2+dZ-1,2))
	  UiGU_p_l =  crossprod(U_p_l[,colsZ],invG_p_l%*%U_p_l) 
	  UiGU_p_r =  crossprod(U_p_r[,colsZ],invG_p_r%*%U_p_r) 
	  ZWZ_p_l = ZWD_p_l[,colsZ] - UiGU_p_l[,colsZ] 
	  ZWZ_p_r = ZWD_p_r[,colsZ] - UiGU_p_r[,colsZ]     
	  ZWY_p_l = ZWD_p_l[,1] - UiGU_p_l[,1] 
	  ZWY_p_r = ZWD_p_r[,1] - UiGU_p_r[,1]     
	  ZWZ_p = ZWZ_p_r + ZWZ_p_l
	  ZWY_p = ZWY_p_r + ZWY_p_l
	  if (covs_drop_coll == 0) gamma_p = chol2inv(chol(ZWZ_p))%*%ZWY_p
	  if (covs_drop_coll == 1) gamma_p = ginv(ZWZ_p, tol = ginv.tol)%*%ZWY_p
	  s_Y = c(1 ,  -gamma_p[,1])
	  
	  gamma_p1_l = t(s_Y%*%t(beta_p_l))
	  gamma_p1_r = t(s_Y%*%t(beta_p_r))
	}
	
	
	###############################################
	### Preparte data for polynomial curve plot ###
	###############################################

	nplot = 500
	x_plot_l = seq(c-h_l, c, length.out = nplot)
	x_plot_r = seq(c, c+h_r, length.out = nplot)
	
	rplot_l = outer(x_plot_l-c, Y = 0:p, FUN = "^")
	rplot_r = outer(x_plot_r-c, Y = 0:p, FUN = "^")
	
	y_hat_l = rplot_l%*%gamma_p1_l
	y_hat_r = rplot_r%*%gamma_p1_r

	if (!is.null(covs) & covs_eval=="mean" ) {
	  gammaZ = colMeans(covs)%*%gamma_p
	  y_hat_l = rplot_l%*%gamma_p1_l + c(gammaZ)
	  y_hat_r = rplot_r%*%gamma_p1_r + c(gammaZ)
	}
	
	
	###############################################
  ### Optimal Bins (using polynomial order k) ###
	###############################################
	k=4
  
  rk_l = outer(x_l, Y = 0:k, FUN = "^")
  rk_r = outer(x_r, Y = 0:k, FUN = "^")
  
  invG_k_l = try(qrXXinv(rk_l),silent=TRUE)
  invG_k_r = try(qrXXinv(rk_r),silent=TRUE)
  
  if (inherits(invG_k_l, "try-error") | inherits(invG_k_r, "try-error")) {
    k = 3
    rk_l = outer(x_l, Y = 0:k, FUN = "^")
    rk_r = outer(x_r, Y = 0:k, FUN = "^")
    invG_k_l = try(qrXXinv(rk_l),silent=TRUE)
    invG_k_r = try(qrXXinv(rk_r),silent=TRUE)
  }

  if (inherits(invG_k_l, "try-error") | inherits(invG_k_r, "try-error")) {
    k = 2
    rk_l = outer(x_l, Y = 0:k, FUN = "^")
    rk_r = outer(x_r, Y = 0:k, FUN = "^")
    invG_k_l = qrXXinv(rk_l)
    invG_k_r = qrXXinv(rk_r)		
  }
  
  gamma_k1_l = invG_k_l%*%crossprod(rk_l, y_l)  
  gamma_k2_l = invG_k_l%*%crossprod(rk_l, y_l^2)
  gamma_k1_r = invG_k_r%*%crossprod(rk_r, y_r)  
  gamma_k2_r = invG_k_r%*%crossprod(rk_r, y_r^2)
  
  ### Bias w/sample
  mu0_k1_l = rk_l%*%gamma_k1_l
  mu0_k1_r = rk_r%*%gamma_k1_r
  mu0_k2_l = rk_l%*%gamma_k2_l
  mu0_k2_r = rk_r%*%gamma_k2_r
  drk_l = matrix(NA,n_l,k)
  drk_r = matrix(NA,n_r,k)
  for (j in 1:k) {
    drk_l[,j] = j*x_l^(j-1)
    drk_r[,j] = j*x_r^(j-1)
  }
  
  ind_l = order(x_l); ind_r = order(x_r)
  x_i_l = x_l[ind_l]; y_i_l = y_l[ind_l]
  x_i_r = x_r[ind_r]; y_i_r = y_r[ind_r]

  dxi_l=(x_i_l[2:n_l]-x_i_l[1:(n_l-1)]); dyi_l=(y_i_l[2:n_l]-y_i_l[1:(n_l-1)])
  dxi_r=(x_i_r[2:n_r]-x_i_r[1:(n_r-1)]); dyi_r=(y_i_r[2:n_r]-y_i_r[1:(n_r-1)])
                
  x_bar_i_l = (x_i_l[2:n_l] + x_i_l[1:(n_l-1)]) /2
  x_bar_i_r = (x_i_r[2:n_r] + x_i_r[1:(n_r-1)]) /2
                       
  drk_i_l = matrix(NA,n_l-1,k);	rk_i_l  = matrix(NA,n_l-1,(k+1))
  drk_i_r = matrix(NA,n_r-1,k);	rk_i_r  = matrix(NA,n_r-1,(k+1))
                       
  for (j in 1:(k+1)) {
    rk_i_l[,j] = x_bar_i_l^(j-1)    
    rk_i_r[,j] = x_bar_i_r^(j-1)
  }
                       
  for (j in 1:k) {
    drk_i_l[,j] = j*x_bar_i_l^(j-1)
    drk_i_r[,j] = j*x_bar_i_r^(j-1)
  }
  
  mu0_i_hat_l = rk_i_l%*%gamma_k1_l
  mu0_i_hat_r = rk_i_r%*%gamma_k1_r
  mu2_i_hat_l = rk_i_l%*%gamma_k2_l
  mu2_i_hat_r = rk_i_r%*%gamma_k2_r
                       
  mu0_hat_l = rk_l%*%gamma_k1_l
  mu0_hat_r = rk_r%*%gamma_k1_r
  mu2_hat_l = rk_l%*%gamma_k2_l
  mu2_hat_r = rk_r%*%gamma_k2_r
                       
  mu1_hat_l = drk_l%*%(gamma_k1_l[2:(k+1)])
  mu1_hat_r = drk_r%*%(gamma_k1_r[2:(k+1)])
                       
  mu1_i_hat_l = drk_i_l%*%(gamma_k1_l[2:(k+1)])
  mu1_i_hat_r = drk_i_r%*%(gamma_k1_r[2:(k+1)])
  
  var_y_l = var(y_l)
  var_y_r = var(y_r)
  
  sigma2_hat_l_bar = mu2_i_hat_l - mu0_i_hat_l^2
  sigma2_hat_r_bar = mu2_i_hat_r - mu0_i_hat_r^2
  ind_s2_l = sigma2_hat_l_bar<0
  ind_s2_r = sigma2_hat_r_bar<0
  sigma2_hat_l_bar[ind_s2_l] = var_y_l 
  sigma2_hat_r_bar[ind_s2_r] = var_y_r  
  
  sigma2_hat_l = mu2_hat_l - mu0_hat_l^2
  sigma2_hat_r = mu2_hat_r - mu0_hat_r^2
  ind_s2_l = sigma2_hat_l<0
  ind_s2_r = sigma2_hat_r<0
  sigma2_hat_l[ind_s2_l] = var_y_l 
  sigma2_hat_r[ind_s2_r] = var_y_r  
  
  B_es_hat_dw = c( ((c-x_min)^2/(12*n))*sum(mu1_hat_l^2),((x_max-c)^2/(12*n))*sum(mu1_hat_r^2))
  V_es_hat_dw = c((0.5/(c-x_min))*sum(dxi_l*dyi_l^2),(0.5/(x_max-c))*sum(dxi_r*dyi_r^2))
  V_es_chk_dw = c((1/(c-x_min))*sum(dxi_l*sigma2_hat_l_bar),(1/(x_max-c))*sum(dxi_r*sigma2_hat_r_bar))
  J_es_hat_dw = J.fun(B_es_hat_dw, V_es_hat_dw, n)
  J_es_chk_dw = J.fun(B_es_hat_dw, V_es_chk_dw, n)
  
  B_qs_hat_dw = c((n_l^2/(24*n))*sum(dxi_l^2*mu1_i_hat_l^2), (n_r^2/(24*n))*sum(dxi_r^2*mu1_i_hat_r^2))
  V_qs_hat_dw = c((1/(2*n_l))*sum(dyi_l^2),(1/(2*n_r))*sum(dyi_r^2))
  V_qs_chk_dw = c((1/n_l)*sum(sigma2_hat_l), (1/n_r)*sum(sigma2_hat_r))
  J_qs_hat_dw = J.fun(B_qs_hat_dw, V_qs_hat_dw, n)
  J_qs_chk_dw = J.fun(B_qs_hat_dw, V_qs_chk_dw, n)
  
  J_es_hat_mv  = c(ceiling((var_y_l/V_es_hat_dw[1])*(n/log(n)^2)), ceiling((var_y_r/V_es_hat_dw[2])*(n/log(n)^2)))
  J_es_chk_mv  = c(ceiling((var_y_l/V_es_chk_dw[1])*(n/log(n)^2)), ceiling((var_y_r/V_es_chk_dw[2])*(n/log(n)^2)))
  J_qs_hat_mv  = c(ceiling((var_y_l/V_qs_hat_dw[1])*(n/log(n)^2)), ceiling((var_y_r/V_qs_hat_dw[2])*(n/log(n)^2)))
  J_qs_chk_mv  = c(ceiling((var_y_l/V_qs_chk_dw[1])*(n/log(n)^2)), ceiling((var_y_r/V_qs_chk_dw[2])*(n/log(n)^2)))
  
  
  #########################################################
  if (binselect=="es") {
    J_star_orig = J_es_hat_dw
    meth="es"
    binselect_type="IMSE-optimal evenly-spaced method using spacings estimators"
    J_IMSE = J_es_hat_dw
    J_MV   = J_es_hat_mv
  }
  if (binselect=="espr") {
    J_star_orig = J_es_chk_dw
    meth="es"
    binselect_type="IMSE-optimal evenly-spaced method using polynomial regression"
    J_IMSE = J_es_chk_dw
    J_MV   = J_es_chk_mv
  }
  if (binselect=="esmv" ) {
    J_star_orig = J_es_hat_mv
    meth="es"
    binselect_type="mimicking variance evenly-spaced method using spacings estimators"
    J_IMSE = J_es_hat_dw
    J_MV   = J_es_hat_mv
  }
  if (binselect=="esmvpr" ) {
    J_star_orig = J_es_chk_mv
    meth="es"
    binselect_type="mimicking variance evenly-spaced method using polynomial regression"
    J_IMSE = J_es_chk_dw
    J_MV   = J_es_chk_mv
  }
  if (binselect=="qs" ) {
    J_star_orig = J_qs_hat_dw
    meth="qs"
    binselect_type="IMSE-optimal quantile-spaced method using spacings estimators"
    J_IMSE = J_qs_hat_dw
    J_MV   = J_qs_hat_mv
  }
  if (binselect=="qspr" ) {
    J_star_orig = J_qs_chk_dw
    meth="qs"
    binselect_type="IMSE-optimal quantile-spaced method using polynomial regression"
    J_IMSE = J_qs_chk_dw
    J_MV   = J_qs_chk_mv
  }
  if (binselect=="qsmv" ) {
    J_star_orig = J_qs_hat_mv
    meth="qs"
    binselect_type="mimicking variance quantile-spaced method using spacings estimators"
    J_IMSE = J_qs_hat_dw
    J_MV   = J_qs_hat_mv
  }
  if (binselect=="qsmvpr" ) {
    J_star_orig = J_qs_chk_mv
    meth="qs"
    binselect_type="mimicking variance quantile-spaced method using polynomial regression"
    J_IMSE = J_qs_chk_dw
    J_MV   = J_qs_chk_mv
  }

  J_star_l = scale_l*J_star_orig[1]
  J_star_r = scale_r*J_star_orig[2]

  if (!is.null(nbins)) {
    J_star_l = nbins_l
    J_star_r = nbins_r
    binselect_type="manually evenly spaced"
  }
  
  if (var_y_l==0) {
    J_star_l = J_star_l_orig = 1
    warning("Not enough variability in the outcome variable below the threshold")
  }

	if (var_y_r==0) {
	  J_star_r = J_star_r_orig = 1
	  warning("Not enough variability in the outcome variable above the threshold")
	}
  
  rscale_l = J_star_l / J_IMSE[1]
  rscale_r = J_star_r / J_IMSE[2]

  jump_l  = range_l/J_star_l;jump_r = range_r/J_star_r;
  
  if (meth=="es") {
    jumps_l=seq(x_min,c,jump_l)
    jumps_r=seq(c,x_max,jump_r)
  }   else if (meth=="qs") {
    jumps_l=quantile(x_l,probs=seq(0,1,1/J_star_l))
    jumps_r=quantile(x_r,probs=seq(0,1,1/J_star_r))
  }

	bin_x_l = findInterval(x_l, jumps_l,  rightmost.closed=TRUE) - J_star_l-1
	bin_x_r = findInterval(x_r, jumps_r,  rightmost.closed=TRUE)

	rdplot_bin_l    = sort(unique(bin_x_l))
	rdplot_mean_y_l = as.vector(tapply(y_l, bin_x_l, mean))
	rdplot_mean_x_l = as.vector(tapply(x_l, bin_x_l, mean))

	rdplot_bin_r    = sort(unique(bin_x_r))
	rdplot_mean_y_r = as.vector(tapply(y_r, bin_x_r, mean))
	rdplot_mean_x_r = as.vector(tapply(x_r, bin_x_r, mean))

	if (!is.null(covs) & covs_eval=="mean") {
	  covs_model_l = lm(y_l~ z_l + factor(bin_x_l))
	  covs_model_r = lm(y_r~ z_r + factor(bin_x_r))
	  yhatZ_l = predict(covs_model_l)
	  yhatZ_r = predict(covs_model_r)
	  rdplot_mean_y_l = as.vector(tapply(yhatZ_l, bin_x_l, mean))
	  rdplot_mean_y_r = as.vector(tapply(yhatZ_r, bin_x_r, mean))
	}
	
	t_ind_l = 1:J_star_l
	t_ind_r = 1:J_star_r
	rdplot_mean_bin_l = rowMeans(cbind(matrix(jumps_l[t_ind_l]),matrix(jumps_l[(t_ind_l+1)])))
	rdplot_mean_bin_r = rowMeans(cbind(matrix(jumps_r[t_ind_r]),matrix(jumps_r[(t_ind_r+1)])))

	#rdplot_mean_bin_l = rdplot_mean_bin_l[rev(-rdplot_bin_l)]
	rdplot_mean_bin_l = rdplot_mean_bin_l[(rdplot_bin_l+J_star_l+1)]
	rdplot_mean_bin_r = rdplot_mean_bin_r[rdplot_bin_r]
	
	bin_x = c(bin_x_l,bin_x_r)
  rdplot_mean_bin = c(rdplot_mean_bin_l, rdplot_mean_bin_r)
  rdplot_mean_x   = c(rdplot_mean_x_l,   rdplot_mean_x_r)
	rdplot_mean_y   = c(rdplot_mean_y_l,   rdplot_mean_y_r)
	
	
	rdplot_N_l    = as.vector(tapply(y_l, -bin_x_l, length))
	rdplot_N_r    = as.vector(tapply(y_r,  bin_x_r, length))
	rdplot_sd_y_l = as.vector(tapply(y_l, -bin_x_l, sd))
	rdplot_sd_y_r = as.vector(tapply(y_r,  bin_x_r, sd))
	
	rdplot_N = c(rev(rdplot_N_l),rdplot_N_r)
	
	quant = -qt((1-(ci/100))/2,pmax(rdplot_N-1,1))
	
	rdplot_sd_y_l[is.na(rdplot_sd_y_l)] =0
	rdplot_sd_y_r[is.na(rdplot_sd_y_r)] =0
	
	rdplot_sd_y = c(rev(rdplot_sd_y_l),rdplot_sd_y_r)
	rdplot_se_y <- rdplot_sd_y/sqrt(rdplot_N)
	
	rdplot_cil_bin = rdplot_mean_y - quant*rdplot_se_y
	rdplot_cir_bin = rdplot_mean_y + quant*rdplot_se_y
	
	temp_plot =NULL
	
	rdplot_min_bin_l = jumps_l[1:J_star_l]
	rdplot_max_bin_l = jumps_l[2:(J_star_l+1)]
	
	rdplot_min_bin_r = jumps_r[1:J_star_r]
	rdplot_max_bin_r = jumps_r[2:(J_star_r+1)]
	
	bin_length_l = rdplot_max_bin_l - rdplot_min_bin_l
	bin_length_r = rdplot_max_bin_r - rdplot_min_bin_r
	
	bin_avg_l = mean(bin_length_l)
	bin_avg_r = mean(bin_length_r)
	
	bin_med_l = median(bin_length_l)
	bin_med_r = median(bin_length_r)
	
	rdplot_min_bin = c(rdplot_min_bin_l[rev(-rdplot_bin_l)], rdplot_min_bin_r[rdplot_bin_r])
	rdplot_max_bin = c(rdplot_max_bin_l[rev(-rdplot_bin_l)], rdplot_max_bin_r[rdplot_bin_r])
	bin_length     = c(bin_length_l, bin_length_r)
	bin_avg        = c(bin_avg_l, bin_avg_r)
	bin_med        = c(bin_med_l, bin_med_r)
	
	vars_bins = data.frame("rdplot_mean_bin"=rdplot_mean_bin,"rdplot_mean_x"=rdplot_mean_x, "rdplot_mean_y"=rdplot_mean_y, "rdplot_min_bin"=rdplot_min_bin, "rdplot_max_bin"=rdplot_max_bin, "rdplot_se_y"=rdplot_se_y, "rdplot_N"=rdplot_N, "rdplot_ci_l"=rdplot_cil_bin, "rdplot_ci_r"=rdplot_cir_bin)
	vars_poly = data.frame("rdplot_x"= c(x_plot_l, x_plot_r), "rdplot_y"= c(y_hat_l, y_hat_r))

    if (is.null(col.lines)) col.lines = "red"
    if (is.null(col.dots))  col.dots  = "darkblue"
    #if (is.null(type.dots)) type.dots = 20
    if (is.null(title)) title="RD Plot"
    if (is.null(x.label)) x.label="X axis"
    if (is.null(y.label)) y.label="Y axis"
    #if (is.null(x.lim)) x.lim=c(min(x_l),max(x_r))
    #if (is.null(y.lim)) y.lim=c(min(c(y_l,y_r)),max(c(y_l,y_r)))
    #if (is.null(y.lim)) y.lim=c(min(rdplot_mean_y),max(rdplot_mean_y))
    
    data_bins <- data.frame(rdplot_mean_bin, rdplot_mean_y, rdplot_cil_bin, rdplot_cir_bin)
    data_poly <- data.frame(x_plot_l, y_hat_l, x_plot_r, y_hat_r)
    
    temp_plot <- ggplot() + theme_bw() +
        geom_point(data=data_bins, aes(x=rdplot_mean_bin, y=rdplot_mean_y), col=col.dots, na.rm=TRUE) +
        geom_line( data=data_poly, aes(x=x_plot_l, y=y_hat_l), col=col.lines, na.rm=TRUE) +
        geom_line( data=data_poly, aes(x=x_plot_r, y=y_hat_r), col=col.lines, na.rm=TRUE) 

    if (isFALSE(flag_no_ci))
    temp_plot <- temp_plot +
      geom_errorbar(data=data_bins, aes(x=rdplot_mean_bin, ymin=rdplot_cil_bin, ymax=rdplot_cir_bin), linetype = 1) 
      if (isTRUE(shade)){
       temp_plot <- temp_plot +
         geom_ribbon(data=data_bins, aes(x=rdplot_mean_bin, ymin=rdplot_cil_bin, ymax=rdplot_cir_bin))
    }
    
    temp_plot <- temp_plot + labs(x = x.label, y = y.label) + ggtitle(title)+
      coord_cartesian(xlim = x.lim, ylim = y.lim) +
      theme(legend.position = "None") +
      geom_vline(xintercept = c, linewidth = 0.5) 
    
    if (isFALSE(hide)) print(temp_plot)


      
    coef = cbind(gamma_p1_l,gamma_p1_r)
    colnames(coef)=c("Left","Right")
    
    out=list(coef=coef, rdplot=temp_plot, vars_bins=vars_bins, vars_poly=vars_poly,
             J=c(J_star_l,J_star_r), J_IMSE=J_IMSE, J_MV=J_MV, 
             scale=c(scale_l,scale_r), rscale=c(rscale_l,rscale_r),
             bin_avg=c(bin_avg_l,bin_avg_r), bin_med=c(bin_med_l,bin_med_r),
             p=p, c=c, h=c(h_l,h_r), N=c(n_l,n_r), N_h=c(n_h_l,n_h_r), 
             binselect=binselect_type, kernel=kernel_type, coef_covs=gamma_p)
    
    out$call <- match.call()
    class(out) <- "rdplot"
    return(invisible(out))
}

.print_rdplot_header <- function(x) {
  cat("Call: rdplot\n\n")
  cat(paste("Number of Obs.           ",  format(x$N[1]+x$N[2], width=10, justify="right"),"\n", sep=""))
  cat(paste("Kernel                   ",  format(x$kernel,      width=10, justify="right"),"\n", sep=""))
  cat(paste("Bin Method               ",  x$binselect, "\n", sep=""))
  cat("\n")
  cat(paste(format("", width=25), format("Left",  width=10, justify="right"), "   ", format("Right", width=10, justify="right"), "\n", sep=""))
  cat(paste("Number of Obs.           ",  format(x$N[1],    width=10, justify="right"),  "   ", format(x$N[2],     width=10, justify="right"), "\n", sep=""))
  cat(paste("Eff. Number of Obs.      ",  format(x$N_h[1],  width=10, justify="right"),  "   ", format(x$N_h[2],   width=10, justify="right"), "\n", sep=""))
  cat(paste("Order poly. fit (p)      ",  format(x$p,       width=10, justify="right"),  "   ", format(x$p,        width=10, justify="right"), "\n", sep=""))
  cat(paste("BW poly. fit (h)         ",  format(sprintf("%10.3f",x$h[1])),               "   ", format(sprintf("%10.3f",x$h[2])),              "\n", sep=""))
  cat(paste("Number of bins scale     ",  format(sprintf("%10.3f",x$scale[1])),           "   ", format(sprintf("%10.3f",x$scale[2])),           "\n", sep=""))
  invisible()
}

print.rdplot <- function(x, ...) {
  .print_rdplot_header(x)
  cat(paste("Bins Selected            ",  format(x$J[1],    width=10, justify="right"),  "   ", format(x$J[2],     width=10, justify="right"),   "\n", sep=""))
  cat("\n")
  invisible(x)
}

summary.rdplot <- function(object, ...) {
  out <- object
  class(out) <- c("summary.rdplot", class(object))
  out
}

print.summary.rdplot <- function(x, ...) {
  .print_rdplot_header(x)
  cat("\n")
  cat(paste("Bins Selected            ",  format(x$J[1],       width=10, justify="right"),  "   ", format(x$J[2],       width=10, justify="right"),  "\n", sep=""))
  cat(paste("Average Bin Length       ",  format(sprintf("%10.3f",x$bin_avg[1])),  "   ", format(sprintf("%10.3f",x$bin_avg[2])),  "\n", sep=""))
  cat(paste("Median Bin Length        ",  format(sprintf("%10.3f",x$bin_med[1])),  "   ", format(sprintf("%10.3f",x$bin_med[2])),  "\n", sep=""))
  cat("\n")
  cat(paste("IMSE-optimal bins        ",  format(x$J_IMSE[1], width=10, justify="right"),  "   ", format(x$J_IMSE[2], width=10, justify="right"),  "\n", sep=""))
  cat(paste("Mimicking Variance bins  ",  format(x$J_MV[1],   width=10, justify="right"),  "   ", format(x$J_MV[2],   width=10, justify="right"),  "\n", sep=""))
  cat("\n")
  cat(paste("Relative to IMSE-optimal:", "\n", sep=""))
  cat(paste("Implied scale            ",  format(sprintf("%10.3f",x$rscale[1])),                       "   ", format(sprintf("%10.3f",x$rscale[2])),                       "\n", sep=""))
  cat(paste("WIMSE variance weight    ",  format(sprintf("%10.3f",1/(1+x$rscale[1]^3))),               "   ", format(sprintf("%10.3f",1/(1+x$rscale[2]^3))),               "\n", sep=""))
  cat(paste("WIMSE bias weight        ",  format(sprintf("%10.3f",x$rscale[1]^3/(1+x$rscale[1]^3))),  "   ", format(sprintf("%10.3f",x$rscale[2]^3/(1+x$rscale[2]^3))),  "\n", sep=""))
  cat("\n")
  invisible(x)
}

coef.rdplot <- function(object, ...) {
  object$coef
}

glance.rdplot <- function(object, ...) {
  data.frame(
    nobs.left  = object$N[1],
    nobs.right = object$N[2],
    nobs.effective.left  = object$N_h[1],
    nobs.effective.right = object$N_h[2],
    cutoff           = object$c,
    order.regression = object$p,
    bins.left        = object$J[1],
    bins.right       = object$J[2],
    bins.imse.left   = object$J_IMSE[1],
    bins.imse.right  = object$J_IMSE[2],
    bins.mv.left     = object$J_MV[1],
    bins.mv.right    = object$J_MV[2],
    h.left           = object$h[1],
    h.right          = object$h[2],
    kernel           = object$kernel,
    binselect        = object$binselect,
    stringsAsFactors = FALSE
  )
}

---
title: "New Features in rdrobust 4.0.0"
author: "Sebastian Calonico, Matias D. Cattaneo, Max H. Farrell, Rocio Titiunik"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{New Features in rdrobust 4.0.0}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(
  collapse  = TRUE,
  comment   = "#>",
  fig.width = 7,
  fig.height = 4.5
)
```

This vignette documents the features added in **rdrobust 4.0.0**:

1. A `data =` argument for `rdrobust()`, `rdbwselect()`, and `rdplot()`, letting you refer to variables by their bare names inside a data frame.
2. Heteroskedasticity-robust variance estimators (`vce = "hc0"`, `"hc1"`, `"hc2"`, `"hc3"`).
3. Cluster-robust variance estimators (`vce = "cr1"`, `"cr2"`, `"cr3"`) with automatic validation.
4. A `plot` method for `rdrobust` objects (`plot.rdrobust`).
5. A `hide = TRUE` flag and `vars_bins` / `vars_poly` / `coef` outputs from `rdplot()`, so you can build custom RD plots on top of `ggplot2` (or any other graphics system) without re-implementing the binning logic.
6. `tidy` / `glance` methods for integration with the **broom** package.

Existing functionality — bandwidth selection, covariates, kernels, fuzzy and kink designs — is unchanged. See `?rdrobust` and `?rdbwselect` for full documentation.

---

## Setup

We use the U.S. Senate incumbency advantage dataset included in the package. The running variable is the Democratic vote margin at election $t$; the outcome is the Democratic vote share at election $t+2$. The `state` variable serves as a natural cluster identifier, since elections within the same state share common political and institutional features.

```{r data}
library(rdrobust)
data(rdrobust_RDsenate)
df <- rdrobust_RDsenate
head(df)
```

---

## 1. The `data =` argument

Previously, callers either had to `attach()` the data frame or pass each column with `df$` prefixing. In 4.0.0, `rdrobust()`, `rdbwselect()`, and `rdplot()` accept a `data =` argument so `y`, `x`, `covs`, `cluster`, `fuzzy`, `weights`, and `subset` can be given as bare names, resolved against the supplied data frame:

```{r data-arg-basic}
r <- rdrobust(y = vote, x = margin, data = df)
round(r$coef, 3)
```

It also works for expressions built from columns (e.g. `cbind()` for multiple covariates, comparisons for `subset`):

```{r data-arg-expr}
r_cov <- rdrobust(y = vote, x = margin,
                  covs = cbind(class, termshouse, termssenate),
                  data = df)
round(r_cov$coef, 3)

# subset via an expression evaluated inside `data`
r_sub <- rdrobust(y = vote, x = margin,
                  subset = year >= 1950, data = df)
round(r_sub$coef, 3)
```

The same pattern works for `rdbwselect()` and `rdplot()`. For the rest of the vignette we pass variables through `data =`.

### 1.1 `covs =`: formula, names, or matrix

The `covs` argument accepts three forms, so the way you specify covariates can match the shape of your problem:

1. **One-sided formula** — recommended when you have factors, interactions, or transformations:

    ```{r covs-formula-factor}
    df$classf <- factor(df$class, labels = c("A", "B", "C"))

    r_fm <- rdrobust(y = vote, x = margin,
                     covs = ~ classf + termshouse + I(termssenate^2),
                     data = df)
    round(r_fm$coef, 3)
    ```

    Factors are expanded to contrast dummies via `model.matrix()`; interactions (`~ z1 * z2`) and transformations (`I(...)`, `log()`, `poly()`) all work, and the intercept column is dropped automatically. Symbols missing from `data` fall through to the caller's environment.

2. **Character vector of column names** — convenient for programmatic specifications:

    ```{r covs-char}
    covs_names <- c("class", "termshouse", "termssenate")
    r_ch <- rdrobust(y = vote, x = margin, covs = covs_names, data = df)
    round(r_ch$coef, 3)
    ```

    Missing column names give a clear error (`Column(s) not found in 'data': ...`). Passing a character vector without `data = ` is rejected up front instead of failing deep inside the estimator.

3. **Numeric vector, matrix, or data frame** — fully backwards compatible:

    ```{r covs-matrix}
    Z <- with(df, cbind(class, termshouse, termssenate))
    r_mat <- rdrobust(y = vote, x = margin, covs = Z, data = df)
    round(r_mat$coef, 3)
    ```

All three forms — formula with the same plain terms, character vector, and pre-built matrix — yield identical results when the covariates are numeric and the expansion is trivial; the formula form is the right choice when factor expansion or interactions matter.

```{r rdplot, fig.cap="RD plot: U.S. Senate incumbency advantage."}
rdplot(y = vote, x = margin, data = df,
       title   = "Senate: incumbency advantage",
       x.label = "Vote margin (t)",
       y.label = "Vote share (t+2)")
```

---

## 2. Heteroskedasticity-robust variance (HC family)

The full `HC0`–`HC3` family is available via `vce =`. `HC0` is the plug-in sandwich; `HC1` adds a finite-sample scalar; `HC2` and `HC3` reweight residuals by leverage, with `HC3` being the most conservative.

```{r hc-variants}
hc_fits <- lapply(c("nn", "hc0", "hc1", "hc2", "hc3"), function(v) {
  rdrobust(y = vote, x = margin, vce = v, data = df)
})
names(hc_fits) <- c("NN", "HC0", "HC1", "HC2", "HC3")

hc_tbl <- do.call(rbind, lapply(names(hc_fits), function(nm) {
  r <- hc_fits[[nm]]
  data.frame(vce    = nm,
             coef   = round(r$coef[2], 3),   # bias-corrected
             se_rb  = round(r$se[3],   3),   # robust SE
             pval   = round(r$pv[3],   3))
}))
print(hc_tbl, row.names = FALSE)
```

As expected, the point estimate is identical across HC variants (the variance estimator does not change it); only the standard errors and p-values move.

---

## 3. Cluster-Robust Variance Estimators (CR family)

When observations are grouped into clusters, standard errors should account for within-cluster correlation. Version 4.0.0 adds three cluster-robust variance estimators:

| `vce` | Description |
|-------|-------------|
| `"cr1"` | CR1 — cluster sandwich with degrees-of-freedom correction; the default when `cluster` is supplied |
| `"cr2"` | CR2 — Bell & McCaffrey (2002); accounts for within-cluster leverage |
| `"cr3"` | CR3 — Pustejovsky & Tipton (2018); leave-one-cluster-out jackknife; most conservative |

```{r cluster-counts}
cat("Number of clusters:", length(unique(df$state)), "\n")
```

```{r crv-variants}
r_cr1 <- rdrobust(y = vote, x = margin, cluster = state, vce = "cr1", data = df)
r_cr2 <- rdrobust(y = vote, x = margin, cluster = state, vce = "cr2", data = df)
r_cr3 <- rdrobust(y = vote, x = margin, cluster = state, vce = "cr3", data = df)

cl_tbl <- data.frame(
  vce    = c("CR1", "CR2", "CR3"),
  coef   = c(r_cr1$coef[2],  r_cr2$coef[2],  r_cr3$coef[2]),
  se_rb  = c(r_cr1$se[3],    r_cr2$se[3],    r_cr3$se[3]),
  pval   = c(r_cr1$pv[3],    r_cr2$pv[3],    r_cr3$pv[3])
)
cl_tbl[, 2:4] <- round(cl_tbl[, 2:4], 3)
print(cl_tbl, row.names = FALSE)
```

The point estimate (bias-corrected) is the same across estimators — only the standard errors differ. CR3 produces the widest intervals and is the safest default when the number of clusters is moderate.

### Automatic validation

When `cluster` is supplied, only `cr1`/`cr2`/`cr3` are valid; the HC variants and `nn` are silently remapped (or warned and remapped, for HC) to the corresponding cluster-robust estimator:

- `vce = "nn"`   + cluster → `cr1` (silent default)
- `vce = "hc0"`  + cluster → `cr1` (with warning)
- `vce = "hc1"`  + cluster → `cr1` (with warning)
- `vce = "hc2"`  + cluster → `cr2` (with warning)
- `vce = "hc3"`  + cluster → `cr3` (with warning)

Conversely, asking for `cr1/cr2/cr3` without supplying a `cluster` falls back to the corresponding HC variant with a warning.

### Full output for CR3

```{r cr3-summary}
summary(r_cr3)
```

---

## 4. `plot.rdrobust`: Visual Summary of RD Results

Version 4.0.0 adds a `plot` S3 method for `rdrobust` objects. It produces a binscatter with polynomial fits and confidence bands on a common scale.

**Signature:**

```r
plot(x, y, x_run, nbins = 20, ci = TRUE, show_effect = FALSE,
     title = NULL, x.label = "Running Variable", y.label = "Outcome", ...)
```

### Basic plot

```{r plot-basic, fig.cap="Binscatter with local polynomial fits and 95% CI bands."}
plot(r_cr1, df$vote, df$margin,
     title   = "Senate: incumbency advantage",
     x.label = "Vote margin (t)",
     y.label = "Vote share (t+2)")
```

### Adding an effect panel

`show_effect = TRUE` appends a second panel with the conventional point estimate, the robust bias-corrected confidence interval, and significance stars.

```{r plot-effect, fig.height=6, fig.cap="Main plot with RD effect panel."}
plot(r_cr1, df$vote, df$margin,
     show_effect = TRUE,
     title       = "Senate: incumbency advantage",
     x.label     = "Vote margin (t)",
     y.label     = "Vote share (t+2)")
```

### Clustered plot

The method works identically for any `rdrobust` object, regardless of how it was estimated.

```{r plot-cr3, fig.cap="CR3 clustered variance, with effect panel."}
plot(r_cr3, df$vote, df$margin,
     show_effect = TRUE,
     title       = "Senate: CR3 clustered SE",
     x.label     = "Vote margin (t)",
     y.label     = "Vote share (t+2)")
```

---

## 5. Custom RD plots from `rdplot()`

`rdplot()` now accepts a `hide = TRUE` flag and returns the bin-level data used to draw the plot, so you can build any visualization on top of `ggplot2`, `lattice`, or base graphics without re-implementing the binning logic.

The returned list contains:

- `vars_bins` — one row per bin: bin midpoint (`rdplot_mean_bin`), within-bin mean of `y` (`rdplot_mean_y`), within-bin standard error (`rdplot_se_y`), and pointwise CI endpoints (`rdplot_ci_l` / `rdplot_ci_r`) when `ci = ` is set.
- `vars_poly` — a fine evaluation grid (`rdplot_x`) with the fitted polynomial values on each side of the cutoff (`rdplot_y`).
- `coef` — the per-side polynomial coefficient matrix (`Left` / `Right` columns) that produced `vars_poly`.

```{r rdplot-hide}
out <- rdplot(y = vote, x = margin, c = 0,
              hide = TRUE, ci = 95, data = df)
str(out$vars_bins, give.attr = FALSE)
str(out$vars_poly, give.attr = FALSE)
out$coef
```

A minimal custom plot using `ggplot2`:

```{r rdplot-manual, fig.cap="Custom RD plot built from rdplot() output."}
library(ggplot2)
poly_l <- subset(out$vars_poly, rdplot_x <  0)
poly_r <- subset(out$vars_poly, rdplot_x >= 0)

ggplot() +
  geom_vline(xintercept = 0, linewidth = 0.4) +
  geom_errorbar(data = out$vars_bins,
                aes(x = rdplot_mean_bin,
                    ymin = rdplot_ci_l, ymax = rdplot_ci_r),
                color = "grey60", width = 0) +
  geom_point(data = out$vars_bins,
             aes(x = rdplot_mean_bin, y = rdplot_mean_y),
             color = "grey40", size = 1.5) +
  geom_line(data = poly_l, aes(x = rdplot_x, y = rdplot_y)) +
  geom_line(data = poly_r, aes(x = rdplot_x, y = rdplot_y)) +
  theme_bw() +
  labs(x = "Vote margin (t)", y = "Vote share (t+2)",
       title = "Senate: regression function fit")
```

You can also reconstruct the fitted polynomial on a custom x-range directly from `coef`, which is convenient when the default evaluation grid doesn't span the range you want:

```{r rdplot-from-coef}
eval_fit <- function(xs, coef, c)
  sapply(xs, function(xi) sum(coef * (xi - c) ^ (0:(length(coef) - 1))))

xs <- seq(-50, 50, length.out = 5)
data.frame(
  x       = xs,
  y_left  = ifelse(xs <  0, eval_fit(xs, out$coef[, "Left"],  0), NA),
  y_right = ifelse(xs >= 0, eval_fit(xs, out$coef[, "Right"], 0), NA)
)
```

The full set of worked examples (default plot, error-bar variant, shaded CI ribbons, coefficient-based reconstruction) is in `rdplot_illustration.R` at the project root, with Python and Stata twins (`rdplot_illustration.py`, `rdplot_illustration.do`).

---

## 6. broom Integration

**rdrobust** 4.0.0 registers `tidy` and `glance` methods so that results integrate seamlessly with the **broom** package and tidyverse workflows.

### `tidy()`: Coefficient-level output

`tidy()` returns one row per estimator (Conventional, Bias-Corrected, Robust), with standard broom column names:

```{r broom-tidy, eval=requireNamespace("broom", quietly=TRUE)}
library(broom)
tidy(r_cr3)
```

For most applications, the **Robust** row (row 3) is the recommended inference — it combines the bias-corrected point estimate with robust standard errors.

### `glance()`: Model-level summaries

`glance()` returns a single-row data frame with key estimation details: sample sizes, bandwidths, VCE method, and whether covariates or clusters were used.

```{r broom-glance, eval=requireNamespace("broom", quietly=TRUE)}
glance(r_cr3)
```

### Collecting results across specifications

The main payoff of broom integration is easy comparison across specifications. Here we compare all three cluster-robust estimators:

```{r broom-collect, eval=requireNamespace("broom", quietly=TRUE)}
specs <- list(CR1 = r_cr1, CR2 = r_cr2, CR3 = r_cr3)

# Robust BC estimates side-by-side
res <- do.call(rbind, lapply(names(specs), function(nm) {
  cbind(spec = nm, tidy(specs[[nm]])[3, ])  # row 3 = robust
}))
res[, c("spec", "estimate", "std.error", "conf.low", "conf.high")]
```

We can also compare model-level details:

```{r broom-glance-compare, eval=requireNamespace("broom", quietly=TRUE)}
gl <- do.call(rbind, lapply(names(specs), function(nm) {
  cbind(spec = nm, glance(specs[[nm]]))
}))
gl[, c("spec", "nobs.effective.left", "nobs.effective.right",
       "h.left", "vce", "clusters.left", "clusters.right")]
```

### Broader comparison: HC3 vs clustered

A common workflow is comparing heteroskedasticity-robust and cluster-robust results to assess sensitivity to within-cluster correlation:

```{r broom-hc-vs-cl, eval=requireNamespace("broom", quietly=TRUE)}
r_hc3 <- rdrobust(y = vote, x = margin, vce = "hc3", data = df)

all_specs <- list("HC3"  = r_hc3,
                  "CR1"  = r_cr1,
                  "CR2"  = r_cr2,
                  "CR3"  = r_cr3)

comparison <- do.call(rbind, lapply(names(all_specs), function(nm) {
  g <- glance(all_specs[[nm]])
  t <- tidy(all_specs[[nm]])[3, ]   # robust row
  data.frame(spec      = nm,
             estimate  = round(t$estimate, 3),
             std.error = round(t$std.error, 3),
             ci        = paste0("[", round(t$conf.low, 3), ", ", round(t$conf.high, 3), "]"),
             h         = round(g$h.left, 3),
             clusters  = g$clusters.left)
}))
print(comparison, row.names = FALSE)
```

---

## 7. Reporting tables: conventional point + RBC CI

A common workflow is to present a small panel of specifications (sharp, covariate-adjusted, cluster-robust, etc.) in a single table. The canonical Calonico-Cattaneo-Titiunik presentation pairs the **conventional point estimate** with the **robust bias-corrected (RBC) confidence interval** — the point estimate is from row 1 of `coef`, the CI is row 3 of `ci`. The same pattern appears in the Stata illustration (`rdrobust_illustration_new_features.do`), built there with `esttab`; the R analog uses `knitr::kable()`.

We fit four specifications side-by-side: sharp, sharp+covs, cluster, cluster+covs.

```{r table-fits}
covs_form <- ~ class + termshouse + termssenate

m1 <- rdrobust(y = vote, x = margin, data = df)
m2 <- rdrobust(y = vote, x = margin, covs = covs_form, data = df)
m3 <- rdrobust(y = vote, x = margin, vce = "cr1", cluster = state, data = df)
m4 <- rdrobust(y = vote, x = margin, vce = "cr1", cluster = state,
               covs = covs_form, data = df)
```

A small helper extracts the conventional point estimate, the RBC confidence interval, and a few model-level stats from a fitted object:

```{r table-row-builder}
rdr_row <- function(fit, label) {
  data.frame(
    spec      = label,
    estimate  = fit$coef[1, 1],                                 # tau_conv
    rbc_ci    = sprintf("[%.3f, %.3f]", fit$ci[3, 1], fit$ci[3, 2]),
    rbc_pval  = fit$pv[3, 1],
    h         = fit$bws[1, 1],                                  # h_left
    N_eff     = fit$N_h[1] + fit$N_h[2],
    clusters  = if (!is.null(fit$n_clust)) sum(fit$n_clust) else NA_integer_
  )
}

tab <- rbind(
  rdr_row(m1, "sharp"),
  rdr_row(m2, "sharp + covs"),
  rdr_row(m3, "cluster"),
  rdr_row(m4, "cluster + covs")
)
```

Render with `knitr::kable()` for the report:

```{r table-kable}
knitr::kable(
  tab,
  digits   = 3,
  col.names = c("Specification", "Estimate", "Robust 95% CI",
                "RBC p-value", "h", "Eff. N", "Clusters"),
  caption  = "RD estimates: conventional point estimate with RBC confidence interval."
)
```

The Estimate column reports `tau_conv` (the standard local-polynomial point estimate); the **Robust 95% CI** is the bias-corrected, robust-SE confidence interval — the inference recommended by CCT 2014 and the same one printed by `summary(fit)` in the Robust row. The CI is centered on `tau_bc`, not on `tau_conv`, which is why a CI may not be symmetric about the reported point estimate. This is a feature of robust bias-corrected inference, not a typo.

### Specs-as-columns variant (paper style)

For a typeset table closer to the Stata `esttab` layout (specifications as columns, statistics as rows), transpose:

```{r table-paper}
paper <- t(tab[, c("estimate", "rbc_ci", "h", "N_eff")])
colnames(paper) <- tab$spec
rownames(paper) <- c("RD effect (conv.)", "Robust 95% CI", "h", "Eff. N")

knitr::kable(
  paper,
  caption = "Senate incumbency: RD estimates across specifications."
)
```

### LaTeX export

To write a publication-ready LaTeX table to a file (the R analog of the Stata `esttab ... using rdrobust_table.tex`):

```{r table-latex, eval=FALSE}
tex <- knitr::kable(
  paper,
  format    = "latex",
  booktabs  = TRUE,
  caption   = "RD estimates for Senate incumbency.",
  label     = "tab:rdrobust"
)
writeLines(tex, "rdrobust_table.tex")
```

For richer formatting (multi-row headers, footnotes, grouping), the `kableExtra` and `gt` packages build on top of `kable()` / `gt::gt()` and accept the same input data frame.

---

## Session Info

```{r session}
sessionInfo()
```

\name{plot.rdrobust}
\alias{plot.rdrobust}
\title{Plot Method for rdrobust Objects}
\description{
Produces a visual summary of an \code{\link{rdrobust}} result. The main panel
shows a binned scatter plot of the outcome against the running variable within
the estimated bandwidth, overlaid with the local-polynomial fit curves and
(optionally) pointwise confidence bands. Bin sizes are scaled to the average
kernel weight of the observations they contain, so bins closest to the cutoff
appear largest. An optional effect panel below the main plot displays the
conventional point estimate with its robust bias-corrected confidence interval
and significance stars.
}
\usage{
\method{plot}{rdrobust}(x, y, x_run, nbins = 20,
     ci = TRUE, show_effect = FALSE,
     title = NULL,
     x.label = "Running Variable",
     y.label = "Outcome",
     x.lim = NULL, y.lim = NULL,
     col.l = "#3B7DD8", col.r = "#D95F3B",
     base_size = 14, \dots)
}
\arguments{
  \item{x}{an object of class \code{"rdrobust"}, as returned by \code{\link{rdrobust}}.}
  \item{y}{numeric vector of outcome values (same length and order as used in the original \code{rdrobust} call, before any \code{subset} was applied).}
  \item{x_run}{numeric vector of running variable values (same length and order as \code{y}).}
  \item{nbins}{number of bins per side used for the binned scatter plot. Default is \code{20}.}
  \item{ci}{logical. If \code{TRUE} (default), shaded pointwise confidence bands are drawn around each polynomial fit curve. The bands use the conventional variance-covariance matrices \code{V_cl_l} and \code{V_cl_r} stored in the \code{rdrobust} object, evaluated at the confidence level stored in the \code{rdrobust} object.}
  \item{show_effect}{logical. If \code{TRUE}, a second panel is printed below the main plot showing the conventional point estimate with its robust bias-corrected (RBC) confidence interval as a horizontal error bar, together with significance stars based on the robust p-value. Combining both panels requires the \pkg{gridExtra} package; if unavailable the two plots are printed sequentially. Default is \code{FALSE}.}
  \item{title}{character string for the plot title. Default is \code{NULL} (no title).}
  \item{x.label}{label for the horizontal axis. Default is \code{"Running Variable"}.}
  \item{y.label}{label for the vertical axis. Default is \code{"Outcome"}.}
  \item{x.lim}{numeric vector of length 2 specifying the horizontal axis limits. Default is \code{NULL} (automatic).}
  \item{y.lim}{numeric vector of length 2 specifying the vertical axis limits. Default is \code{NULL} (automatic).}
  \item{col.l}{color for the left-side (control) elements. Default is \code{"#3B7DD8"}.}
  \item{col.r}{color for the right-side (treated) elements. Default is \code{"#D95F3B"}.}
  \item{base_size}{base font size in points passed to the underlying \pkg{ggplot2} theme. All text elements scale relative to this value. Default is \code{14}.}
  \item{\dots}{additional arguments (currently unused).}
}
\details{
Only observations within the main estimation bandwidth \code{[c - h_l, c + h_r]}
are shown. Each side is divided into \code{nbins} equal-width bins; the plotted
point for each bin is the bin mean of both \code{x_run} and \code{y}. The size
of each bin's dot is proportional to the average kernel weight of the
observations it contains, computed from the kernel stored in
\code{x$kernel} (triangular, Epanechnikov, or uniform) evaluated at the
normalized distance \code{(x_i - c) / h}.

The polynomial fit curve on each side is the fitted conditional mean implied by
the local-polynomial coefficients \code{beta_Y_p_l} and \code{beta_Y_p_r}. The
pointwise confidence bands (\code{ci = TRUE}) are computed as
\deqn{\hat{\mu}(t) \pm z_{\alpha/2} \sqrt{\mathbf{x}(t)^\top V \mathbf{x}(t)}}
where \eqn{\mathbf{x}(t) = (1, t-c, \ldots, (t-c)^p)^\top} and \eqn{V} is
the conventional variance-covariance matrix of the polynomial coefficients.

When \code{show_effect = TRUE} the effect panel reports:
\itemize{
  \item Point estimate: conventional (\code{coef["Conventional"]}).
  \item Interval: robust bias-corrected (\code{ci["Robust",]}).
  \item Stars: based on \code{pv["Robust",]}: \code{*} p<0.10, \code{**} p<0.05, \code{***} p<0.01.
}
}
\value{
Invisibly returns the \pkg{ggplot2} plot object when \code{show_effect = FALSE},
or a named list \code{list(rd_plot, effect_plot)} of two \pkg{ggplot2} objects
when \code{show_effect = TRUE}. In both cases the plot(s) are printed as a
side-effect.
}
\references{
Calonico, S., M. D. Cattaneo, and R. Titiunik. 2014.
\href{https://rdpackages.github.io/references/Calonico-Cattaneo-Titiunik_2014_ECMA.pdf}{Robust Nonparametric Confidence Intervals for Regression-Discontinuity Designs.}
\emph{Econometrica} 82(6): 2295-2326.

Calonico, S., M. D. Cattaneo, M. H. Farrell, and R. Titiunik. 2019.
\href{https://rdpackages.github.io/references/Calonico-Cattaneo-Farrell-Titiunik_2019_RESTAT.pdf}{Regression Discontinuity Designs using Covariates.}
\emph{Review of Economics and Statistics}, 101(3): 442-451.
}
\author{
Sebastian Calonico, University of California, Davis, CA.  \email{scalonico@ucdavis.edu}.

Matias D. Cattaneo, Princeton University, Princeton, NJ.  \email{matias.d.cattaneo@gmail.com}.

Max H. Farrell, University of California, Santa Barbara, CA. \email{mhfarrell@gmail.com}.

Rocio Titiunik, Princeton University, Princeton, NJ.  \email{rocio.titiunik@gmail.com}.
}
\examples{
x <- runif(500, -1, 1)
y <- 5 + 3*x + 2*(x >= 0) + rnorm(500)
r <- rdrobust(y, x)

## Main RD plot with CI bands
plot(r, y, x)

## Add effect panel
plot(r, y, x, show_effect = TRUE)
}
\seealso{
  \code{\link{rdrobust}}, \code{\link{rdbwselect}}, \code{\link{rdplot}}
}
\keyword{RDD}
\keyword{Plots}

\name{rdrobust-package}
\alias{rdrobust-package}
\docType{package}
\title{Robust Data-Driven Statistical Inference in RD Designs}
\description{
   Regression-discontinuity (RD) designs are quasi-experimental research designs popular in social, behavioral and natural sciences. The RD design is usually employed to study the (local) causal effect of a treatment, intervention or policy. This package provides tools for data-driven graphical and analytical statistical inference in RD designs: \code{\link{rdrobust}} to construct local-polynomial point estimators and robust confidence intervals for average treatment effects at the cutoff in Sharp, Fuzzy and Kink RD settings, \code{\link{rdbwselect}} to perform bandwidth selection for the different procedures implemented, and \code{\link{rdplot}} to conduct exploratory data analysis (RD plots).
}
\details{
\tabular{ll}{
Package: \tab rdrobust\cr
Type: \tab Package\cr
Version: \tab 4.0.0\cr
Date: \tab 2026-05-15\cr
License: \tab GPL-3\cr
}
Function for statistical inference: \code{\link{rdrobust}}\cr
Function for bandwidths selection: \code{\link{rdbwselect}}\cr
Function for exploratory data analysis (RD plots): \code{\link{rdplot}}
}
\author{
Sebastian Calonico, University of California, Davis, CA.  \email{scalonico@ucdavis.edu}.

Matias D. Cattaneo, Princeton University, Princeton, NJ.  \email{matias.d.cattaneo@gmail.com}.

Max H. Farrell, University of California, Santa Barbara, CA. \email{mhfarrell@gmail.com}.

Rocio Titiunik, Princeton University, Princeton, NJ.  \email{rocio.titiunik@gmail.com}.
}

\name{rdrobust_RDsenate}
\alias{rdrobust_RDsenate}
\docType{data}
\title{RD Senate Data}
\description{
Extract of the dataset constructed by Cattaneo, Frandsen, and Titiunik (2015), which includes measures of incumbency advantage in the U.S. Senate for the period 1914-2010. The dataset contains the running variable (Democratic vote margin at election \eqn{t}), the main outcome (Democratic vote share at election \eqn{t+2}), additional outcomes and lagged vote shares useful as covariates, and a state identifier suitable for cluster-robust inference.
}
\usage{data(rdrobust_RDsenate)}
\format{
  A data frame with 1390 observations on the following 17 variables.
  \describe{
    \item{\code{state}}{character. U.S. state name; can be used as a cluster variable for cluster-robust variance estimation.}
    \item{\code{year}}{numeric. Election year.}
    \item{\code{margin}}{numeric. Democratic vote margin at election \eqn{t} (running variable); equals \code{demmv} in the original dataset.}
    \item{\code{vote}}{numeric. Democratic vote share at election \eqn{t+2} (main outcome); equals \code{demvoteshfor2} in the original dataset.}
    \item{\code{class}}{numeric. U.S. Senate class (1, 2, or 3).}
    \item{\code{termshouse}}{numeric. Number of terms served in the U.S. House.}
    \item{\code{termssenate}}{numeric. Number of terms served in the U.S. Senate.}
    \item{\code{dopen}}{numeric. Indicator for open-seat election.}
    \item{\code{population}}{numeric. State population.}
    \item{\code{presdemvoteshlag1}}{numeric. Lagged Democratic presidential vote share.}
    \item{\code{demvoteshlag1}}{numeric. Democratic vote share at election \eqn{t} (same cycle as running variable); useful as a covariate.}
    \item{\code{demvoteshlag2}}{numeric. Democratic vote share at election \eqn{t-2}; useful as a covariate.}
    \item{\code{demvoteshfor1}}{numeric. Democratic vote share at election \eqn{t+1}.}
    \item{\code{demwinprv1}}{numeric. Indicator for Democratic win at election \eqn{t-1}.}
    \item{\code{demwinprv2}}{numeric. Indicator for Democratic win at election \eqn{t-2}.}
    \item{\code{dmidterm}}{numeric. Indicator for midterm election year.}
    \item{\code{dpresdem}}{numeric. Indicator for Democratic president in office.}
  }
}
\source{
    Cattaneo, M. D., Frandsen, B., and R. Titiunik. 2015. \href{https://rdpackages.github.io/references/Cattaneo-Frandsen-Titiunik_2015_JCI.pdf}{Randomization Inference in the Regression Discontinuity Design: An Application to the Study of Party Advantages in the U.S. Senate.} \emph{Journal of Causal Inference} 3(1): 1-24.
}
\references{
    Cattaneo, M. D., Frandsen, B., and R. Titiunik. 2015. \href{https://rdpackages.github.io/references/Cattaneo-Frandsen-Titiunik_2015_JCI.pdf}{Randomization Inference in the Regression Discontinuity Design: An Application to the Study of Party Advantages in the U.S. Senate.} \emph{Journal of Causal Inference} 3(1): 1-24.
}
\keyword{datasets}

\name{rdrobust}
\alias{rdrobust} 
\alias{print.rdrobust}
\alias{summary.rdrobust}
\title{Local-Polynomial RD Estimation with Robust Confidence Intervals}
\description{

\code{\link{rdrobust}} implements local polynomial Regression Discontinuity (RD) point estimators with robust bias-corrected confidence intervals and inference procedures developed in Calonico, Cattaneo and Titiunik (2014a), Calonico, Cattaneo and Farrell (2018), Calonico, Cattaneo, Farrell and Titiunik (2019), and Calonico, Cattaneo and Farrell (2020). It also computes alternative estimation and inference procedures available in the literature. 

Companion commands are: \code{\link{rdbwselect}} for data-driven bandwidth selection, and \code{\link{rdplot}} for data-driven RD plots (see Calonico, Cattaneo and Titiunik (2015a) for details).

A detailed introduction to this command is given in Calonico, Cattaneo and Titiunik (2015b), and Calonico, Cattaneo, Farrell and Titiunik (2017). A companion \code{Stata} package is described in Calonico, Cattaneo and Titiunik (2014b).

For more details, and related Stata and R packages useful for analysis of RD designs, visit \url{https://rdpackages.github.io/}

}
\usage{rdrobust(y, x, c = NULL,  fuzzy = NULL, 
          deriv = NULL, p = NULL, q = NULL, 
          h = NULL, b = NULL, rho = NULL, 
          covs = NULL,  covs_drop = TRUE, ginv.tol = 1e-20,
          kernel = "tri", weights = NULL, bwselect = "mserd", 
          vce = "nn", cluster = NULL, 
          nnmatch = 3, level = 95, scalepar = 1, scaleregul = 1, 
          sharpbw = FALSE, subset = NULL,
          masspoints = "adjust", bwcheck = NULL,
          bwrestrict = TRUE, stdvars = FALSE, data = NULL)}
\arguments{
  \item{y}{is the dependent variable.}
  \item{x}{is the running variable (a.k.a. score or forcing variable).}
  \item{c}{specifies the RD cutoff in \code{x}; default is \code{c = 0}.}
  \item{fuzzy}{specifies the treatment status variable used to implement fuzzy RD estimation (or Fuzzy Kink RD if \code{deriv=1} is also specified). Default is Sharp RD design and hence this option is not used.} 
  \item{deriv}{specifies the order of the derivative of the regression functions to be estimated. Default is \code{deriv=0} (for Sharp RD, or for Fuzzy RD if \code{fuzzy} is also specified). Setting \code{deriv=1} results in estimation of a Kink RD design (up to scale), or Fuzzy Kink RD if \code{fuzzy} is also specified.}
  \item{p}{specifies the order of the local-polynomial used to construct the point-estimator; default is \code{p = 1} (local linear regression).}
  \item{q}{specifies the order of the local-polynomial used to construct the bias-correction; default is \code{q = 2} (local quadratic regression).}
  \item{h}{specifies the main bandwidth used to construct the RD point estimator. If not specified, bandwidth \code{h} is computed by the companion command \code{\link{rdbwselect}}. If two bandwidths are specified, the first bandwidth is used for the data below the cutoff and the second bandwidth is used for the data above the cutoff.}
  \item{b}{specifies the bias bandwidth used to construct the bias-correction estimator. If not specified, bandwidth \code{b} is computed by the companion command \code{\link{rdbwselect}}. If two bandwidths are specified, the first bandwidth is used for the data below the cutoff and the second bandwidth is used for the data above the cutoff.}
  \item{rho}{specifies the value of \code{rho}, so that the bias bandwidth \code{b} equals \code{h/rho}. Default is \code{rho = 1} if \code{h} is specified but \code{b} is not.}
  \item{covs}{additional covariates to be used for efficiency improvements. One of:
  \itemize{
    \item a one-sided \code{\link{formula}}, e.g. \code{~ z1 + z2 + factor(g) + I(z3^2)}: processed with \code{\link{model.matrix}}, so factors are expanded to contrasts, interactions are supported, and transformations such as \code{I(\dots)}, \code{log()}, and \code{poly()} work. The intercept column is dropped automatically. Symbols are looked up in \code{data} first (if supplied), then in the caller's environment;
    \item a character vector of column names (requires \code{data =}), e.g. \code{c("z1", "z2")}: selected as \code{data[, covs]} and coerced to a matrix. Useful for programmatic specifications;
    \item a numeric vector, matrix, or data frame: passed through unchanged (backwards compatible).
  }}
  \item{covs_drop}{if TRUE, it checks for collinear additional covariates and drops them. Default is TRUE.}
  \item{ginv.tol}{tolerance used to invert matrices involving covariates when \code{covs_drop=TRUE}.}
  \item{kernel}{is the kernel function used to construct the local-polynomial estimator(s). Options are \code{triangular} (default option), \code{epanechnikov} and \code{uniform}.}
  \item{weights}{is the variable used for optional weighting of the estimation procedure. The unit-specific weights multiply the kernel function.}

\item{bwselect}{specifies the bandwidth selection procedure to be used. By default it computes both \code{h} and \code{b}, unless \code{rho} is specified, in which case it only computes \code{h} and sets \code{b=h/rho}.}
Options are:

\code{mserd} one common MSE-optimal bandwidth selector for the RD treatment effect estimator.

\code{msetwo} two different MSE-optimal bandwidth selectors (below and above the cutoff) for the RD treatment effect estimator.

\code{msesum} one common MSE-optimal bandwidth selector for the sum of regression estimates (as opposed to difference thereof).

\code{msecomb1} for min(\code{mserd},\code{msesum}).

\code{msecomb2} for median(\code{msetwo},\code{mserd},\code{msesum}), for each side of the cutoff separately.

\code{cerrd} one common CER-optimal bandwidth selector for the RD treatment effect estimator.

\code{certwo} two different CER-optimal bandwidth selectors (below and above the cutoff) for the RD treatment effect estimator.

\code{cersum} one common CER-optimal bandwidth selector for the sum of regression estimates (as opposed to difference thereof).

\code{cercomb1} for min(\code{cerrd},\code{cersum}).

\code{cercomb2} for median(\code{certwo},\code{cerrd},\code{cersum}), for each side of the cutoff separately.

Note: MSE = Mean Square Error; CER = Coverage Error Rate.
Default is \code{bwselect=mserd}. For details on implementation see Calonico, Cattaneo and Titiunik (2014a), Calonico, Cattaneo and Farrell (2018), and Calonico, Cattaneo, Farrell and Titiunik (2019), and the companion software articles.

\item{vce}{specifies the procedure used to compute the variance-covariance matrix estimator.
Options are:

\code{nn} for heteroskedasticity-robust nearest neighbor variance estimator with \code{nnmatch} the (minimum) number of neighbors to be used.

\code{hc0} for heteroskedasticity-robust plug-in residuals variance estimator without weights.

\code{hc1} for heteroskedasticity-robust plug-in residuals variance estimator with \code{hc1} weights.

\code{hc2} for heteroskedasticity-robust plug-in residuals variance estimator with \code{hc2} weights.

\code{hc3} for heteroskedasticity-robust plug-in residuals variance estimator with \code{hc3} weights.

\code{cr1} for CR1 cluster-robust variance estimator with degrees-of-freedom correction at the cluster level. Requires \code{cluster} to be specified.

\code{cr2} for CR2 cluster-robust variance estimator with leverage adjustments (Bell-McCaffrey). Requires \code{cluster} to be specified.

\code{cr3} for CR3 cluster-robust variance estimator with leave-one-cluster-out jackknife (Pustejovsky-Tipton); most conservative. Requires \code{cluster} to be specified.

Default is \code{vce=nn}. When \code{cluster} is specified, only \code{vce=cr1}, \code{vce=cr2}, and \code{vce=cr3} are valid; other options are automatically mapped to the corresponding cluster variant with a warning.

The CR2/CR3 leverage correction applies to both the conventional and the robust bias-corrected standard errors, including when the point-estimation bandwidth \code{h} differs from the bias-correction bandwidth \code{b}; in that case the cluster leverage is computed from the bias (\code{b}) regression.}
  \item{cluster}{indicates the cluster ID variable used for cluster-robust variance estimation. Must be combined with \code{vce=cr1}, \code{vce=cr2}, or \code{vce=cr3}.}
  \item{nnmatch}{to be combined with for \code{vce=nn} for heteroskedasticity-robust nearest neighbor variance estimator with \code{nnmatch} indicating the minimum number of neighbors to be used. Default is \code{nnmatch=3}}
  \item{level}{sets the confidence level for confidence intervals; default is \code{level = 95}.}
  \item{scalepar}{specifies scaling factor for RD parameter of interest. This option is useful when the population parameter of interest involves a known multiplicative factor (e.g., sharp kink RD).  Default is \code{scalepar = 1} (no scaling).}
  \item{scaleregul}{specifies scaling factor for the regularization term added to the denominator of the bandwidth selectors. Setting \code{scaleregul = 0} removes the regularization term from the bandwidth selectors; default is \code{scaleregul = 1}.}
  \item{sharpbw}{option to perform fuzzy RD estimation using a bandwidth selection procedure for the sharp RD model. This option is automatically selected if there is perfect compliance at either side of the cutoff.}
  \item{subset}{an optional vector specifying a subset of observations to be used.}
  \item{masspoints}{checks and controls for repeated observations in the running variable. Options are:
  
    (i)  \code{off}: ignores the presence of mass points; 
    
    (ii) \code{check}: looks for and reports the number of unique observations at each side of the cutoff. 
    
    (iii) \code{adjust}: controls that the preliminary bandwidths used in the calculations contain a minimal number of unique observations. By default it uses 10 observations, but it can be manually adjusted with the option \code{bwcheck}. 
    
    Default option is \code{masspoints=adjust}.}
    \item{bwcheck}{if a positive integer is provided, the preliminary bandwidth used in the calculations is enlarged so that at least \code{bwcheck} unique observations are used.}
    
    \item{bwrestrict}{if \code{TRUE}, computed bandwidths are restricted to lie within the range of \code{x}; default is \code{bwrestrict = TRUE}.}
      
    \item{stdvars}{if \code{TRUE}, \code{x} and \code{y} are standardized before computing the bandwidths; default is \code{stdvars = FALSE}.}
  \item{data}{an optional data frame. When supplied, \code{y}, \code{x}, \code{covs}, \code{cluster}, \code{fuzzy}, \code{weights}, and \code{subset} may be given as bare variable names referring to columns of \code{data}.}
      
}
\value{
  \item{N}{vector with the sample sizes used to the left and to the right of the cutoff.}
  \item{N_h}{vector with the effective sample sizes used to the left and to the right of the cutoff.}
  \item{N_b}{vector with the effective sample sizes used to the left and to the right of the cutoff for bias estimation.}
  \item{M}{vector with the number of unique observations to the left and to the right of the cutoff (when \code{masspoints} is not \code{off}).}
  \item{c}{cutoff value.}
  \item{p}{order of the polynomial used for estimation of the regression function.}
  \item{q}{order of the polynomial used for estimation of the bias of the regression function.}
  \item{bws}{matrix containing the bandwidths used: row \code{h} for the estimation bandwidth, row \code{b} for the bias bandwidth; columns \code{left} and \code{right}.}
  \item{tau_cl}{conventional local-polynomial estimate to the left and to the right of the cutoff.}
  \item{tau_bc}{bias-corrected local-polynomial estimate to the left and to the right of the cutoff.}
  \item{coef}{vector containing conventional and bias-corrected local-polynomial RD estimates.}
  \item{se}{vector containing conventional and robust standard errors of the local-polynomial RD estimates.}
  \item{bias}{estimated bias for the local-polynomial RD estimator below and above the cutoff.}
  \item{beta_Y_p_l}{conventional p-order local-polynomial estimates to the left of the cutoff for the outcome variable.}
  \item{beta_Y_p_r}{conventional p-order local-polynomial estimates to the right of the cutoff for the outcome variable.}
  \item{beta_T_p_l}{conventional p-order local-polynomial estimates to the left of the cutoff for the first stage (fuzzy RD).}
  \item{beta_T_p_r}{conventional p-order local-polynomial estimates to the right of the cutoff for the first stage (fuzzy RD).}
  \item{coef_covs}{coefficients of the additional covariates, only returned when \code{covs} is specified.}
  \item{V_cl_l}{conventional variance-covariance matrix estimated below the cutoff.}
  \item{V_cl_r}{conventional variance-covariance matrix estimated above the cutoff.}
  \item{V_rb_l}{robust variance-covariance matrix estimated below the cutoff.}
  \item{V_rb_r}{robust variance-covariance matrix estimated above the cutoff.}
  \item{z}{vector containing the z-statistics associated with conventional, bias-corrected and robust local-polynomial RD estimates.}
  \item{pv}{vector containing the p-values associated with conventional, bias-corrected and robust local-polynomial RD estimates.}
  \item{ci}{matrix containing the confidence intervals associated with conventional, bias-corrected and robust local-polynomial RD estimates.}
  \item{kernel}{kernel function used.}
  \item{vce}{variance estimation method used.}
  \item{bwselect}{bandwidth selection method used.}
  \item{level}{confidence level used.}
  \item{masspoints}{mass points option used.}
  \item{rdmodel}{character string describing the model estimated: design type (sharp/fuzzy/kink), whether covariates were included, and whether standard errors are clustered.}
  \item{n_clust}{vector with the number of clusters to the left and to the right of the cutoff. \code{NULL} when \code{cluster} is not specified.}
  \item{tau_T, se_T, z_T, pv_T, ci_T}{first-stage estimates, standard errors, z-statistics, p-values, and confidence intervals for the treatment indicator (fuzzy RD only).}
}
\references{
Calonico, S., M. D. Cattaneo, and M. H. Farrell. 2018. \href{https://rdpackages.github.io/references/Calonico-Cattaneo-Farrell_2018_JASA.pdf}{On the Effect of Bias Estimation on Coverage Accuracy in Nonparametric Inference.} \emph{Journal of the American Statistical Association}, 113(522): 767-779.

Calonico, S., M. D. Cattaneo, and M. H. Farrell. 2020. \href{https://rdpackages.github.io/references/Calonico-Cattaneo-Farrell_2020_ECTJ.pdf}{Optimal Bandwidth Choice for Robust Bias Corrected Inference in Regression Discontinuity Designs.} \emph{Econometrics Journal}, 23(2): 192-210.

Calonico, S., M. D. Cattaneo, M. H. Farrell, and R. Titiunik. 2017. \href{https://rdpackages.github.io/references/Calonico-Cattaneo-Farrell-Titiunik_2017_Stata.pdf}{rdrobust: Software for Regression Discontinuity Designs.} \emph{Stata Journal}, 17(2): 372-404.

Calonico, S., M. D. Cattaneo, M. H. Farrell, and R. Titiunik. 2019. \href{https://rdpackages.github.io/references/Calonico-Cattaneo-Farrell-Titiunik_2019_RESTAT.pdf}{Regression Discontinuity Designs using Covariates.} \emph{Review of Economics and Statistics}, 101(3): 442-451. 

Calonico, S., M. D. Cattaneo, and R. Titiunik. 2014a. \href{https://rdpackages.github.io/references/Calonico-Cattaneo-Titiunik_2014_ECMA.pdf}{Robust Nonparametric Confidence Intervals for Regression-Discontinuity Designs.} \emph{Econometrica} 82(6): 2295-2326.

Calonico, S., M. D. Cattaneo, and R. Titiunik. 2014b. \href{https://rdpackages.github.io/references/Calonico-Cattaneo-Titiunik_2014_Stata.pdf}{Robust Data-Driven Inference in the Regression-Discontinuity Design.} \emph{Stata Journal} 14(4): 909-946.

Calonico, S., M. D. Cattaneo, and R. Titiunik. 2015a. \href{https://rdpackages.github.io/references/Calonico-Cattaneo-Titiunik_2015_JASA.pdf}{Optimal Data-Driven Regression Discontinuity Plots.} \emph{Journal of the American Statistical Association} 110(512): 1753-1769.

Calonico, S., M. D. Cattaneo, and R. Titiunik. 2015b. \href{https://rdpackages.github.io/references/Calonico-Cattaneo-Titiunik_2015_R.pdf}{rdrobust: An R Package for Robust Nonparametric Inference in Regression-Discontinuity Designs.} \emph{R Journal} 7(1): 38-51.

Cattaneo, M. D., B. Frandsen, and R. Titiunik. 2015. \href{https://rdpackages.github.io/references/Cattaneo-Frandsen-Titiunik_2015_JCI.pdf}{Randomization Inference in the Regression Discontinuity Design: An Application to the Study of Party Advantages in the U.S. Senate.} \emph{Journal of Causal Inference} 3(1): 1-24.
}
\author{
Sebastian Calonico, University of California, Davis, CA.  \email{scalonico@ucdavis.edu}.

Matias D. Cattaneo, Princeton University, Princeton, NJ.  \email{matias.d.cattaneo@gmail.com}.

Max H. Farrell, University of California, Santa Barbara, CA. \email{mhfarrell@gmail.com}.

Rocio Titiunik, Princeton University, Princeton, NJ.  \email{rocio.titiunik@gmail.com}.
}
\examples{
x <- runif(1000,-1,1)
y <- 5 + 3*x + 2*(x>=0) + rnorm(1000)
rdrobust(y,x)

# Using a data frame with covariates via the `data =` argument
df <- data.frame(y = y, x = x, z1 = rnorm(1000), z2 = rnorm(1000))
rdrobust(y, x, covs = ~ z1 + z2, data = df)
}
\keyword{ RDD }
\keyword{ Robust Estimation }
\seealso{
  \code{\link{rdbwselect}}, \code{\link{rdplot}}, \code{\link{plot.rdrobust}}
}

\name{rdplot}
\alias{rdplot}
\alias{print.rdplot}
\alias{summary.rdplot}
\title{Data-Driven Regression Discontinuity Plots}
\description{

\code{\link{rdplot}}  implements several data-driven Regression Discontinuity (RD) plots, using either evenly-spaced or quantile-spaced partitioning. Two types of RD plots are constructed: (i) RD plots with binned sample means tracing out the underlying regression function, and (ii) RD plots with binned sample means mimicking the underlying variability of the data. For technical and methodological details see Calonico, Cattaneo and Titiunik (2015a).

Companion commands are: \code{\link{rdrobust}} for point estimation and inference procedures, and \code{\link{rdbwselect}} for data-driven bandwidth selection.

A detailed introduction to this command is given in Calonico, Cattaneo and Titiunik (2015b) and Calonico, Cattaneo, Farrell and Titiunik (2017). A companion \code{Stata} package is described in Calonico, Cattaneo and Titiunik (2014).

For more details, and related Stata and R packages useful for analysis of RD designs, visit \url{https://rdpackages.github.io/}
}
\usage{
rdplot(y, x, c = 0, p = 4, nbins = NULL, binselect = "esmv",  
          scale = NULL, kernel = "uni", weights = NULL, h = NULL, 
          covs = NULL, covs_eval = "mean", covs_drop = TRUE, ginv.tol = 1e-20,
          support = NULL, subset = NULL, masspoints = "adjust",
          hide = FALSE, ci = NULL, shade = FALSE, title = NULL, 
          x.label = NULL, y.label = NULL, x.lim = NULL, y.lim = NULL, 
          col.dots = NULL, col.lines = NULL, data = NULL)
}
\arguments{
  \item{y}{is the dependent variable.}
  \item{x}{is the running variable (a.k.a. score or forcing variable).}
  \item{c}{specifies the RD cutoff in \code{x}; default is \code{c = 0}.}
  \item{p}{specifies the order of the global-polynomial used to approximate the population conditional mean functions for control and treated units; default is \code{p = 4}.} 
  \item{nbins}{specifies the number of bins used to the left of the cutoff, denoted \eqn{J_-}, and to the right of the cutoff, denoted \eqn{J_+}, respectively. If not specified, \eqn{J_+} and \eqn{J_-} are estimated using the method and options chosen below.}
  \item{binselect}{specifies the procedure to select the number of bins. This option is available only if \eqn{J_-} and \eqn{J_+} are not set manually. Options are:  

        \code{es}: IMSE-optimal evenly-spaced method using spacings estimators.
        
        \code{espr}: IMSE-optimal evenly-spaced method using polynomial regression.
        
        \code{esmv}: mimicking variance evenly-spaced method using spacings estimators. This is the default option.
        
        \code{esmvpr}: mimicking variance evenly-spaced method using polynomial regression.
        
        \code{qs}: IMSE-optimal quantile-spaced method using spacings estimators.
        
        \code{qspr}: IMSE-optimal quantile-spaced method using polynomial regression.
        
        \code{qsmv}: mimicking variance quantile-spaced method using spacings estimators.
        
        \code{qsmvpr}: mimicking variance quantile-spaced method using polynomial regression.
     }
       \item{scale}{specifies a multiplicative factor to be used with the optimal numbers of bins selected. Specifically, the number of bins used for the treatment and control groups will be \code{scale}\eqn{\times \hat{J}_+} and \code{scale}\eqn{\times \hat{J}_-}, where \eqn{\hat{J}_\cdot} denotes the estimated optimal numbers of bins originally computed for each group; default is \code{scale = 1}.}
  \item{kernel}{specifies the kernel function used to construct the local-polynomial estimator(s). Options are: \code{triangular}, \code{epanechnikov}, and \code{uniform}. Default is \code{kernel=uniform} (i.e., equal/no weighting to all observations on the support of the kernel).}
  \item{weights}{is the variable used for optional weighting of the estimation procedure. The unit-specific weights multiply the kernel function.}
  \item{h}{specifies the bandwidth used to construct the (global) polynomial fits given the kernel choice \code{kernel}.
If not specified, the bandwidths are chosen to span the full support of the data. If two bandwidths are specified, the first bandwidth is used for the data below the cutoff and the second bandwidth is used for the data above the cutoff.}
  \item{covs}{specifies additional covariates to be used in the polynomial regression. One of:
  \itemize{
    \item a one-sided \code{\link{formula}}, e.g. \code{~ z1 + z2 + factor(g) + I(z3^2)}: processed with \code{\link{model.matrix}}, so factors are expanded to contrasts, interactions are supported, and transformations such as \code{I(\dots)}, \code{log()}, and \code{poly()} work. The intercept column is dropped automatically. Symbols are looked up in \code{data} first (if supplied), then in the caller's environment;
    \item a character vector of column names (requires \code{data =}), e.g. \code{c("z1", "z2")}: selected as \code{data[, covs]} and coerced to a matrix. Useful for programmatic specifications;
    \item a numeric vector, matrix, or data frame: passed through unchanged (backwards compatible).
  }}
  \item{covs_eval}{sets the evaluation points for the additional covariates, when included in the estimation. Options are: \code{covs_eval = 0} and \code{covs_eval = "mean"} (default)}
  \item{covs_drop}{if TRUE, it checks for collinear additional covariates and drops them. Default is TRUE.}
    \item{ginv.tol}{tolerance used to invert matrices involving covariates when \code{covs_drop=TRUE}.}
  \item{support}{specifies an optional extended support of the running variable to be used in the construction of the bins;  default is the sample range.}
  \item{subset}{an optional vector specifying a subset of observations to be used.}
    \item{masspoints}{checks and controls for repeated observations in the running variable. Options are:
  
    (i)  \code{off}: ignores the presence of mass points; 
    
    (ii) \code{check}: looks for and reports the number of unique observations at each side of the cutoff. 
    
    (iii) \code{adjust}: sets \code{binselect()} as polynomial regression when mass points are present. 
    
    Default option is \code{masspoints=adjust}.}

  \item{hide}{logical. If \code{TRUE}, it omits the RD plot; default is \code{hide = FALSE}.}
  \item{ci}{optional graphical option to display confidence intervals of selected level for each bin.}
  \item{shade}{optional graphical option to replace confidence intervals with shaded areas.}
  \item{title}{optional title for the RD plot.}
  \item{x.label}{optional label for the x-axis of the RD plot.}
  \item{y.label}{optional label for the y-axis of the RD plot.}
  \item{x.lim}{optional setting for the range of the x-axis in the RD plot.}
  \item{y.lim}{optional setting for the range of the y-axis in the RD plot.}
  \item{col.dots}{optional setting for the color of the dots in the RD plot.}
  \item{col.lines}{optional setting for the color of the lines in the RD plot.}
  \item{data}{an optional data frame. When supplied, \code{y}, \code{x}, \code{covs}, \code{weights}, and \code{subset} may be given as bare variable names referring to columns of \code{data}.}
}
\value{
  \item{binselect}{method used to compute the optimal number of bins.}
  \item{N}{sample sizes used to the left and right of the cutoff.}
  \item{N_h}{effective sample sizes used to the left and right of the cutoff.}
  \item{c}{cutoff value.}
  \item{p}{order of the global polynomial used.}
  \item{h}{bandwidth used to the left and right of the cutoff.}
  \item{kernel}{kernel used.}
  \item{J}{selected number of bins to the left and right of the cutoff.}
  \item{J_IMSE}{IMSE optimal number of bins to the left and right of the cutoff.}
  \item{J_MV}{Mimicking variance number of bins to the left and right of the cutoff.}
  \item{coef}{matrix containing the coefficients of the \eqn{p^{th}} order global polynomial estimated both sides of the cutoff.}
  \item{coef_covs}{coefficients of the additional covariates, only returned when \code{covs()} are used.}
  \item{scale}{selected scale value.}
  \item{rscale}{implicit scale value.}
  \item{bin_avg}{average bin length.}
  \item{bin_med}{median bin length.}
  \item{vars_bins}{data frame containing the variables used to construct the bins: bin id, cutoff values, mean of x and y within each bin, cutoff points and confidence interval bounds.}
  \item{vars_poly}{data frame containing the variables used to construct the global polynomial plot.}
  \item{rdplot}{a standard \code{ggplot} object that can be used for further customization.}
}
\references{
Calonico, S., M. D. Cattaneo, M. H. Farrell, and R. Titiunik. 2017. \href{https://rdpackages.github.io/references/Calonico-Cattaneo-Farrell-Titiunik_2017_Stata.pdf}{rdrobust: Software for Regression Discontinuity Designs.} \emph{Stata Journal} 17(2): 372-404. 

Calonico, S., M. D. Cattaneo, and R. Titiunik. 2014. \href{https://rdpackages.github.io/references/Calonico-Cattaneo-Titiunik_2014_Stata.pdf}{Robust Data-Driven Inference in the Regression-Discontinuity Design.} \emph{Stata Journal} 14(4): 909-946.

Calonico, S., M. D. Cattaneo, and R. Titiunik. 2015a. \href{https://rdpackages.github.io/references/Calonico-Cattaneo-Titiunik_2015_JASA.pdf}{Optimal Data-Driven Regression Discontinuity Plots.} \emph{Journal of the American Statistical Association} 110(512): 1753-1769.

Calonico, S., M. D. Cattaneo, and R. Titiunik. 2015b. \href{https://rdpackages.github.io/references/Calonico-Cattaneo-Titiunik_2015_R.pdf}{rdrobust: An R Package for Robust Nonparametric Inference in Regression-Discontinuity Designs.} \emph{R Journal} 7(1): 38-51.

Cattaneo, M. D., B. Frandsen, and R. Titiunik. 2015. \href{https://rdpackages.github.io/references/Cattaneo-Frandsen-Titiunik_2015_JCI.pdf}{Randomization Inference in the Regression Discontinuity Design: An Application to the Study of Party Advantages in the U.S. Senate.} \emph{Journal of Causal Inference} 3(1): 1-24.
}
\author{
Sebastian Calonico, University of California, Davis, CA.  \email{scalonico@ucdavis.edu}.

Matias D. Cattaneo, Princeton University, Princeton, NJ.  \email{matias.d.cattaneo@gmail.com}.

Max H. Farrell, University of California, Santa Barbara, CA. \email{mhfarrell@gmail.com}.

Rocio Titiunik, Princeton University, Princeton, NJ.  \email{rocio.titiunik@gmail.com}.
}
\examples{
x <- runif(1000,-1,1)
y <- 5 + 3*x + 2*(x>=0) + rnorm(1000)
rdplot(y,x)
}
\keyword{ regression discontinuity }
\keyword{ RD plots }
\keyword{ binning }
\keyword{ partitioning}
\keyword{ tuning parameter selection}
\seealso{
  \code{\link{rdbwselect}}, \code{\link{rdrobust}}, \code{\link{plot.rdrobust}}
}
